#!/usr/bin/perl


############################################################################
# Software developed at the Bioinformatics Group, Department of Computer
# Science, Albert-Ludwigs-University Freiburg by
# Adapted for RNAAlignClust by Milad Miladi 2014
# Based on fasta2shrep_gspan developed by Sita Lange et al.  2011-2013
#
# Version 0.1-Align 
#
#	This script creates a GSPAN file from input sequences that are compatible
# with the NSPDK or EDEN software from Fabrizio Costa. It creates graphs
# from RNA sequences that portray sequence and structure features employing
# RNAshapes for structure prediction.
#
# 
############################################################################

#use feature ':5.10';
use File::Basename;
use lib dirname($0);    # search skript directory for modules
use strict 'vars';
use warnings;
use Getopt::Long;
use Pod::Usage;
use List::Util qw/ min max /;



use Cwd qw/abs_path getcwd/;

use FindBin;                 # locate this script
use lib "$FindBin::Bin/../lib/";  # use the parent directory

use StructureLibrary::Sequence;
use StructureLibrary::Gspan;
use StructureLibrary::Tools;
use Cwd qw(getcwd abs_path);
use File::Temp qw(tempdir);
use File::Copy;
use POSIX qw(ceil);



=head1 NAME


alignment_gspan.pl -fasta mysequences.fasta -wins "50,100,150,200" -shift 5 -M 8

=head1 SYNOPSIS

Options:

  HELP
  -help   brief help message
  -man    full documentation

  COMPULSORY
  -fasta	<STRING> e.g. "sequence.fasta"
          All sequences in fasta format.

  OPTIONS
  -wins   [INTEGER] e.g. "50,100,200"
          A list of window sizes to use. If none are given (empty string ''),
          then the entire sequence is taken with no windows. Each window > 1
          required!
  -shift  <INTEGER> e.g. 20
          The shift of the window, relative to the window size given in percent.
          So you give which percent of the window size shall be used for the
          shift. Of course the shift is rounded down to the nearest whole
          number.
          Example: 20 % of a window 150 would result in a step size of 30 nt. It
          is a relative parameter, as you can give different window sizes. If
          you do not give this parameter there is a default shift of 1 nt.
  -cue    Crop unpaired ends. If you give this flag, then the unpaired ends of
          each single structure are ignored. E.g. the structure
          ...(((...))).. becomes just (((...)))
  -stack  Adds stacking information to graphs. This adds an additional vertex
          (type P) for each pair of stacked base-pairs and four edges (type p)
          from each of the involved bases to the new vertex.
  -e      <FLOAT> e.g. 5.0
          Energy range in kcal/mol (RNAshapes)
          Use only one of -e and -c!
  -c      <INTEGER> e.g. 10
          Relative energy range, i.e. percentage (%) of MFE energy (RNAshapes)
          Use only one of -e and -c!
  -t      <INTEGER> [1-5] e.g. 3 OR "3=0,4=100,5=200"
          The shape type (RNAshapes). Default is 3.
          With the list format, the shape level can be changed for different
          window length "4=100" means that shape level 4 is used from length
          100nt (window length) The first given length has to be 0! Not
          continuous given levels are allowed!
  -M      <INTEGER> e.g. 10
          Max number of shreps that should be taken per window.
  -u      Ignore unstable structures (RNAshapes).
          This option filters out closed structures with positive free energy.
  -r      Calculate structure probabilities for shreps (RNAshapes)
  -i      <INT> e.g. 10
          Turn on structure sampling and gives number of sampling iterations.
          Default no sampling (i=0)
  -sample-len 
          <INT> e.g. 100
          Only in sampling mode: Sampling is only used for seqs/windows >= given
          length, Default: sample all lengths (0), if -i > 0
  -q      Turn on shape probabilities for RNAshapes, no sampling mode allowed
  -Tp     <FLOAT> e.g 0.001
          Filter cutoff for shape probabilities, applied before -M filter!
  -seq-graph-win  
          add for each window a graph which contains no structure
  -seq-graph-t    
          add for each 't #' a graph which contains no structure
  -seq-graph-alph 
          change the alphabet of unstructured graphs
  -annotate    
          <STRING> annotation.tab
          A file with annotations to be added as abstract graphs on the sequence
          leven (if given) and on the structure (SHREP) level. The format is has
          the following TAB-delimited columns: SEQID, START, END,
          NAMESPACE#LABEL. Labels with the same name-space and SEQID form
          connected components, which is a sequence of label vertices ordered by
          the START position in the sequence.
  -abstr  Add abstract structure graphs to the single shrep graph instances.
  -abstr-out      
          <STRING> write sequence representation of abstract structure graphs to
          this file
  -nostr  Calculate no structures, only add sequence information,
          if this is given, then -seq-graph-win AND/OR -seq-graph-t
          are required.
  -match-shape
          <SHAPE>
          all seqs/windows will be constraint folded into that shape via
          RNAshapes (if structure is given in another way this struct will be
          kept), if this shape is not possible within given energy range,
          produce a specific t graph with only one vertex 'X'. By this the
          instance becomes very unsimilar to all other graphs (for knn)
  -vp     enable graph computation with viewpoints:
          svmsgdnspdk will center on those nucleotides that are given via
          capital letters and ignore those given as lowercase letters
  -tmp    <STRING> e.g. "/scratch/1/sita/tmp"
          A directory for writing temporary files
  -o      <STRING> e.g. "ProjectX/MySequences/GSPAN/"
          Output directory for gspan files containing graphs.
  -group  <INTEGER> e.g. 5
          Combine/group that number of input seqs into 1 gspan file output name
          is then '<INT>.group.gspan.bz2'
  -sge    Use SGE cluster for each sequence separately
  -sge-submit-host
          <STRING>
          host used to submit SGE jobs
  -sge-submit-user
          <STRING>
          user for ssh access to submit host. needed if called from a clustered
          job because with these getlogin() automation fails.
  -sge-script
          <STRING>
          Use this file instead of the default sge-submit-script.
  -sge-logdir
          stdout directory for SGE call
  -sge-errdir
          sdterr directory for SGE call
  -stdout send graphs to stdout instead of writing to files
  -ignore-header
          don't write fasta id part after first space to gspan
  -debug  additional debug output


  DEFAULT VALUES
  -wins   ""
  -shift  1 nt
  -c      10
  -t      3
  -M      0 # selects all shreps
  -tmp    "/var/tmp/fasta2shrep"
  -o      "CURRENT_DIR/GSPAN/"

  SGE mode
  -sge-submit-host  "biui.informatik.uni-freiburg.de"
  -sge-logdir       "CURRENT_DIR/GSPAN/SGE_log"
  -sge-errdir       "CURRENT_DIR/GSPAN/SGE_log"

=head1 DESCRIPTION

=cut

###############################################################################
# end handler for temporary directory
# adds an error handler that deletes the directory in case of error
# SIGUSR{1/2} are sent by the sge prior to the uncatchable SIGKILL if the
# option -notify was set
###############################################################################
$SIG{'INT'}  = 'end_handler';
$SIG{'TERM'} = 'end_handler';
$SIG{'ABRT'} = 'end_handler';
$SIG{'USR1'} = 'end_handler';
$SIG{'USR2'} = 'end_handler';

sub end_handler {
  print STDERR "signal ", $_[0], " caught, cleaning up temporary files\n";

  # change into home directory. deletion of the temporary directory will
  # fail if it is the current working directory
  chdir();
  File::Temp::cleanup();
  die();
}

###############################################################################
# PARSE COMMAND LINE OPTIONS
###############################################################################

# command line options
my (
  $i_help,               $i_man,          $i_debug,
  $i_fas,                $i_wins,         $i_shift,
  $i_crop_unpaired_ends, $i_r,            $i_e,
  $i_c,                  $i_t,            $i_u,
  $i_M,                  $i_o,            $i_sge,
  $i_sge_submit_host,    $i_sge_submit_user,
  $i_sge_script,         $i_jobid,        $i_tmp,
  $i_ignore_seq_header,  $i_stacks,       $i_stdout,
  $i_q,                  $i_T,            $i_i,
  $i_sample_min_length,  $i_sge_logDir,   $i_sge_errDir,
  $i_groupsize,          $i_annotate,     $i_abstr,
  $i_abstr_out,          $i_no_structure, $i_vp,
  $i_matchShape
);

my ( $i_add_seq_graph_win, $i_add_seq_graph_t, $i_change_seq_graph_alph );

my $options = GetOptions(
  "help"            => \$i_help,
  "man"             => \$i_man,
  "debug"           => \$i_debug,
  "fasta=s"         => \$i_fas,
  "wins=s"          => \$i_wins,
  "shift=f"         => \$i_shift,
  "cue"             => \$i_crop_unpaired_ends,
  "stack"           => \$i_stacks,
  "r"               => \$i_r,
  "e=f"             => \$i_e,
  "c=i"             => \$i_c,
  "t=s"             => \$i_t,
  "u"               => \$i_u,
  "M=i"             => \$i_M,
  "tmp=s"           => \$i_tmp,
  "o=s"             => \$i_o,
  "i=i"             => \$i_i,
  "sample-len=i"    => \$i_sample_min_length,
  "q"               => \$i_q,
  "Tp=f"            => \$i_T,
  "seq-graph-win"   => \$i_add_seq_graph_win,
  "seq-graph-t"     => \$i_add_seq_graph_t,
  "-seq-graph-alph" => \$i_change_seq_graph_alph,
  "sge"             => \$i_sge,
  "sge-submit-host=s" => \$i_sge_submit_host,
  "sge-submit-user=s" => \$i_sge_submit_user,
  "sge-script=s"    => \$i_sge_script,
  "jobid=s"         => \$i_jobid,
  "ignore-header"   => \$i_ignore_seq_header,
  "stdout"          => \$i_stdout,
  "sge-logdir=s"    => \$i_sge_logDir,
  "sge-errdir=s"    => \$i_sge_errDir,
  "group=i"         => \$i_groupsize,
  "annotate=s"      => \$i_annotate,
  "abstr"           => \$i_abstr,
  "abstr-out=s"     => \$i_abstr_out,
  "nostr"           => \$i_no_structure,
  "vp"              => \$i_vp,
  "match-shape=s"   => \$i_matchShape,
);

pod2usage( -exitstatus => 1, -verbose => 1 ) if $i_help;
pod2usage( -exitstatus => 0, -verbose => 2 ) if $i_man;
($options) or pod2usage(2);

# check compulsory options
($i_fas) or pod2usage("Error: the option -fasta is compulsory!\n");
( -e $i_fas ) or pod2usage("Error: no such file - $i_fas!\n");
$i_fas = abs_path($i_fas);

# check other options and set default values
pod2usage("Error: either -e OR -c can be given, but NOT both!\n")
  if ( $i_e && $i_c );
( $i_e or $i_c ) or $i_c = 10;    # set -c 10, if neither -e or -c are given
($i_M) or $i_M = 0;    # max number of shreps is 0 (=means take all computed)
($i_i) or $i_i = 0;    # default no sampling else sampling iterations
($i_sample_min_length) or $i_sample_min_length = 0;
pod2usage("\nError: use --sample-len <INT> only with -i <1..INT> !\n")
  if ( not $i_i and $i_sample_min_length );
($i_T) or $i_T = 0;
($i_q) or $i_q = 0;
pod2usage(
  "\nError: Sampling (-i) not possible with shape probabilities (-q)!\n")
  if ( $i_i and $i_q );
($i_add_seq_graph_win)     or $i_add_seq_graph_win     = 0;
($i_add_seq_graph_t)       or $i_add_seq_graph_t       = 0;
($i_change_seq_graph_alph) or $i_change_seq_graph_alph = 0;
($i_no_structure)          or $i_no_structure          = 0;

if ($i_change_seq_graph_alph) {
  ($i_add_seq_graph_t)
    or ($i_add_seq_graph_win)
    or pod2usage( "Error: "
      . "When giving the parameter -seq-graph-alph, then either -seq-graph-t or -seq-graph-win"
      . " must also be given!\n" );
}
( -e $i_annotate )
  or pod2usage("Error: no such file - $i_annotate!\n")
  if ($i_annotate);

(defined $i_sge_submit_host) or $i_sge_submit_host = 'biui.informatik.uni-freiburg.de';

# set the appropriate exports for old and new cluster respectively
my $SGE_BIN_PATH;
my $SGE_EXPORTS;

if ($i_sge_submit_host eq 'biui.informatik.uni-freiburg.de') {
  $SGE_BIN_PATH='/opt/sge-6.0/bin/lx24-amd64/';
  $SGE_EXPORTS='export SGE_ROOT=/opt/sge-6.0/;';
} elsif ($i_sge_submit_host eq 'ui.bi.uni-freiburg.de') {
  $SGE_BIN_PATH='/opt/ogs/bin/linux-x64/';
  $SGE_EXPORTS='export SGE_ROOT=/opt/ogs; export SGE_CELL=BI1;';
}

$i_add_seq_graph_t = 1 if ($i_no_structure);

($i_t) or $i_t = 3;    # default abstraction type is 3
my $change_shape_level = 0;
my @level_lens = ( -1, -1, -1, -1, -1 )
  ;    ## array_idx-1=shape level, value=start length of this level

if ( $i_t !~ /^\d+$/ ) {

  my @t_minlens = split( ",", $i_t );    ## -t "3=100,4=200"

  foreach my $idx ( 1 .. @t_minlens ) {

    my $level = $t_minlens[ $idx - 1 ];    ##  $level = "3=100"

    die "$level Wrong -t format! Example: -t 3=0,4=100,5=200\n"
      if ( $level !~ /^\d+\=\d+$/ );
    my @lev_len = split( "=", $level );    ##  $level = "3=100"

    die
"Wrong -t format! First level given needs to be length 0! Example: -t 3=0,4=100,5=200\n"
      if ( $idx == 1 && $lev_len[1] != 0 );
    die "Wrong -t format! Only level 1-5 allowed! Example: -t 3=0,4=100,5=200\n"
      if ( $lev_len[0] < 1 or $lev_len[0] > 5 );
    die "Wrong -t format! Length >= 0 expected! Example: -t 3=0,4=100,5=200\n"
      if ( $lev_len[1] < 0 );
    $change_shape_level = 1;
    $level_lens[ $lev_len[0] - 1 ] = $lev_len[1];
  }

  ($i_debug)
    and print STDERR
    "i_t = $i_t - change $change_shape_level -  shape level lengths#"
    . join( ":", @level_lens ) . "#"
    . join( ":", @t_minlens ) . "#\n";
}



my $CURRDIR = getcwd;

# set up tmp directory
# default tmp is /var/tmp, usually not NFS mounted!
( defined $i_tmp ) or $i_tmp = '/var/tmp/';
my $tmp_template = 'fasta2shrep-XXXXXX';

# CLEANUP => 1 : automatically delete at exit
$i_tmp = tempdir( $tmp_template, DIR => $i_tmp, CLEANUP => 1 );

# create GSPAN directory when not printing to stdout
if ( not $i_stdout ) {
  if ($i_o) {
    ( -e $i_o ) or system("mkdir -p $i_o");
  } else {
    system("mkdir -p GSPAN");
    $i_o = $CURRDIR . "/GSPAN/";
  }
}

###############################################################################
# GLOBAL VARIABLES
###############################################################################

my @WINDOWS = ();
@WINDOWS = split( ",", $i_wins ) if ($i_wins);
my $globalFolding;
$globalFolding = 1 unless @WINDOWS;
my $CURRUSER = getlogin;
(defined $i_sge_submit_user) and $CURRUSER=$i_sge_submit_user;
my $SEQNO    = 0;          # used to id the sequences
my $GSPANNO  = 0;          # used for gspan filenames

# minimum length of sequence that will work with RNAshapes
# sequences with one or two nucleotides will be restricted to sequence-only graphs
# this ensures that no sequences are skipped and external info kept synchronized
my $GSPAN_SEQ_MINLEN = 3;

##for alignments we set one zip file per alignment file
my $one_zip_per_alignment;
###############################################################################
# EXECUTION CODE
###############################################################################

# read fasta file into hash
my ( $headers_aref, $sequences_aref, $metainfo_aref ) =
  read_fasta_with_nonunique_headers_meta($i_fas);

    my @parts_cons = get_path_base_suffix($i_fas); 
	
	#my $file_path = $parts_cons[0];
	my $alignment_file_name = $parts_cons[1];
	#my $file_extension = $parts_cons[2];
# call script again on the sge cluster in a batch job
if ($i_sge) {
  my $SCRIPTNAME = $0;
  my $CLUSTERSUBMIT;
  if ( defined $i_sge_script ) {
    $CLUSTERSUBMIT = $i_sge_script;
  } else {
    $CLUSTERSUBMIT =
      $FindBin::Bin . "/StructureLibrary/generic_submit_to_cluster.sh";
  }

  die "Cannot find SGE submit script $CLUSTERSUBMIT! Exit...\n\n"
    if ( !-e $CLUSTERSUBMIT );

  my $SGE_jobs = @{$sequences_aref};
  $SGE_jobs = ceil( @{$sequences_aref} / $i_groupsize ) if ($i_groupsize);

  my $params = "-fasta $i_fas ";
  $params .= "-wins $i_wins "                     if ($i_wins);
  $params .= "-shift $i_shift "                   if ($i_shift);
  $params .= "-e $i_e "                           if ($i_e);
  $params .= "-c $i_c "                           if ($i_c);
  $params .= "-t $i_t "                           if ($i_t);
  $params .= "-u "                                if ($i_u);
  $params .= "-r "                                if ($i_r);
  $params .= "-M $i_M "                           if ($i_M);
  $params .= "-o $i_o "                           if ($i_o);
  $params .= "-i $i_i "                           if ($i_i);
  $params .= "--sample-len $i_sample_min_length " if ($i_sample_min_length);
  $params .= "-q "                                if ($i_q);
  $params .= "-Tp $i_T "                          if ($i_T);
  $params .= "--seq-grap-win "                    if ($i_add_seq_graph_win);
  $params .= "--seq-graph-t "                     if ($i_add_seq_graph_t);
  $params .= "--seq-graph-alph "                  if ($i_change_seq_graph_alph);
  $params .= "-ignore-header "                    if ($i_ignore_seq_header);
  $params .= "-cue "                              if ($i_crop_unpaired_ends);
  $params .= "-stack "                            if ($i_stacks);
  $params .= "--group $i_groupsize "              if ($i_groupsize);
  $params .= "-annotate $i_annotate "             if ($i_annotate);
  $params .= "-abstr "                            if ($i_abstr);
  $params .= "-nostr "                            if ($i_no_structure);
  $params .= "--debug "                           if ($i_debug);
  $params .= "-vp "                               if ($i_vp);
  $params .= "--match-shape $i_matchShape "       if ($i_matchShape);

  print "used script:" . $SCRIPTNAME . "\n";
  print "used submit script:" . $CLUSTERSUBMIT . "\n";

  $i_sge_logDir = "$i_o/SGE_log" if ( !$i_sge_logDir );
  mkdir($i_sge_logDir);

  $i_sge_errDir = $i_sge_logDir if ( !$i_sge_errDir );
  mkdir($i_sge_errDir);

  my $ssh = 1;    ## can be used for debug shell script call
  if ($ssh) {
    my $syscall =
        "ssh $CURRUSER\@$i_sge_submit_host "
      . "'cd $CURRDIR; "
      . "$SGE_EXPORTS "
      . "$SGE_BIN_PATH/qsub -t 1-$SGE_jobs -o $i_sge_errDir/ -e $i_sge_errDir/ "
      . "$CLUSTERSUBMIT $CURRDIR $SCRIPTNAME \"$params\" ' ";
    say STDERR "starting SGE job:";
    say STDERR $syscall;
    system($syscall );
  } else {
    system("$CLUSTERSUBMIT $CURRDIR $SCRIPTNAME '$params' ");
  }
  exit;
}

## compute shreps and gspan, either local or after SGE submission

# TODO read and process annotations

my @used_seq_headers;
my @used_seqs;
my @used_meta;
my $group_idx;

if ($i_jobid) {
  ## just process the one sequence as given by the jobid number
  my $used_grouping = 1;   ## if no group is given, make 1 seq per job = group=1
  $used_grouping = $i_groupsize if ($i_groupsize);
  my $st  = ( $i_jobid - 1 ) * $used_grouping + 1;
  my $end = $st + $used_grouping - 1;
  $end = @{$sequences_aref} if ( $end > @{$sequences_aref} );

  foreach my $idx ( $st .. $end ) {
    push( @used_seq_headers, $headers_aref->[ $idx - 1 ] );
    push( @used_seqs,        $sequences_aref->[ $idx - 1 ] );
    push( @used_meta,        $metainfo_aref->[ $idx - 1 ] );
  }

  $group_idx = $i_jobid;

#($i_debug) and print STDERR "st $st end $end gr $group_idx job $i_jobid gs $i_groupsize\n";

  $GSPANNO = $i_jobid - 1 if ( !$i_groupsize );

} else {
  ## process all sequences at once
  @used_seq_headers = @{$headers_aref};
  @used_seqs        = @{$sequences_aref};
  @used_meta        = @{$metainfo_aref};
}

my $out;
my $gspanfile;
my $out_no_match_shape;
my $out_abstr;

if ( $i_matchShape && !$i_stdout && !$i_groupsize ) {
  open( $out_no_match_shape, ">$i_o/fasta2shrep.no_match" );
}

# set up file for abstract annotation output
if ( $i_abstr and defined $i_abstr_out ) {
  open( $out_abstr, '>', $i_abstr_out )
    or die("error: could not open file '$i_abstr_out'");
}
 
 #initilize gi counter outside seq iterations
 my $gi = 0;
 my $loop_counter = 0;		
# for each sequence in the fasta file
while ( my $seq = shift @used_seqs ) {
	
  $loop_counter++;		
  my $tmp_header = shift @used_seq_headers;
  my $tmp_meta   = shift @used_meta;

  $i_debug and say STDERR "calculating for sequence " . $tmp_header;

  my ( $seq_id, $seq_header ) = ( $tmp_header =~ /(\S+)\s*([\S*\s*]*)/ );
  $i_ignore_seq_header and $seq_header = '';

  my $seq_fasta = generate_single_fasta_from_sequence_X( $seq_id, $seq );
  my $seq_len = length($seq);

  # only print sequence graphs for sequences below this threshold
  my $no_structure_override = ( $seq_len < $GSPAN_SEQ_MINLEN ) ? 1 : 0;

  $GSPANNO++;

  # set outstream for gspan output to correct file/STDOUT
  if ($i_stdout) {
    $out = \*STDOUT;
  } elsif ( !$i_groupsize ) {
    $gspanfile = $i_tmp . '/' . $GSPANNO . '.gspan';
    open( $out, "| bzip2 -f > $gspanfile.bz2" );
  } elsif ( ( $GSPANNO - 1 ) % $i_groupsize == 0 ) {

    if ( $GSPANNO > 1 ) {
      close($out);
      system("mv $gspanfile.bz2 $i_o/$group_idx.group.gspan.bz2");
      if ($out_no_match_shape) {
        close($out_no_match_shape);
        system("mv $gspanfile.no_match $i_o/$group_idx.group.gspan.no_match");
      }
    }

    if ( !$i_jobid ) {
      $group_idx = int( ( $GSPANNO - 1 ) / $i_groupsize ) + 1;
    }

    $gspanfile = "$i_tmp/$group_idx.group.gspan";
    open( $out,                "| bzip2 -f > $gspanfile.bz2" );
    open( $out_no_match_shape, ">$gspanfile.no_match" )
      if ($out_no_match_shape);
  }

  ## do not use folding windows in special cases
  if ( $globalFolding || $i_no_structure || $no_structure_override ) {
    @WINDOWS = ();
    push( @WINDOWS, $seq_len );
  }

  ##check win sizes for global folding
  my @WINDOWS_used = sort { $a <=> $b } @WINDOWS;
  foreach my $w_idx ( 0 .. ( @WINDOWS_used - 1 ) ) {
    if ( $WINDOWS_used[$w_idx] >= $seq_len && $WINDOWS_used[$w_idx] > 1 ) {
      if ( $w_idx < $#WINDOWS_used ) {
        @WINDOWS_used = @WINDOWS_used[ 0 .. $w_idx ];
        last;
      }
    }
  }

  ## use seq graph only if no shape folding wanted
  $i_add_seq_graph_t = 1 if ( $i_no_structure || $no_structure_override );

  ## no shape info in graphheader if we have fixed structure (according LocaRNA handling)
  ## use tags #FS or #S for provided structure
  my $graph_header;
  my @struct_meta = grep { $_ =~ /#FS/ || $_ =~ /#S/ } keys %{$tmp_meta};

  ## Milad: print alignment header/formely graph header, only once, todo: move this part above the loop 	
#  if($loop_counter == 1)
  {	
	  if (@struct_meta) {
	    $graph_header = getAlignmentHeader(
	      $alignment_file_name, #,$seq_id,
	                    $seq_header, \@WINDOWS,             $i_shift,
	      $i_e,                 $i_c,        $i_t,                  $i_u,
	      $i_r,                 $i_M,        $i_crop_unpaired_ends, $i_i,
	      $i_sample_min_length, $i_q,        $i_T,                  $seq_len,
	      1
	    );
	  } else {
	    $graph_header = getAlignmentHeader(
	      $alignment_file_name, #,$seq_id,
	      $seq_header,
	      \@WINDOWS,             $i_shift,
	      $i_e,                  $i_c,
	      $i_t,                  $i_u,
	      $i_r,                  $i_M,
	      $i_crop_unpaired_ends, $i_i,
	      $i_sample_min_length,  $i_q,
	      $i_T,                  $seq_len,
	      $i_no_structure || $no_structure_override
	    );
	  }
 if($loop_counter == 1)
 		{ print $out $graph_header; }
  }
#lets consider all the sequences a single graph
#  my $gi = 0;    # current graph index

  ## add graph with no structure at all depending on $i_add_seq_graph_t
  if ($i_add_seq_graph_t) {

    ($gi) = convertSeqWindow( $seq, $seq_len, 1, $gi, $graph_header, $out,
      $i_annotate, $seq );

  }

  ## encode fixed structure only if provided and structures wanted in general
  if ( @struct_meta && !$i_no_structure && !$no_structure_override ) {

    my $struct_meta = "#FS";
    $struct_meta = "#S" if ( !exists $tmp_meta->{$struct_meta} );

    my $seq_shrep =
      [ $tmp_meta->{$struct_meta}, "ENERGY", "0.00", "SHAPE", $struct_meta ];
    $gi =
      convertShapeWindow( [$seq_shrep], $seq, $seq_len, 1, $gi, $out,
      $graph_header, $i_annotate, $i_abstr, $i_crop_unpaired_ends, $i_stacks,
      $seq );
    @WINDOWS_used = ();    ## no shape folding if we have a fixed structure

  }

  ## ignore RNAshapes folding if wanted, but do correct file move afterwards
  ## (just "next" does not work due to output)
  @WINDOWS_used = () if ( $i_no_structure or $no_structure_override );

#  #for each window size in list
#  foreach my $win_size (@WINDOWS_used) {
#
#    ## no (match) shape found at all for this seq
#   
#  }    ## foreach WINDOW_used

  
  if ( !$i_stdout && !$i_groupsize && !$one_zip_per_alignment) {
    close($out);
    move "$gspanfile.bz2", "$i_o/$GSPANNO.gspan.bz2";
  }

  system("rm $seq_fasta");

}    ## while @used_seqs


if($one_zip_per_alignment)
{
  close($out);
  system("mv $gspanfile.bz2 $i_o/$alignment_file_name.gspan.bz2");
  close($out_no_match_shape) if ($out_no_match_shape);
  move "$gspanfile.no_match", "$i_o/$alignment_file_name.gspan.no_match"
    if ($out_no_match_shape);
}
elsif ($i_groupsize) {
  close($out);
  system("mv $gspanfile.bz2 $i_o/$alignment_file_name.gspan.bz2");
  close($out_no_match_shape) if ($out_no_match_shape);
  move "$gspanfile.no_match", "$i_o/$one_zip_per_alignment.gspan.no_match"
    if ($out_no_match_shape);
} elsif ( $out_no_match_shape && !$i_stdout && !$i_groupsize ) {
  close($out_no_match_shape);
}

###############################################################################
# METHODS
###############################################################################
############################################################################
# Generates fasta file for a single sequence (one-lined-fasta). This
# fasta is stored in the temp directory and should be deleted at the end.
# Input:
# seq_id : the sequence ID
# seq : the sequence
#
# Output:
# The fasta file name
############################################################################
sub generate_single_fasta_from_sequence_X {
  my ( $seq_id, $seq ) = @_;

  $seq = uc($seq);
  $seq =~ tr/T/U/;
  $seq =~ s/[^AUCGN]/N/g;

  my $outfas = $i_tmp . "/seq_" . $SEQNO++ . ".fasta"; 
  my $host   = readpipe("hostname");
  open( FAS, ">$outfas" ) or die "$host Cannot open file $outfas! Exit...\n\n";
  print FAS ">$seq_id\n$seq";
  close(FAS);

  return $outfas;
}
##################################################################################
# This method copied from StructureLibrary::Sequence.pm
# This method parses a fasta file and is useful if the header ID is not-unique!!
# It returns the header lines in an array and the sequence lines in the same
# order. It is then up to the user to extract the parts from the header that is
# necessary for the script.
# Furthermore, the method deals with multiple lines, and returns a single sequence
# without line breaks.
#
# In addition this function is able to parse additional meta info given with each
# sequence. This can be used e.g. to provide a given structure or a
# folding constraint string with a sequence.
# the format of the meta info is:
# >seqid header
# AUGCUGCUGUAAGAGUAGUAGUAGAUG\n
# ............((((.......)))) #/\S+/\n
#
# The only required part is " #/\S+/", this means an name or identifier for that
# meta info. (#\S+) is used as hash key for in the meta-info hash for that sequence.
# meta infos with tha same key for one seq are concatanated like a seq.
# All meta infos are returned as array of hash-refs, like for header and seqs,
# but using a hash
#
#
# Input:
#		file		The name of the fasta file
# Output:
#	(1)	An array reference for each header line
#	(2) An array reference for each sequence in same order as the headers
##################################################################################
sub read_fasta_with_nonunique_headers_metaLocal {
  my ($file) = @_;
  my $FUNCTION = "read_fasta_file in Sequences.pm";

  my $header    = "";
  my $seqstring = "";
  my @headers   = ();
  my @sequences = ();
  my @meta      = ();
  my %seq_meta  = ();
  open( IN_HANDLE, "<$file" ) || die "ERROR in $FUNCTION:\n" . "Couldn't open the following file in package Tool," . " sub read_fasta_file: $file\n";

  while ( my $line = <IN_HANDLE> ) {
    chomp($line);

    # header (can contain one space after > symbol)
    if ( $line =~ /^\>(.*)/ ) {
      if ($header) {
        $seqstring =~ s/\s*//g;    ## do not allow spaces in sequence
        push( @headers,   $header );
        push( @sequences, $seqstring );
        push( @meta,      +{%seq_meta} );    ## anonymous hash reference
#        print keys %seq_meta;
        $seqstring = "";
        undef(%seq_meta);
      }
      $header = $1;

    } elsif ( $line =~ /(.+)\s+(#\S+)$/ && $header ) {

      if ( exists $seq_meta{$2} ) {
        $seq_meta{$2} .= $1;
      } else {
        $seq_meta{$2} = $1;
      }

    } elsif ($header) {
      $seqstring .= $line
    }
  }

  if ($header) {
    $seqstring =~ s/\s*//g;    ## do not allow spaces in sequence
    push( @headers,   $header );
    push( @sequences, $seqstring );
    push( @meta,      +{%seq_meta} );
    $seqstring = "";
    %seq_meta  = ();
  }
  return ( \@headers, \@sequences, \@meta );
}
###########################################################################
# This method copied from Structure::Gspan.pm
# This method gives us the overall graph header for this sequence. This
# graph includes all window calculations and all shreps for all windows.
# Input:
# seq_id : the sequence ID
# seq_head : the header information for the sequence
# wins_aref : the windows sizes for folding
# h_shift : input parameter -shift
# h_e : input parameter -e
# h_c : input parameter -c
# h_t : input parameter -t
# h_u : input parameter -u
# h_r : input parameter -r
# h_M : input parameter -M
# h_crop_unpaired_ends : input parameter -cue
#
# Output:
# The header line as a string
############################################################################
sub getAlignmentHeader {
  my (
    $seq_id,       $seq_head, $wins_aref,            $h_shift,
    $h_e,          $h_c,      $h_t,                  $h_u,
    $h_r,          $h_M,      $h_crop_unpaired_ends, $h_i,
    $h_sample_len, $h_q,      $h_Tp,                 $h_seqlen,
    $h_noStr
  ) = @_;

  my $ret;

  $ret = "t # SEQID $seq_id ";
  $ret .= "$seq_head " if ( defined $seq_head );
  $ret .= "SEQLEN $h_seqlen ";

  return $ret . "\n" if ($h_noStr);

  $ret .= "MAXWINSHREPS $h_M ";
  $ret .= "CUE 1 " if ($h_crop_unpaired_ends);
  $ret .= "RNASHAPES -w " . ( join( ",", @{$wins_aref} ) ) . " ";
  $ret .= "-SHIFT%w $h_shift " if ( not $globalFolding and $h_shift ); #todo: globalFolding is a global variable 
  $ret .= "-SHIFT 1 " if ( not $globalFolding and not $h_shift );
  $ret .= "-e $h_e "                         if ( defined $h_e );
  $ret .= "-c $h_c "                         if ( defined $h_c );
  $ret .= "-t $h_t "                         if ( defined $h_t );
  $ret .= "-u 1 "                            if ($h_u);
  $ret .= "-r 1 "                            if ($h_r);
  $ret .= "-i $h_i SAMPLE_LEN $h_sample_len" if ( not $h_q and $h_i );
  $ret .= "-q 1 "                            if ($h_q);
  $ret .= "-T $h_Tp"                         if ( $h_q and defined $h_Tp );
  $ret .= "\n";

  return $ret;
}

1;

#############################################################################
# Programming description
#############################################################################
#	Substructure graphs for machine learning with Fabrizio
#	-------------------------------------------------------
#
#	(1) Parameters (RNAshapes parameter):
#		- Window sizes [] (-w)
#		- window shift size (-W)
#	- calculate structure probabilities (-r)
#		- energy range kcal/mol (-e) OR energy relative percentage (%) to MFE (-c)
#		- shape type 1 to 5 (-t)
#		- ignore unstable substructures (-u)
#		- max shreps
#
#
#	(2) For each sequence, generate one graph/file that consists of all windows. The general format for one graph is as follows:
#
#	t # seq_id parameters
#	v graph_index nt_type window_size window_centre abs_seq_pos
#	...
#	e m m+1 > 1 (backbone)
#	...
#	e base_i_graph_index base_j_graph_index s shrep_e shrep_p ...
#
#	For each window (subgraph) we create a subgraph (of the subgraph) for each substructure.
#	We have a running index (gi) for each subgraph. All vertex and edge indices of the subgraph add
#	the running graph index to the actual window position. For example
#
#
#	Sequence: AAACC CUUUG GG
#		  01234 56789 01
#
#	Window=10 substructure1 = (((...))). centre 5.5
#
#	v 0 A 10 5.5 0
#	v 1 A 10 5.5 1
#	v 2 A 10 5.5 2
#	v 3 C 10 5.5 3
#	v 4 C 10 5.5 4
#	v 5 C 10 5.5 5
#	v 6 U 10 5.5 6
#	v 7 U 10 5.5 7
#	v 8 U 10 5.5 8
#	v 9 G 10 5.5 9
#	e 0 1 > 1
#	e 1 2 > 1
#	e 2 3 > 1
#	e 3 4 > 1
#	e 4 5 > 1
#	e 5 6 > 1
#	e 6 7 > 1
#	e 7 8 > 1
#	e 8 9 > 1
#	e 0 8 s -15.0 0.1223
#	e 1 7 s -15.0 0.1223
#	e 2 6 s -15.0 0.1223
#
#	gi = 9+1 = 10
#
#	Window = 10 substructure2 = .(((...))) centre 6.5
#	v 10 A 10 6.5 2
#	v 11 C 10 6.5 3
#	v 12 C 10 6.5 4
#	v 13 C 10 6.5 5
#	v 14 U 10 6.5 6
#	v 15 U 10 6.5 7
#	v 16 U 10 6.5 8
#	v 17 G 10 6.5 9
#	v 18 G 10 6.5 10
#	v 19 G 10 6.5 11
#	e 10 11 > 1
#	e 11 12 > 1
#	e 12 13 > 1
#	e 13 14 > 1
#	e 14 15 > 1
#	e 15 16 > 1
#	e 16 17 > 1
#	e 17 18 > 1
#	e 18 19 > 1
#	e 11 19 s -17.0 0.156
#	e 12 18 s -17.0 0.156
#	e 13 17 s -17.0 0.156
#
#	gi = 19+1 = 20
#
#
#
#	Write one perl script to create graphs for a set of sequences, fasta2shrep_gspan.pl.
#
#	INPUT:
#		-f fasta file with all sequences to compute
#		parameters as above
#
#	OUTPUT:
#		one file per sequence that contains graph called seq_id.gspan
#
#	(1) for each window size call RNAshapes and write to a tmp file
#	(2) parse result of RNAshapes (catch RNAshapes error - sequence too long?) check for max shreps.
#	(3) convert RNAshapes result to subgraph -> write to file (readpipe) look at efficiency and errors
#	(4) repeat (1) to (3) for each sequence

# ABSTRACT STRUCTURE GRAPH
# To each shrep graph (s #) add an abstract structure graph with special
# abstract relations. For example, to add the abstract structure graph to
# the previous shrep structure with gi numbers from 10 to 19, we first
# identify the abstract shape, i.e. EL-S-HL and then create nodes (labelled
# with a given name-space followed the actual node name and separated by a
# hash) and edges for this graph as follows:
# NOTE: at the moment we add the adjacent base-pairs to the loop definitions
# but these can be removed if necessary in the future
# v 20 abstruct#EL
# v 21 abstruct#S
# v 22 abstruct#HL
# e 20 21 n
# e 21 22 n
# v 23 ^R
# e 20 23 ^r
# e 23 10 @r
# e 23 11 @r
# e 23 19 @r
# v 24 ^R
# e 21 24 ^r
# e 24 11 @r
# e 24 12 @r
# e 24 13 @r
# e 24 17 @r
# e 24 18 @r
# e 24 19 @r
# v 25 ^R
# e 22 25 ^r
# e 25 13 @r
# e 25 14 @r
# e 25 15 @r
# e 25 16 @r
# e 25 17 @r

# gi = 25+1 = 26

# ANOTATION FILE
# This is a file that labels a sequence region with a given annotation. In one
# file we can have annotations within different name-spaces, for example target
# sites predicted with different tools.
# File format is a tab-delimited file with the following columns:
# SEQID (same ID as in the fasta file header)
# i (left positon of region)
# j (right position of region, if one position i=j)
# NAMESPACE#LABEL (give each annotation type one name-space and choose a label depending on the task)

# E.g we have 2 different miRNAs and we predict the target-sites with (1) IntaRNA and (2) TargetSearch,
# then this could be IntaRNA#miR1, IntaRNA#miR2, TargetSearch#miR1, and TargetSearch#miR2.
# All labels can be used more than once for one or more sequences.
# all labels with the same namespace and the same sequence ID are grouped into
# one abstract graph, according to the order of i.
# Example
# SEQID 	i 	j 	NAMESPACE#LABEL
# s1	10 	20	IntaRNA#miR1
# s1	54	60	IntaRNA#miR2
# s1	15	25	TargetSearch#miR1
# s1	54	60 	TargetSearch#miR2
# s2 ...

# We create connected component per sequence ID per name-space,
# so that all labels with
# the same namespace are grouped together (for one sequence) as follows:
# v 26 IntaRNA#miR1
# v 27 IntaRNA#miR2
# e 26 27 n
# then create R nodes and ^r and @r relations as before (from i to j)

# If a sequence graph option is given, then these graphs (u #) have to
# be labelled with this annotation

