#!/bin/bash

################################################################################ 
## This script evaluates Graph Kernel similarity features for an input gspan graph file.
## Evaluation is performed by calculating the portion of nearest neighboorhod 
## objects of each element where both element and its nearest neighbour belong 
## to the same refernce family.
## Milad Miladi miladim@informatik.uni-freiburg.de
################################################################################
# ./analsys.sh gspan_file output_dir metric class_column
################################################################################
##path to the graph kernel
#EDeN='/usr/local/user/EDeN/1.2.0/EDeN'
#EDeN='/home/projects/cmf_cluster/EDeN/precompiled/EDeN'
EDeN='/home/ajunge/bin/EDeN-1.3.7/EDeN'
#EDeN='/home/ajunge/bin/EDeN/EDeN'
NSPDK='/home/projects/cmf_cluster/GraphClust/installed/bin/NSPDK'
BIN_DIR="$(dirname "$(readlink -f "$0")")"

## parsing input  parameters
gspan_file=$1
output_dir=$2
metric=$3 ##KNN, ROC,APR,ROC50
class_column=$4 # column where unique class name can be found on gspan feature header statred with #t
		# 4 for RNAalignClust, 9 for GraphClust
cluster_options=$5 # options for clustering, e.g.
                   # "--cluster_type K_QUICK_SHIFT -r 3 -d 5 --approximate --cluster_threshold 1"

echo -e "\nPERF $metric analysis for gspan file:\n\t$1\n"

## param validity check
if [ ! -n "$gspan_file" ] || [ ! -n "$output_dir" ] || [ ! -n "$metric" ] || [ ! -n "$class_column" ]
then
	echo "ERROR: gspan output_dir, metric and class_column number are mandatory. Format is:"
	echo -e "./analsys.sh gspan_file output_dir metric class_column\n\n "
	exit
fi 

if [ ! -f $gspan_file ]
then
	echo "ERROR: given GSPAN file cannot be found"
	exit
fi

if [ ! -d $output_dir ]
then
	echo -e "ERROR: output_dir does not exist\n\n"
	exit
fi

if [ $metric == "CLUSTER" ] && [ ! -n "$cluster_options" ]
then
	echo "ERROR: specify parameters to be passed to K quick shift clustering."
	exit
fi

#echo -e "column $class_column \n"


## extract basename
base_name=$( basename $gspan_file .gz )
base_name=$( basename $base_name .gspan )
output_path_base=$output_dir/$base_name
echo -e "base_name: $base_name\n\n"
#echo "output_path_base: $output_path_base"

##assign output and aux file names
target=$output_path_base.target
class=$output_path_base.class
dict2_file=$output_path_base.dict2
matrix_file=$output_path_base.matrix
knn_targetv_file=$output_path_base.knn_target_value 
all_results_file=$output_path_base.analysis_all

##OUTPUT REAULT HASH
declare -A results
##calculate auxilary files if not been extracted before
if [ ! -f $target ] || [ ! -f $dict2_file ]
then
	zcat -f $gspan_file | grep "^t" | awk -v column=$class_column '{print $column}'  | sed 's/^\([^.]*\).*/\1/' > $class
	cat $class | sort | uniq -c | cat -n  | awk '{print $1"\t"$3"\t"$2}' > $dict2_file
	cat $class | awk -v HF=$dict2_file 'BEGIN{while(getline <HF){H[$2]=$1}} {print H[$1]}' > $target
fi

# K quick shift clustering
cluster_folder=$output_path_base-clusters/
if [ $metric == "CLUSTER" ]
then
	if [ -d $cluster_folder ]
	then
		echo -e "Output cluster folder exists.\n"
	else
		## run EDeN clustering
		$EDeN --action CLUSTER --file_type GRAPH -i $gspan_file -y $cluster_folder $cluster_options
	fi

	# TODO analyze clustering results
	
	[ $metric == "ALL" ] || exit
fi

## KNN neareast neighbourhood analysis
if [ $metric == "KNN" ] || [ $metric == "ALL" ]
then
	if [ -f $knn_targetv_file ]
	then
		echo -e "Output knn_target_value files exist, reusing history...\n"
	else
		rm -f knn_target_value knn knn_feature knn_kernel_value
		## run EDeN for knn
		$EDeN -i $1 -t  $target -a NEAREST_NEIGHBOR -r 3 -d 5
		mv knn_target_value $knn_targetv_file
	fi

	## report number of elements that their nearest neighbor belongs to their classes
	## hit reation of knn s
	printf "Improved hit count\n";
	printf "Neigh.\t#hits\t#total\t#ratio(improved)\n" ; 
	for n in {1..9}; 
	do 
 	  cat  $knn_targetv_file \
 	  | awk -v nn=$n -v HF=$dict2_file 'BEGIN{while(getline < HF){H[$1]=$3}} {if(H[$2]>nn) {print $2,$(nn+2), H[$2]}}' \
 	  | awk -v n=$n -v count=0 '$1==$2{count++} END{print n"\t"count"\t"NR"\t"count/NR}' 
	  
	  knn_ratio=$( cat  $knn_targetv_file \
	  | awk -v nn=$n -v HF=$dict2_file 'BEGIN{while(getline < HF){H[$1]=$3}} {if(H[$2]>nn) {print $2,$(nn+2), H[$2]}}' \
	  | awk -v n=$n -v count=0 '$1==$2{count++} END{print count/NR}'
	  )
	  
	  results[$n"knn"]=$knn_ratio
	  
		
		 
	 done

	#$BIN_DIR/hit-ratio.sh $knn_targetv_file $dict2_file
	
	[ $metric == "ALL" ] || exit
fi

##perf measurment
if  [ $metric == "ROC" ] || [ $metric == "R50" ] || [ $metric == "APR" ] || [ $metric == "ALL" ]
then
	if [ -f $matrix_file ]
	then
		echo -e "Output matrix files exist, reusing history...\n"
	else
		rm -f matrix
		## run EDeN
		$EDeN --action MATRIX  --file_type GRAPH --cluster_type K_QUICK_SHIFT  -i $1 -r 3 -d 5 --shared_neighborhood 
		mv matrix $matrix_file
	fi
	
	if [ $metric == "ALL" ]
	then
	  for perfs in R50 APR ROC;
	  do
	    if [ ! -f $matrix_file.$perfs.txt ]
	    then
	      perl $BIN_DIR/perf_measurer.pl $target $matrix_file $perfs
	    fi
	    last_line=$(tail -1 $matrix_file.$perfs.txt)
	    if [[ $last_line =~ "result:".* ]]
	    then
	      results["$perfs"]=$(echo $last_line | sed 's/result://g')
	    else
	      echo -e "Warning $perfs result not found\n"
	    fi
	    
	  done

	 else
	  perl $BIN_DIR/perf_measurer.pl $target $matrix_file $metric
	  exit
	fi
fi

if [ $metric == "ALL" ]
then
 echo -e "Analysis result:\n"
 keys="${!results[@]}"
 sorted_keys=$(echo $keys | sed 's/ /\n/g' | sort -n)
 rm -f $all_results_file
 for key in $sorted_keys; do
    echo -e "$key\t${results["$key"]}"
    echo -e "$key\t${results["$key"]}" >> $all_results_file
 done
 exit
fi

echo "ERROR metric $metric is not suported"
exit 1
