__author__ = 'Alexander Junge, RTH, University of Copenhagen, Denmark <alexander.junge@gmail.com>'

import sys
import itertools
import os

from Bio import SeqIO

import inspect
current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parent_dir = os.path.dirname(current_dir)
tools_dir = os.path.join(parent_dir, 'tools')
sys.path.insert(0, tools_dir)

import pairwise_sequence_identity as pwseqid


def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist" % arg)
    return arg

import argparse
parser = argparse.ArgumentParser(description='''
                        Computes the average pairwise sequence identity for a set of sequences contained in a set of
                        files in FASTA or Stockholm format.

                        The sequences are read from file(s) given on the command line.
                        For each pair of input files A and B, all sequences in A are compared to all sequences in B and
                        and the average pairwise sequence identity of all comparisons is written in the following
                        tab-delimited format to stdout:
                        <A>\\t<B>\\t<average pairwise identity, a float>
                        ''')

parser.add_argument('filename', type=is_valid_path, nargs='+',
                    help='File name(s) that contain the sequences to be read. File names in this list must be unique.')
parser.add_argument('-f', '--seqformat', default='stockholm', choices=['stockholm', 'fasta'],
                    help='Format of the files storing the sequences to be read (default: \'stockholm\').')
parser.add_argument('-p', '--poolseq', dest='pool_sequences',
                    help='If -p is set, sequences contained in different input files are pooled before computing their '
                         'pairwise sequence identity. This means that only a single float is returned on stdout - '
                         'the average pairwise sequence identity of all input sequences.',
                    action='store_true')
parser.add_argument('-r', '--realign', dest='realign_sequences',
                    help='If -r is set, each pair of sequences is globally realigned before computing their sequence '
                         'identity. Otherwise, no alignment is computed and the input files are assumed to contain '
                         'already aligned sequences.',
                    action='store_true')
parser.add_argument('-a', '--aveseqlength', dest='average_sequence_length',
                    help='If -a is set, the average pairwise identity of two sequences is computed as '
                         '(number of identical positions)/(average length of the two sequences). '
                         'However, the default is '
                         '(number of identical positions)/(max ungapped sequence length of input sequences).',
                    action='store_true')
args = parser.parse_args()

input_file_names = args.filename
sequence_format = args.seqformat
pool_sequences = args.pool_sequences
realign_sequences = args.realign_sequences
normalize_by_average_sequence_length = args.average_sequence_length

if len(set(input_file_names)) != len(input_file_names):
    raise IOError('Names of files must be unique.')

if len(input_file_names) <= 1 and not pool_sequences:
    raise IOError('Only a single file was given. Set -p flag to pool the sequences and compute the average pairwise '
                  'sequence identity of the sequences in this file.')


def main():
    # maps each file name to the sequences read from the file (in a list)
    file_to_sequence_lists = {}
    for input_file in input_file_names:
        sequence_list = []
        with open(input_file, 'r') as input_file_handler:
            for record in SeqIO.parse(input_file_handler, sequence_format):
                # Convert all nucleotides to upper case
                sequence_list.append(str(record.seq).upper())
        file_to_sequence_lists[input_file] = sequence_list

    if pool_sequences:
        # following puts all aggregates all values in the codomain of file_to_sequence_lists in a list, believe me ;)
        sequence_list = list(itertools.chain(*(file_to_sequence_lists.values())))
        all_identities = []
        for pair in itertools.combinations(sequence_list, r=2):
            first_sequence_gaps, second_sequence_gaps = pair
            sequence_identity = pwseqid.get_sequence_identity(first_sequence_gaps, second_sequence_gaps,
                                                              realign_sequences, normalize_by_average_sequence_length)
            all_identities.append(sequence_identity)

        if len(all_identities) > 0:
            average_pairwise_sequence_identity = sum(all_identities)/len(all_identities)
        else:
            average_pairwise_sequence_identity = 'NA'
        sys.stdout.write(str(average_pairwise_sequence_identity) + '\n')
    else:
        for file_a, file_b in itertools.combinations(file_to_sequence_lists.keys(), r=2):
            sequences_a = file_to_sequence_lists[file_a]
            sequences_b = file_to_sequence_lists[file_b]

            all_identities = []
            for seq_a in sequences_a:
                for seq_b in sequences_b:
                    sequence_identity = pwseqid.get_sequence_identity(seq_a, seq_b, realign_sequences,
                                                                      normalize_by_average_sequence_length)
                    all_identities.append(sequence_identity)

            if len(all_identities) > 0:
                average_pairwise_sequence_identity = sum(all_identities)/len(all_identities)
            else:
                average_pairwise_sequence_identity = 'NA'
            sys.stdout.write(file_a + '\t' + file_b + '\t' + str(average_pairwise_sequence_identity) + '\n')


if __name__ == "__main__":
    main()