__author__ = 'Alexander Junge <alexander.junge@gmail.com>'

import sys
import itertools
import argparse
import re
import os

from Bio import SeqIO
from Bio import pairwise2


def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist" % arg)
    return arg


def is_supported_sequence_file_format(arg):
    if arg.lower() == 'stockholm':
        return 'stockholm'
    if arg.lower() == 'fasta':
        return 'fasta'
    else:
        parser.error("Given alignment file type is not supported." % arg)


parser = argparse.ArgumentParser()
argparse.ArgumentParser(description=
                        '''Computes the average sequence identity for a set of sequences present in a file in FASTA or
                        Stockholm format. The sequences are read from a list of file names given on the command line and
                        the average sequence identity is written to stdout.''')

parser.add_argument('filename', type=is_valid_path, nargs='+',
                    help='list of file names containing sequences to be read')
parser.add_argument('-f', '--seqformat', type=is_supported_sequence_file_format,
                    help='format of the sequences read from stdin - "Stockholm" or "FASTA" format')
parser.add_argument('-r', '--realign', dest='realign_sequences',
                    help='perform a realignment of each pair of sequences before computing their sequence identity',
                    action='store_true')
args = parser.parse_args()
seqformat = args.seqformat
input_file_names = args.filename


def compute_pairwise_sequence_identity(alignment_score, length_first_sequence, length_second_sequence):
    """

    :param alignment_score: int - The number of identical nucleotides in the sequences.
    :param length_first_sequence: int - The length of the first input sequence.
    :param length_second_sequence: int - The length of the second input sequence.
    :return: float - The computed sequence identity of the sequences.
    """

    return alignment_score / ((length_first_sequence + length_second_sequence) / 2.0)


def count_matching_characters(sequence_one, sequence_two):
    """

    :param sequence_one: string - the first sequence to be compared
    :param sequence_two: string - the second sequence to be compared
    :return: the number of matching characters between the two strings, gap character matches are ignored
    """
    if len(sequence_one) != len(sequence_two):
        raise ValueError('Given sequences have different length')
    identical_chars = 0
    for idx in range(len(sequence_one)):
        first_char = sequence_one[idx]
        sec_char = sequence_two[idx]
        if first_char == sec_char and first_char != '-':
            identical_chars += 1
    return identical_chars


sequence_list = []
for input_file in input_file_names:
    with open(input_file, 'r') as input_file_handler:
        for record in SeqIO.parse(input_file_handler, seqformat):
            # Convert all nucleotides to upper case
            sequence_list.append(str(record.seq).upper())

all_identities = []
for pair in itertools.combinations(sequence_list, r=2):
    first_sequence_gaps, second_sequence_gaps = pair
    first_sequence = re.sub('-', '', first_sequence_gaps)
    second_sequence = re.sub('-', '', second_sequence_gaps)
    # Pairwise alignment to determine sequence identity.
    if args.realign_sequences:
        # Perform a pairwise realignment of the sequences where identical characters have score 1,
        # otherwise 0 and no gap penalties are used.
        sequence_alignment = pairwise2.align.globalxx(first_sequence, second_sequence)[0]
        sequence_1, sequence_2, score, begin, end = sequence_alignment
    else:
        # Use the sequence alignment from Rfam
        score = count_matching_characters(first_sequence_gaps, second_sequence_gaps)
    sequence_identity = compute_pairwise_sequence_identity(score, len(first_sequence), len(second_sequence))
    all_identities.append(sequence_identity)

if len(all_identities) > 0:
    average_pairwise_sequence_identity = sum(all_identities)/len(all_identities)
else:
    average_pairwise_sequence_identity = 'NA'
sys.stdout.write(str(average_pairwise_sequence_identity) + '\n')