#!/usr/bin/env python
import matplotlib
matplotlib.use('Agg')
import argparse
import logging
import os
import gzip
import pandas as pd
import seaborn as sns
import itertools
import collections
from collections import Counter
import multiprocessing
import numpy as np
from Bio import SeqIO
from Bio import SeqUtils
import pairwise_sequence_identity as pwseqid
import matplotlib.pyplot as plt

logging.basicConfig(level=logging.INFO)

__author__ = 'Alexander Junge (alexander.junge@gmail.com)'


def is_valid_dir(arg):
    if not os.path.isdir(arg):
        parser.error('Given path %s is not a directory.' % arg)
    return arg


def is_not_existing_path(arg):
    if os.path.exists(arg):
        parser.error("Given path %s exists. Please specify a non-existing output path." % arg)
    return arg


def is_positive_int(arg):
    int_arg = None
    try:
        int_arg = int(arg)
    except ValueError:
        parser.error('Argument {} is not an integer.'.format(arg))
    if int_arg < 1:
        parser.error('Argument {} is not a positive integer.'.format(arg))
    return int_arg


def count_class_members(class_vec):
    """

    :param class_vec: list containing class label of each object in data set
    :return: a pandas DataFrame with two columns. First column is the class, second call is the number of class members.
    """
    classes = list(set(class_vec))
    classes.sort()
    class_members = []
    for cl in classes:
        c_mems = sum([cl == cc for cc in class_vec])
        class_members.append(c_mems)
    return pd.DataFrame({'Class': classes, 'Members': class_members})


def get_data_set_size(class_count_df):
    """

    :param class_count_df: pandas DataFrame with two columns. First column is class, second call number of class members
    :return: pandas DataFrame with two columns. First column is the minimal number of class members, second is
             the number of classes having at least this many members.
    """
    member_counts = class_count_df['Members'].tolist()
    object_counts = []
    class_counts = []
    min_member_counts = range(1, 11)
    for min_members in min_member_counts:
        object_count = sum([mem_c for mem_c in member_counts if mem_c >= min_members])
        class_count = sum([mem_c >= min_members for mem_c in member_counts])
        object_counts.append(object_count)
        class_counts.append(class_count)
    df = pd.DataFrame({'K': min_member_counts, 'Number of Rfam families': class_counts,
                       'Number of Alignments': object_counts})
    return df[['K', 'Number of Rfam families', 'Number of Alignments']]


def count_sequences(stockholm_files):
    """

    :param stockholm_files - a list of string - file paths of the input alignments in Stockholm format
    :return: a pandas DataFrame with two columns. First column ist the file name, second column the number of sequences
             in the alignment
    """
    input_file_seq_counts = []
    file_names = []
    for my_file in stockholm_files:
        file_names.append(os.path.basename(my_file))
        my_file_seq_count = 0
        with open(my_file, 'r') as input_file_handler:
            for _ in SeqIO.parse(input_file_handler, 'stockholm'):
                my_file_seq_count += 1
        input_file_seq_counts.append(my_file_seq_count)
    return pd.DataFrame({'Object': file_names, 'SequenceCount': input_file_seq_counts})


def __object_apsi__(stockholm_file):
    sequence_list = []
    with open(stockholm_file, 'r') as input_file_handler:
        try:
            for record in SeqIO.parse(input_file_handler, 'stockholm'):
                # Convert all nucleotides to upper case
                sequence_list.append(str(record.seq).upper())
        except ValueError:
            logging.warn('Error while reading file {}. Continuing.'.format(stockholm_file))
            return float('nan')

    apsi_list = []
    for pair in itertools.combinations(sequence_list, r=2):
        first_sequence_gaps, second_sequence_gaps = pair
        sequence_identity = pwseqid.get_sequence_identity(first_sequence_gaps, second_sequence_gaps,
                                                          False, False)
        apsi_list.append(sequence_identity)

    if len(apsi_list) > 0:
        curr_apsi = sum(apsi_list)/len(apsi_list)
    else:
        curr_apsi = float('nan')
    return curr_apsi


def compute_object_apsi(stockholm_files, process_pool):
    """
    :param stockholm_files - a list of string - file paths of the input alignments in Stockholm format
    :param process_pool - a multiprocessing.Pool for concurrent computations. If None, serial execution is performed.
    :return a pandas DataFrame with two columns. First column is the file name, second column the APSI of the alignment
            NOTE: as the returned APSI. is used to measure the covariation in the structural alignments
                  as obtained from Rfam, no sequence-based realignment is performed before computing PSIs
    """
    input_file_apsis = []
    file_names = []
    for my_file in stockholm_files:
        file_names.append(os.path.basename(my_file))
        if not process_pool:
            curr_apsi = __object_apsi__(my_file)
            # logging.warn('{} has ASPI {:f} and {:d} sequences.\n'.format(my_file, apsi, len(sequence_list)))
            input_file_apsis.append(curr_apsi)
    if process_pool:
        input_file_apsis = process_pool.map(__object_apsi__, stockholm_files)

    return pd.DataFrame({'Object': file_names, 'APSI': input_file_apsis})


def __class_apsi__(stockholm_files):
    class_sequence_list = []
    for my_file in stockholm_files:
        with open(my_file, 'r') as input_file_handler:
            try:
                for record in SeqIO.parse(input_file_handler, 'stockholm'):
                    # Convert all nucleotides to upper case
                    class_sequence_list.append(str(record.seq).upper())
            except ValueError:
                logging.warn('Error while reading file {}. Continuing.'.format(my_file))
                return float('nan')

    apsi_list = []
    for pair in itertools.combinations(class_sequence_list, r=2):
        first_sequence_gaps, second_sequence_gaps = pair
        sequence_identity = pwseqid.get_sequence_identity(first_sequence_gaps, second_sequence_gaps,
                                                          False, False)
        apsi_list.append(sequence_identity)

    if len(apsi_list) > 0:
        curr_apsi = sum(apsi_list)/len(apsi_list)
    else:
        curr_apsi = float('nan')
    return curr_apsi


def compute_class_apsis(class_to_stockholm_files, process_pool):
    """
    :param class_to_stockholm_files - a dict string -> lists of strings. Maps each class name to a list containing file
                            paths of the input alignments of the class in Stockholm format.
    :param process_pool - a multiprocessing.Pool for concurrent computations. If None, serial execution is performed.
    :return a pandas DataFrame with two columns. First column is the class name, second column the APSI of the alignment
            of this class.
            NOTE: as the returned APSI. is used to measure the covariation in the structural alignments
                  as obtained from Rfam, no sequence-based realignment is performed before computing PSIs
    """
    class_names = list(class_to_stockholm_files.keys())
    class_names.sort()
    class_apsis = []
    class_file_list = []
    for my_class in class_names:
        curr_class_files = class_to_stockholm_files[my_class]
        if not process_pool:
            curr_apsi = __class_apsi__(curr_class_files)
            # logging.warn('{} has ASPI {:f} and {:d} sequences.\n'.format(curr_class, apsi, len(sequence_list)))
            class_apsis.append(curr_apsi)
        else:
            class_file_list.append(curr_class_files)
    if process_pool:
        class_apsis = process_pool.map(__class_apsi__, class_file_list)

    return pd.DataFrame({'Class': class_names, 'APSI': class_apsis})


def __compute_pairwise_similarity__(first_stk_file, second_stk_file):
    sequence_lists = [None, None]
    file_list = (first_stk_file, second_stk_file)
    for i, stk_file in enumerate(file_list):
        sequence_list = []
        with open(stk_file, 'r') as input_file_handler:
            try:
                for record in SeqIO.parse(input_file_handler, 'stockholm'):
                    # Convert all nucleotides to upper case
                    sequence_list.append(str(record.seq).upper())
            except ValueError:
                logging.warn('Error while reading file {}. Continuing.'.format(stk_file))
                return float('nan')
        sequence_lists[i] = sequence_list

    apsi_list = []
    for pair in itertools.product(*sequence_lists):
        first_sequence_gaps, second_sequence_gaps = pair
        sequence_identity = pwseqid.get_sequence_identity(first_sequence_gaps, second_sequence_gaps,
                                                          True, False)
        apsi_list.append(sequence_identity)

    if len(apsi_list) > 0:
        curr_apsi = sum(apsi_list)/len(apsi_list)
    else:
        curr_apsi = float('nan')
    return curr_apsi


def __compute_pairwise_similarity_explode__(stockholm_pair):
    if len(stockholm_pair) != 2:
        raise ValueError('Please give exactly two ')
    return __compute_pairwise_similarity__(*stockholm_pair)


def compute_within_class_apsis(class_to_stockholm_files, process_pool):
    """
    :param class_to_stockholm_files - a dict string -> lists of strings. Maps each class name to a list containing file
                            paths of the input alignments of the class in Stockholm format.
    :param process_pool - a multiprocessing.Pool for concurrent computations. If None, serial execution is performed.
    :return a pandas DataFrame with three columns. First column is the name of the first stockholm file,
            second column the name of the second stockholm file, third is the APSI of all pairs of sequences coming from
            the first and second file.  First and second file always belong to the same class.
            NOTE: as the returned APSI is used to measure the similarity between structural alignments
                  a sequence-based realignment is performed before computing PSIs
    """
    class_names = list(class_to_stockholm_files.keys())
    class_names.sort()
    within_class_apsis = []
    within_class_file_pairs = []
    for my_class in class_names:
        curr_class_files = class_to_stockholm_files[my_class]
        for file_pair in itertools.combinations(curr_class_files, r=2):
            within_class_file_pairs.append(file_pair)
            if not process_pool:
                curr_apsi = __compute_pairwise_similarity_explode__(file_pair)
                # logging.warn('{} has ASPI {:f} and {:d} sequences.\n'.format(curr_class, apsi, len(sequence_list)))
                within_class_apsis.append(curr_apsi)
    if process_pool:
        within_class_apsis = process_pool.map(__compute_pairwise_similarity_explode__, within_class_file_pairs)

    return pd.DataFrame({'Object1': [x[0] for x in within_class_file_pairs],
                         'Object2': [x[1] for x in within_class_file_pairs],
                         'APSI': within_class_apsis})


def compute_between_class_apsis(class_to_stockholm_files, process_pool):
    """
    :param class_to_stockholm_files - a dict string -> lists of strings. Maps each class name to a list containing file
                            paths of the input alignments of the class in Stockholm format.
    :param process_pool - a multiprocessing.Pool for concurrent computations. If None, serial execution is performed.
    :return a pandas DataFrame with three columns. First column is the name of the first stockholm file,
            second column the name of the second stockholm file, third is the APSI of all pairs of sequences coming from
            the first and second file. First and second file always belong to different classes.
            NOTE: as the returned APSI is used to measure the similarity between structural alignments
                  a sequence-based realignment is performed before computing PSIs
    """
    class_names = list(class_to_stockholm_files.keys())
    class_names.sort()
    between_class_apsis = []
    between_class_file_pairs = []
    for class_pair in itertools.combinations(class_names, r=2):
        first_class, second_class = class_pair
        first_class_files = class_to_stockholm_files[first_class]
        second_class_files = class_to_stockholm_files[second_class]
        for file_pair in itertools.product(first_class_files, second_class_files):
            between_class_file_pairs.append(file_pair)
            if not process_pool:
                curr_apsi = __compute_pairwise_similarity_explode__(file_pair)
                # logging.warn('{} has ASPI {:f} and {:d} sequences.\n'.format(curr_class, apsi, len(sequence_list)))
                between_class_apsis.append(curr_apsi)
    if process_pool:
        between_class_apsis = process_pool.map(__compute_pairwise_similarity_explode__, between_class_file_pairs)

    return pd.DataFrame({'Object1': [x[0] for x in between_class_file_pairs],
                         'Object2': [x[1] for x in between_class_file_pairs],
                         'APSI': between_class_apsis})


def __object_asl__(stockholm_file):
    sequence_lengths = []
    with open(stockholm_file, 'r') as input_file_handler:
        try:
            for record in SeqIO.parse(input_file_handler, 'stockholm'):
                sequence_lengths.append(len(str(record.seq.ungap('-'))))
        except ValueError:
            logging.warn('Error while reading file {}. Continuing.'.format(stockholm_file))
            return float('nan')
    if len(sequence_lengths) > 0:
        asl = sum(sequence_lengths)/float(len(sequence_lengths))
    else:
        asl = float('nan')
    return asl


def compute_object_asl(stockholm_files, process_pool):
    """
    :param stockholm_files - a list of strings - file paths of the input alignments in Stockholm format
    :param process_pool - a multiprocessing.Pool for concurrent computations. If None, serial execution is performed.
    :return a pandas DataFrame with two columns. First column is the file name, second column the ASL of the alignment
    """
    input_file_asls = []
    file_names = []
    for my_file in stockholm_files:
        file_names.append(os.path.basename(my_file))
        if not process_pool:
            asl = __object_asl__(my_file)
            input_file_asls.append(asl)

    if process_pool:
        input_file_asls = process_pool.map(__object_asl__, stockholm_files)

    return pd.DataFrame({'Object': file_names, 'ASL': input_file_asls})


def __class_asl__(stockholm_files):
    sequence_lengths = []

    for stockholm_file in stockholm_files:
        with open(stockholm_file, 'r') as input_file_handler:
            try:
                for record in SeqIO.parse(input_file_handler, 'stockholm'):
                    sequence_lengths.append(len(str(record.seq.ungap('-'))))
            except ValueError:
                logging.warn('Error while reading file {}. Continuing.'.format(stockholm_file))
                return float('nan')

    if len(sequence_lengths) > 0:
        asl = sum(sequence_lengths)/float(len(sequence_lengths))
    else:
        asl = float('nan')
    return asl


def compute_class_asls(class_to_stockholm_files, process_pool):
    """
    :param class_to_stockholm_files - a dict string -> lists of strings. Maps each class name to a list containing file
                            paths of the input alignments of the class in Stockholm format.
    :param process_pool - a multiprocessing.Pool for concurrent computations. If None, serial execution is performed.
    :return a pandas DataFrame with two columns. First column is the file name, second column the ASL of the class
           alignments
    """
    class_names = list(class_to_stockholm_files.keys())
    class_names.sort()
    class_asls = []
    class_file_list = []
    for my_class in class_names:
        curr_class_files = class_to_stockholm_files[my_class]
        if not process_pool:
            asl = __class_asl__(curr_class_files)
            # logging.warn('{} has ASL {:f} and {:d} sequences.\n'.format(curr_class, apsi, len(sequence_list)))
            class_asls.append(asl)
        else:
            class_file_list.append(curr_class_files)
    if process_pool:
        class_asls = process_pool.map(__class_asl__, class_file_list)

    return pd.DataFrame({'Class': class_names, 'ASL': class_asls})


def __object_gcc__(stockholm_file):
    sequence_gccs = []
    with open(stockholm_file, 'r') as input_file_handler:
        try:
            for record in SeqIO.parse(input_file_handler, 'stockholm'):
                sequence_gccs.append(SeqUtils.GC(str(record.seq.ungap('-'))))
        except ValueError:
            logging.warn('Error while reading file {}. Continuing.'.format(stockholm_file))
            return float('nan')
    if len(sequence_gccs) > 0:
        gcc = sum(sequence_gccs)/float(len(sequence_gccs))
    else:
        gcc = float('nan')
    return gcc


def compute_object_gcc(stockholm_files, process_pool):
    """
    :param stockholm_files - a list of strings - file paths of the input alignments in Stockholm format
    :param process_pool - a multiprocessing.Pool for concurrent computations. If None, serial execution is performed.
    :return a pandas DataFrame with two columns. First column is the file name, second column the GC content of the
            alignment
    """
    input_file_gccs = []
    file_names = []
    for my_file in stockholm_files:
        file_names.append(os.path.basename(my_file))
        if not process_pool:
            gcc = __object_gcc__(my_file)
            input_file_gccs.append(gcc)

    if process_pool:
        input_file_gccs = process_pool.map(__object_gcc__, stockholm_files)

    return pd.DataFrame({'Object': file_names, 'G+C_Content': input_file_gccs})


def __class_gcc__(stockholm_files):
    sequence_gccs = []

    for stockholm_file in stockholm_files:
        with open(stockholm_file, 'r') as input_file_handler:
            try:
                for record in SeqIO.parse(input_file_handler, 'stockholm'):
                    sequence_gccs.append(SeqUtils.GC(str(record.seq.ungap('-'))))
            except ValueError:
                logging.warn('Error while reading file {}. Continuing.'.format(stockholm_file))
                return float('nan')

    if len(sequence_gccs) > 0:
        gcc = sum(sequence_gccs)/float(len(sequence_gccs))
    else:
        gcc = float('nan')
    return gcc


def compute_class_gcc(class_to_stockholm_files, process_pool):
    """
    :param class_to_stockholm_files - a dict string -> lists of strings. Maps each class name to a list containing file
                            paths of the input alignments of the class in Stockholm format.
    :param process_pool - a multiprocessing.Pool for concurrent computations. If None, serial execution is performed.
    :return a pandas DataFrame with two columns. First column is the file name, second column the GC content of the
            class alignments
    """
    class_names = list(class_to_stockholm_files.keys())
    class_names.sort()
    class_gccs = []
    class_file_list = []
    for my_class in class_names:
        curr_class_files = class_to_stockholm_files[my_class]
        if not process_pool:
            gcc = __class_gcc__(curr_class_files)
            # logging.warn('{} has GC content {:f} and {:d} sequences.\n'.format(curr_class, apsi, len(sequence_list)))
            class_gccs.append(gcc)
        else:
            class_file_list.append(curr_class_files)
    if process_pool:
        class_gccs = process_pool.map(__class_gcc__, class_file_list)

    return pd.DataFrame({'Class': class_names, 'G+C_Content': class_gccs})


def plot_histogram(values, bin_number, plot_range, show_mean, title, xlabel, ylabel, output_file, y_log_scale=False):
    plt.figure()
    if plot_range:
        plt.hist(values, bins=bin_number, range=plot_range)
    else:
        plt.hist(values, bins=bin_number)
    if show_mean:
        mean_val = float(sum(values))/len(values)
        logging.info('Mean is: {:f}'.format(mean_val))
        plt.axvline(mean_val, color='black', linestyle='dashed', linewidth=2)
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if y_log_scale:
        plt.yscale('log', nonposy='clip')
    plt.savefig(output_file)
    plt.close()


def plot_overlay_histogram(values1, values2, bin_number, plot_range, show_mean, title, xlabel, ylabel, label1, label2,
                           fpath, y_log_scale, plot_kde):
    plt.figure()
    if plot_range:
        if plot_kde:
            sns.kdeplot(np.array(values1), label=label1)
            sns.kdeplot(np.array(values2), label=label2)
        else:
            plt.hist(values1, bins=bin_number, range=plot_range, alpha=0.5, label=label1)
            plt.hist(values2, bins=bin_number, range=plot_range, alpha=0.5, label=label2)
    else:
        if plot_kde:
            sns.kdeplot(np.array(values1), label=label1)
            sns.kdeplot(np.array(values2), label=label2)
        else:
            plt.hist(values1, bins=bin_number, alpha=0.5, label=label1)
            plt.hist(values2, bins=bin_number, alpha=0.5, label=label2)
    plt.legend(loc='upper right')

    if show_mean:
        mean_val = (float(sum(values1)) + float(sum(values2)))/(len(values1) + len(values2))
        logging.info('Mean is: {:f}'.format(mean_val))
        plt.axvline(mean_val,
                    color='black', linestyle='dashed', linewidth=2)

    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if y_log_scale:
        plt.yscale('log', nonposy='clip')
    plt.savefig(fpath)
    plt.close()


if __name__ == '__main__':
    sns.set_palette("deep", desat=.6)

    parser = argparse.ArgumentParser(description="""
                                                 Computes overview statistics for a set of structural alignments.
                                                 """)
    parser.add_argument('input_dir', type=is_valid_dir,
                        help="""
                             A directory containing the set of structural alignment files.

                             All output statistics are computed for these alignments. Alignments are assumed to be in
                             Stockholm format.
                             """)
    parser.add_argument('output_dir',
                        help="""
                             Output files and plots are written to this directory.

                             If the directory does not exist it will be created at runtime.
                             """)
    parser.add_argument('thread_count', type=is_positive_int,
                        help="""
                             The number of threads to be started concurrently. Must be a positive integer.
                             """)
    parser.add_argument('-n', '--no_within_between_class_apsi', action='store_true',
                        help="""
                             If set, within and between class Mean PSIs will not be computed.
                             """
                        )
    parser.add_argument('-m', '--min_family_size',
                        type=is_positive_int, default=1,
                        help='The minimal number of alignments per family; families with fewer alignments are removed.')
    args = parser.parse_args()

    input_dir = args.input_dir
    output_dir = args.output_dir
    thread_count = args.thread_count
    no_within_between_class_apsi = args.no_within_between_class_apsi
    min_family_size = args.min_family_size

    if thread_count > 1:
        pool = multiprocessing.Pool(thread_count)
    else:
        pool = None

    if not os.path.isdir(output_dir):
        try:
            os.makedirs(output_dir, mode=0o774)
        except OSError:
            raise OSError('Failed to create output directory {}.'.format(output_dir))

    input_file_paths = [os.path.join(input_dir, p)
                        for p in sorted(os.listdir(input_dir)) if os.path.isfile(os.path.join(input_dir, p))]
    logging.info('Found {:d} files in input directory'.format(len(input_file_paths)))

    non_stk_files = [not os.path.basename(f).endswith('.stk') for f in input_file_paths]
    if sum(non_stk_files) > 0:
        logging.warn('Detected {} files in input directory that do not end on .stk. Removing files'.format(
                sum(non_stk_files)))
        input_file_paths = [input_file_paths[idx] for idx, is_no_stk_file in enumerate(non_stk_files)
                            if not is_no_stk_file]
        logging.warn('{:d} .stk files remaining.'.format(len(input_file_paths)))

    if not all([len(os.path.basename(f).split('.')) == 3 for f in input_file_paths]):
        raise ValueError('Did not detect class label for all input files. Expected naming is: <class>.<integer>.stk')

    axes = {'titlesize': 'xx-large',
            'labelsize': 'x-large'}
    plt.rc('axes', **axes)

    input_file_classes = [os.path.basename(f).split('.')[0] for f in input_file_paths]
    logging.info('Detected class assignment for {:d} classes.'.format(len(set(input_file_classes))))

    logging.info('Removing classes with fewer than {} member(s).'.format(min_family_size))
    class_counter = Counter(input_file_classes)
    remove_classes = set()
    for cls, occ in class_counter.iteritems():
        if occ < min_family_size:
            remove_classes.add(cls)
    remove_indices = set()
    for idx, cls in enumerate(input_file_classes):
        if cls in remove_classes:
            remove_indices.add(idx)
    tmp_input_file_paths = []
    tmp_input_file_classes = []
    for idx, path_cls in enumerate(zip(input_file_paths, input_file_classes)):
        path, cls = path_cls
        if idx not in remove_indices:
            tmp_input_file_paths.append(path)
            tmp_input_file_classes.append(cls)
    input_file_paths = tmp_input_file_paths
    input_file_classes = tmp_input_file_classes
    logging.info('{:d} classes and {:d} objects remaining after filtering.'.format(len(set(input_file_classes)),
                                                                                   len(input_file_paths)))

    logging.info('Counting number of objects per Rfam family.')
    class_counts_tsv = os.path.join(output_dir, 'class_counts.tsv.gz')
    class_counts_hist = os.path.join(output_dir, 'class_counts.pdf')
    if os.path.exists(class_counts_tsv):
        logging.info('Skipping computations because previous class counts were detected.')
    else:
        class_counts_df = count_class_members(input_file_classes)
        with gzip.open(class_counts_tsv, 'w') as class_count_file:
            class_counts_df.to_csv(class_count_file, sep='\t', index=False)
        if len(class_counts_df) > 1:
            class_counts_list = class_counts_df['Members'].tolist()
            class_counts_bins = max(class_counts_list)
            plot_histogram(class_counts_list, class_counts_bins, None, False, '',
                           'Number of alignments per Rfam family',
                           'Count (log)', class_counts_hist, y_log_scale=True)
        else:
            logging.warn('Omit class_counts.pdf histogram as only {:d} classes found.'.format(len(class_counts_df)))

        data_set_size_df = get_data_set_size(class_counts_df)
        data_set_size_tsv = os.path.join(output_dir, 'data_set_size.tsv.gz')
        with gzip.open(data_set_size_tsv, 'w') as data_set_size_file:
            data_set_size_df.to_csv(data_set_size_file, sep='\t', index=False)

        data_set_size_tex = os.path.join(output_dir, 'data_set_size.tex')
        latex_table = data_set_size_df.to_latex(index=False)
        with open(data_set_size_tex, 'w') as tex_file:
            tex_file.write(latex_table)

    logging.info('Counting number of sequences per alignment.')
    seq_counts_tsv = os.path.join(output_dir, 'seq_counts.tsv.gz')
    seq_counts_hist = os.path.join(output_dir, 'seq_counts.pdf')
    if os.path.exists(seq_counts_tsv):
        logging.info('Skipping computations because previous sequence counts were detected.')
    else:
        seq_counts_df = count_sequences(input_file_paths)
        with gzip.open(seq_counts_tsv, 'w') as seq_counts_file:
            seq_counts_df.to_csv(seq_counts_file, sep='\t', index=False)
        counts = seq_counts_df['SequenceCount'].tolist()
        seq_count_bins = len(set(counts))
        plot_histogram(counts, seq_count_bins, None, False, '', 'Number of sequences per alignment',
                       'Count', seq_counts_hist, y_log_scale=False)

    logging.info('Computing object-wise Mean PSIs.')
    object_apsi_tsv = os.path.join(output_dir, 'object_apsi.tsv.gz')
    object_apsi_hist = os.path.join(output_dir, 'object_apsi.pdf')
    if os.path.exists(object_apsi_tsv):
        logging.info('Skipping computations because previous object Mean PSIs were detected.')
    else:
        object_apsi_df = compute_object_apsi(input_file_paths, pool)
        with gzip.open(object_apsi_tsv, 'w') as obj_apsi_file:
            object_apsi_df.to_csv(obj_apsi_file, sep='\t', index=False)
        plot_histogram(object_apsi_df['APSI'].tolist(), 50, [0, 1], True, '',
                       'Mean PSI per alignment (based on Rfam alignments)',
                       'Count',
                       object_apsi_hist)

    # collect Stockholm files belonging to same class
    class_to_input_file_paths = collections.defaultdict(list)
    for curr_class in set(input_file_classes):
        curr_class_paths = [file_path for idx, file_path in enumerate(input_file_paths)
                            if input_file_classes[idx] == curr_class]
        class_to_input_file_paths[curr_class] = curr_class_paths

    logging.info('Computing class-wise Mean PSIs.')
    class_apsi_tsv = os.path.join(output_dir, 'class_apsi.tsv.gz')
    class_apsi_hist = os.path.join(output_dir, 'class_apsi.pdf')
    if os.path.exists(class_apsi_tsv):
        logging.info('Skipping computations because previous class Mean PSIs were detected.')
    else:
        class_apsi_df = compute_class_apsis(class_to_input_file_paths, pool)
        with gzip.open(class_apsi_tsv, 'w') as class_apsi_file:
            class_apsi_df.to_csv(class_apsi_file, sep='\t', index=False)
        if len(class_apsi_df) > 1:
            plot_histogram(class_apsi_df['APSI'].tolist(), 50, [0, 1], True, '',
                           'Mean PSI per Rfam family (based on Rfam alignments)',
                           'Count',
                           class_apsi_hist)
        else:
            logging.warn('Omit class_apsi.pdf histogram as only {:d} classes found.'.format(len(class_apsi_df)))

    if no_within_between_class_apsi:
        logging.info('Skipping computations of between and within class similarities.')
    else:
        within_class_df = []
        within_class_mean_df = []
        logging.info('Computing within-class Mean PSI.')
        within_class_tsv = os.path.join(output_dir, 'within_class_APSI.tsv.gz')
        within_class_hist = os.path.join(output_dir, 'within_class_APSI.pdf')
        if os.path.exists(within_class_tsv):
            logging.info('Skipping computations because previous within cluster Mean PSIs were detected.')
        else:
            within_class_df = compute_within_class_apsis(class_to_input_file_paths, pool)
            with gzip.open(within_class_tsv, 'w') as within_c_file:
                within_class_df.to_csv(within_c_file, sep='\t', index=False)
            if len(within_class_df) > 1:
                plot_histogram(within_class_df['APSI'].tolist(), 50, [0, 1], True, '',
                               'Mean PSI of objects belonging to same class', 'Count', within_class_hist,
                               y_log_scale=False)
            else:
                logging.warn('Omit {} histogram as only {:d} within class pairs found.'.format(within_class_hist,
                                                                                               len(within_class_df)))

            obj1_names = within_class_df['Object1'].tolist()
            obj2_names = within_class_df['Object2'].tolist()
            apsi_values = within_class_df['APSI'].tolist()
            class_to_apsis = collections.defaultdict(list)
            for o1, o2, apsi in zip(obj1_names, obj2_names, apsi_values):
                o1_fname = os.path.basename(o1)
                o2_fname = os.path.basename(o2)
                c1 = o1_fname.split('.')[0]
                c2 = o2_fname.split('.')[0]
                if c1 != c2:
                    raise ValueError('Objects {} and {} do not belong to the same class'.format(o1_fname, o2_fname))
                key = c1
                class_to_apsis[key].append(apsi)

            class_1 = []
            apsis = []
            for c, vals in class_to_apsis.iteritems():
                class_1.append(c)
                apsis.append(sum(vals)/len(vals))
            within_class_mean_df = pd.DataFrame({'Class1': class_1, 'Class2': class_1, 'APSI': apsis})
            within_class_mean_hist = os.path.join(output_dir, 'within_class_mean_APSI.pdf')
            within_class_mean_tsv = os.path.join(output_dir, 'within_class_mean_APSI.tsv.gz')
            with gzip.open(within_class_mean_tsv, 'w') as within_c_file:
                within_class_mean_df.to_csv(within_c_file, sep='\t', index=False)
            plot_histogram(within_class_mean_df['APSI'].tolist(), 50, [0, 1], True, '',
                           'Mean PSI within classes', 'Count', within_class_mean_hist,
                           y_log_scale=False)

        between_class_df = []
        between_class_mean_df = []
        logging.info('Computing between-class Mean PSI.')
        between_class_tsv = os.path.join(output_dir, 'between_class_APSI.tsv.gz')
        between_class_hist = os.path.join(output_dir, 'between_class_APSI.pdf')
        if os.path.exists(between_class_tsv):
            logging.info('Skipping computations because previous between cluster Mean PSIs were detected.')
        else:
            between_class_df = compute_between_class_apsis(class_to_input_file_paths, pool)
            with gzip.open(between_class_tsv, 'w') as between_c_file:
                between_class_df.to_csv(between_c_file, sep='\t', index=False)
            if len(between_class_df) > 1:
                plot_histogram(between_class_df['APSI'].tolist(), 50, [0, 1], True, '',
                               'PSI of objects belonging to different class', 'Count', between_class_hist,
                               y_log_scale=False)
            else:
                logging.warn('Omit {} histogram as only {:d} between class pairs found.'.format(
                             between_class_hist, len(between_class_df)))

            obj1_names = between_class_df['Object1'].tolist()
            obj2_names = between_class_df['Object2'].tolist()
            apsi_values = between_class_df['APSI'].tolist()
            class_pairs_to_apsis = collections.defaultdict(list)
            for o1, o2, apsi in zip(obj1_names, obj2_names, apsi_values):
                o1_fname = os.path.basename(o1)
                o2_fname = os.path.basename(o2)
                c1 = o1_fname.split('.')[0]
                c2 = o2_fname.split('.')[0]
                key = tuple(sorted((c1, c2)))
                class_pairs_to_apsis[key].append(apsi)

            class_1 = []
            class_2 = []
            apsis = []
            for c_pair, vals in class_pairs_to_apsis.iteritems():
                class_1.append(c_pair[0])
                class_2.append(c_pair[1])
                apsis.append(sum(vals)/len(vals))
            between_class_mean_df = pd.DataFrame({'Class1': class_1, 'Class2': class_2, 'APSI': apsis})
            between_class_mean_hist = os.path.join(output_dir, 'between_class_mean_APSI.pdf')
            between_class_mean_tsv = os.path.join(output_dir, 'between_class_mean_APSI.tsv.gz')
            with gzip.open(between_class_mean_tsv, 'w') as between_c_file:
                between_class_mean_df.to_csv(between_c_file, sep='\t', index=False)
            plot_histogram(between_class_mean_df['APSI'].tolist(), 50, [0, 1], True, '',
                           'Mean PSI between class', 'Count', between_class_mean_hist,
                           y_log_scale=False)

        within_between_class_hist = os.path.join(output_dir, 'within_between_class_APSI.pdf')
        within_between_class_kde = os.path.join(output_dir, 'within_between_class_APSI_kde.pdf')
        if os.path.exists(within_between_class_hist):
            logging.info('Skipping within- and between-class Mean PSI histogram as {} exists.'.format(
                within_between_class_hist))
        else:
            logging.info('Plotting within- and between-class Mean PSIs in one plot.')
            if len(within_class_df) <= 1 or len(between_class_df) <= 1:
                logging.warn('Omit {} histogram.'.format(within_between_class_hist))
            else:
                plot_overlay_histogram(within_class_df['APSI'].tolist(),
                                       between_class_df['APSI'].tolist(),
                                       100, [0, 1], False, '',
                                       'Mean PSI of objects belonging to same and different classes',
                                       'Count',
                                       label1='Within class',
                                       label2='Between class',
                                       fpath=within_between_class_hist,
                                       y_log_scale=False,
                                       plot_kde=False)
                plot_overlay_histogram(within_class_df['APSI'].tolist(),
                                       between_class_df['APSI'].tolist(),
                                       100, [0, 1], False, '',
                                       'Mean PSI of objects belonging to same and different classes',
                                       'Density',
                                       label1='Within class',
                                       label2='Between class',
                                       fpath=within_between_class_kde,
                                       y_log_scale=False,
                                       plot_kde=True)

        within_between_mean_class_hist = os.path.join(output_dir, 'within_between_class_mean_APSI.pdf')
        within_between_mean_class_kde = os.path.join(output_dir, 'within_between_class_mean_APSI_kde.pdf')
        if os.path.exists(within_between_mean_class_hist):
            logging.info('Skipping within- and between-class Mean PSI histogram as {} exists.'.format(
                within_between_mean_class_hist))
        else:
            logging.info('Plotting within- and between-class Mean PSIs in one plot.')
            if len(within_class_mean_df) <= 1 or len(between_class_mean_df) <= 1:
                logging.warn('Omit {} histogram.'.format(within_between_mean_class_hist))
            else:
                plot_overlay_histogram(within_class_mean_df['APSI'].tolist(),
                                       between_class_mean_df['APSI'].tolist(),
                                       100, [0, 1], False, '',
                                       'Mean PSI between and within classes',
                                       'Count',
                                       label1='Within class',
                                       label2='Between class',
                                       fpath=within_between_mean_class_hist,
                                       y_log_scale=False,
                                       plot_kde=False)
                plot_overlay_histogram(within_class_mean_df['APSI'].tolist(),
                                       between_class_mean_df['APSI'].tolist(),
                                       100, [0, 1], False, '',
                                       'Mean PSI within and between classes',
                                       'Density',
                                       label1='Within class',
                                       label2='Between class',
                                       fpath=within_between_mean_class_kde,
                                       y_log_scale=False,
                                       plot_kde=True)

    logging.info('Computing object-wise ASLs.')
    object_asl_tsv = os.path.join(output_dir, 'object_asl.tsv.gz')
    object_asl_hist = os.path.join(output_dir, 'object_asl.pdf')
    if os.path.exists(object_asl_tsv):
        logging.info('Skipping computations because previous object ASLs were detected.')
    else:
        object_asl_df = compute_object_asl(input_file_paths, pool)
        with gzip.open(object_asl_tsv, 'w') as obj_asl_file:
            object_asl_df.to_csv(obj_asl_file, sep='\t', index=False)
        plot_histogram(object_asl_df['ASL'].tolist(), 50, None, False, '', 'Mean sequence length per alignment',
                       'Count (log)',
                       object_asl_hist, y_log_scale=True)

    logging.info('Computing class-wise ASLs.')
    class_asl_tsv = os.path.join(output_dir, 'class_asl.tsv.gz')
    class_asl_hist = os.path.join(output_dir, 'class_asl.pdf')
    if os.path.exists(class_asl_tsv):
        logging.info('Skipping computations because previous class ASLs were detected.')
    else:
        class_asl_df = compute_class_asls(class_to_input_file_paths, pool)
        with gzip.open(class_asl_tsv, 'w') as class_asl_file:
            class_asl_df.to_csv(class_asl_file, sep='\t', index=False)
        if len(class_asl_df) > 1:
            plot_histogram(class_asl_df['ASL'].tolist(), 50, None, False, '', 'Mean sequence length per Rfam family',
                           'Count (log)',
                           class_asl_hist,
                           y_log_scale=True)
        else:
            logging.warn('Omit class_asl.pdf histogram as only {:d} classes found.'.format(len(class_asl_df)))

    logging.info('Computing object-wise GC content.')
    object_gcc_tsv = os.path.join(output_dir, 'object_gcc.tsv.gz')
    object_gcc_hist = os.path.join(output_dir, 'object_gcc.pdf')
    if os.path.exists(object_gcc_tsv):
        logging.info('Skipping computations because previous object GC contents were detected.')
    else:
        object_gcc_df = compute_object_gcc(input_file_paths, pool)
        with gzip.open(object_gcc_tsv, 'w') as obj_gcc_file:
            object_gcc_df.to_csv(obj_gcc_file, sep='\t', index=False)
        plot_histogram(object_gcc_df['G+C_Content'].tolist(), 50, None, True, '', 'Mean GC content per alignment',
                       'Count',
                       object_gcc_hist)

    logging.info('Computing class-wise GC content.')
    class_gcc_tsv = os.path.join(output_dir, 'class_gcc.tsv.gz')
    class_gcc_hist = os.path.join(output_dir, 'class_gcc.pdf')
    if os.path.exists(class_gcc_tsv):
        logging.info('Skipping computations because previous class GC contents were detected.')
    else:
        class_gcc_df = compute_class_gcc(class_to_input_file_paths, pool)
        with gzip.open(class_gcc_tsv, 'w') as class_gcc_file:
            class_gcc_df.to_csv(class_gcc_file, sep='\t', index=False)
        if len(class_gcc_df) > 1:
            plot_histogram(class_gcc_df['G+C_Content'].tolist(), 50, None, True, '', 'Mean GC content per Rfam family',
                           'Count',
                           class_gcc_hist)
        else:
            logging.warn('Omit class_gcc.pdf histogram as only {:d} classes found.'.format(len(class_gcc_hist)))
