#!/usr/bin/env python
# Module providing functions for computing the sequence identity of a pair of sequences.
__author__ = 'Alexander Junge (alexander.junge@gmail.com)'

import re

#from skbio.alignment import global_pairwise_align_nucleotide as pw_align_scikit
import nwalign

import os
import inspect
current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


def __pairwise_sequence_identity_average_sequence_length__(alignment_score, length_first_sequence,
                                                           length_second_sequence):
    """
    :param alignment_score: int - The number of identical nucleotides in the sequences.
    :param length_first_sequence: int - The length of the first input sequence.
    :param length_second_sequence: int - The length of the second input sequence.
    :return: float - The computed sequence identity of the sequences measured as the alignment score divided by the
                     average sequence length.
    """
    return alignment_score/((length_first_sequence + length_second_sequence) / 2.0)


def __pairwise_sequence_identity_max_sequence_length__(alignment_score, max_sequence_length):
    """
    :param alignment_score: int - The number of identical nucleotides in the alignment.
    :param max_sequence_length: int - The ungapped length of the longer sequence in the alignment.
    :return: float - The computed sequence identity of the sequences measured as the alignment score divided by the
                     ungapped length of the longer sequence in the alignment.
    """
    return alignment_score/float(max_sequence_length)


def pairwise_sequence_identity(alignment_score, norm_by_average_sequence_length, length_first_sequence,
                               length_second_sequence):
    """
    Computes pairwise sequence identity of two aligned sequences.

    :param alignment_score: int - The number of identical nucleotides in the alignment.
    :param norm_by_average_sequence_length: bool - If True, the pairwise sequence identity is computed as
                                                (number of identical positions)/
                                                (average length of the ungapped two sequences).
                                                If False, the pairwise sequence identity is computed as
                                                (number of identical positions)/
                                                (length of longer ungapped sequence).
    :param length_first_sequence: int - The length of the first input sequence.
    :param length_second_sequence: int - The length of the second input sequence.
    :return: float - the computed sequence identity of the sequences.
    """
    if norm_by_average_sequence_length:
        return __pairwise_sequence_identity_average_sequence_length__(alignment_score, length_first_sequence,
                                                                      length_second_sequence)
    else:
        return __pairwise_sequence_identity_max_sequence_length__(alignment_score,
                                                                  max(length_first_sequence, length_second_sequence))


def __count_matching_characters__(sequence_one, sequence_two):
    """
    :param sequence_one: string - the first sequence to be compared
    :param sequence_two: string - the second sequence to be compared
    :return: the number of matching characters between the two strings, gap character matches are ignored
    """
    if len(sequence_one) != len(sequence_two):
        raise ValueError('Given sequences have different length although they are assumed to be aligned. '
                         'Sequences are:\n%s\n%s\n' % (sequence_one, sequence_two))
    identical_chars = 0
    for idx in range(len(sequence_one)):
        first_char = sequence_one[idx]
        sec_char = sequence_two[idx]
        if first_char == sec_char and first_char != '-':
            identical_chars += 1
    return identical_chars


def get_sequence_identity(first_sequence_gapped, second_sequence_gapped, perform_sequence_alignment,
                          norm_by_average_sequence_length):
    """
    Computes the sequence identity of the given pair of sequences.

    :param first_sequence_gapped: string - the first sequence to be compared
    :param second_sequence_gapped: string - the second sequence to be compared
    :param perform_sequence_alignment: bool - True if the input sequences are to be aligned pefore computing their
                                     sequence identity.
                                     False if input sequences are aligned and sequence identity should be
                                     computed from input alignment
    :param norm_by_average_sequence_length: bool - True if the average pairwise identity of two sequences is to be
                                            computed as
                                            (number of identical positions)/
                                            (average ungapped length of the two sequences).
                                            False if
                                            (number of identical positions)/
                                            (length of longer ungapped sequence) should be
                                            used.
    :return: float - the sequence identity of the given sequences
    """
    first_sequence = re.sub('-', '', first_sequence_gapped)
    second_sequence = re.sub('-', '', second_sequence_gapped)
    # Pairwise alignment to determine sequence identity.
    if perform_sequence_alignment:
        # Perform a pairwise realignment of the sequences where identical characters have score 1,
        # otherwise 0 and no gap penalties are used.
        #if (len(first_sequence) * len(second_sequence) > 1000000 ):
        #  If sequences were (on geom. average) larger than 1000 nucleotides
        sequence_1, sequence_2 = nwalign.global_align(first_sequence, second_sequence,
                                                      gap_open=0, gap_extend=0, match=1, matrix=current_dir+"/M1MM0")
        nwalign_score = nwalign.score_alignment(sequence_1, sequence_2, gap_open=0, gap_extend=0,
                                                matrix=current_dir+"/M1MM0")
        score = float(nwalign_score)

        # sanity check for alignment
        assert len(sequence_1) == len(sequence_2)
    else:
        # Use the sequence alignment from Rfam
        score = __count_matching_characters__(first_sequence_gapped, second_sequence_gapped)

    sequence_identity = pairwise_sequence_identity(score, norm_by_average_sequence_length,
                                                   len(first_sequence), len(second_sequence))
    return sequence_identity