#!/usr/bin/env python
import operator
import shutil

__author__ = 'Alexander Junge (alexander.junge@gmail.com)'

import argparse
import os
import logging
import collections
import itertools

from Bio import SeqIO
import pairwise_sequence_identity as pwseqid

logging.basicConfig(level=logging.DEBUG)


def is_positive_float(arg):
    float_arg = None
    try:
        float_arg = float(arg)
    except:
        parser.error('Given string {} is not a float.'.format(arg))
    if not 0 <= float_arg <= 1:
        parser.error('Given float {} is not in [0,1].'.format(arg))
    return float_arg


def is_valid_dir(arg):
    if not os.path.isdir(arg):
        parser.error("Given path %s is not a directory." % arg)
    return arg


def is_not_valid_path(arg):
    if os.path.exists(arg):
        parser.error("Given path %s does exist. Please specify a non-existing path.")
    return arg


parser = argparse.ArgumentParser(description=
                                 """
                                 Filter Rfamome data set to remove sequence redundancy.

                                 Human representative sequences from the same family are removed such that no two
                                 sequences have a PSI above a user-defined threshold.
                                 """)
parser.add_argument('input_dir', type=is_valid_dir,
                    help="""
                         The directory the filtered sequences are read from.
                         One Stockholm formatted file per human representative is assumed to be in this folder.
                         """)
parser.add_argument('sequence_identity', type=is_positive_float,
                    help="""
                         The maximal PSI tbetween sequences from the same family that is allowed.
                         """)
parser.add_argument('output_dir', type=is_not_valid_path,
                    help="""
                         The directory the filtered sequences are written to.
                         """)
args = parser.parse_args()

input_dir = args.input_dir
output_dir = args.output_dir
seq_id_threshold = args.sequence_identity

try:
    os.makedirs(output_dir, mode=0o774)
except OSError:
    raise OSError('Failed to create output directory {}.'.format(output_dir))

input_file_paths = [os.path.join(input_dir, p)
                    for p in sorted(os.listdir(input_dir)) if os.path.isfile(os.path.join(input_dir, p))]
logging.info('Found {:d} files in input directory'.format(len(input_file_paths)))

if not all([os.path.basename(f).endswith('.stk') for f in input_file_paths]):
    raise ValueError('Detected files in input directory that do not end on .stk. Stopping.')

if not all([len(os.path.basename(f).split('.')) == 3 for f in input_file_paths]):
    raise ValueError('Did not detect class label for all input files. Expected naming is: <class>.<integer>.stk')

input_file_classes = [os.path.basename(f).split('.')[0] for f in input_file_paths]
logging.info('Detected class assignment for {:d} classes.'.format(len(set(input_file_classes))))

# collect Stockholm files belonging to same class
class_to_input_file_paths = collections.defaultdict(list)
for curr_class in set(input_file_classes):
    curr_class_paths = [file_path for idx, file_path in enumerate(input_file_paths)
                        if input_file_classes[idx] == curr_class]
    class_to_input_file_paths[curr_class] = curr_class_paths

# map each Stockholm file to its human representative sequence
file_to_sequence = {}
for input_f_path in input_file_paths:
    seq_count = 0
    sequence = None
    with open(input_f_path, 'r') as stk_file:
        for record in SeqIO.parse(stk_file, 'stockholm'):
            # Convert all nucleotides to upper case
            sequence = str(record.seq.ungap('-')).upper()
            seq_count += 1
    if seq_count != 1:
        raise IOError('Stockholm file {} does contain more than one sequence record. Exiting.'.format(input_f_path))
    file_to_sequence[input_f_path] = sequence

# each class, compute sequence identity between human representatives
class_pair_to_sequence_id = {}
logging.info('Computing sequence identities of representatives.')
for curr_class, curr_files in class_to_input_file_paths.iteritems():
    logging.debug('Processing class {}.'.format(curr_class))
    for file_pair in itertools.combinations(curr_files, r=2):
        key = tuple(sorted(file_pair))
        sequence_identity = pwseqid.get_sequence_identity(file_to_sequence[key[0]], file_to_sequence[key[1]],
                                                          True, False)
        class_pair_to_sequence_id[key] = sequence_identity
        logging.debug('Sequence identity of {} and {} is {:f}'.format(key[0], key[1], sequence_identity))

# for each class, find representatives such that no pair of sequences has sequence_id higher than the threshold
logging.info('Performing redundancy reduction.')
for curr_class, class_files in class_to_input_file_paths.iteritems():
    keep_class_files = list(class_files)
    file_to_high_sim_file_count = {}
    for class_f in class_files:
        file_to_high_sim_file_count[class_f] = 0
    while len(file_to_high_sim_file_count) > 1:
        file_to_high_sim_file_count = collections.defaultdict(int)
        for file_pair in itertools.combinations(keep_class_files, r=2):
            curr_key = tuple(sorted(file_pair))
            curr_seq_id = class_pair_to_sequence_id[curr_key]
            if curr_seq_id > seq_id_threshold:
                file_to_high_sim_file_count[file_pair[0]] += 1
                file_to_high_sim_file_count[file_pair[1]] += 1
        if len(file_to_high_sim_file_count) > 1:
            sorted_x = sorted(file_to_high_sim_file_count.items(), key=operator.itemgetter(1), reverse=True)
            rm_file = sorted_x[0][0]
            del file_to_high_sim_file_count[rm_file]
            keep_class_files.remove(rm_file)
    logging.debug('Class {}: Retaining {:d} out of {:d} representatives.'.format(curr_class, len(keep_class_files),
                                                                                 len(class_files)))
    for keep_file in keep_class_files:
        shutil.copy2(keep_file, output_dir)