#!/usr/bin/env python

__author__ = 'Alexander Junge, RTH, University of Copenhagen <ajunge@th.dk>'

import argparse
import os
from Bio import SeqIO
import itertools
import pairwise_sequence_identity as pwseqid
import collections
import sys
import shutil
import gzip
import csv
import matplotlib.pyplot as plt
import seaborn as sns  # makes plots look nicer, can be commented out without doing any harm
sns.set_palette("deep", desat=.6)


def is_valid_dir(arg):
    if not os.path.exists(arg) or not os.path.isdir(arg):
        parser.error("Given path %s is not a file." % arg)
    return arg


def is_not_valid_path(arg):
    if os.path.exists(arg):
        parser.error("Given path %s exists. Please specify an empty directory." % arg)
    return arg


def is_float_between_0_and_1(arg):
    value = None
    try:
        value = float(arg)
    except ValueError:
        parser.error("Given APSI threshold must be a float but is %s." % arg)
    if value < 0:
        parser.error("Given APSI threshold must be >= than 0 but is %f." % value)
    if value > 1:
        parser.error("Given APSI threshold must be <= than 1 but is %f." % value)
    return value


def is_positive_integer(arg):
    value = None
    try:
        value = int(arg)
    except ValueError:
        parser.error("Given length threshold must be an integer but is %s." % arg)
    if value < 0:
        parser.error("Given length threshold must be positive but is %d." % value)
    return value


def compute_apsi_per_file(file_list):
    file_name_to_apsi = {}

    for my_file in file_list:
        sequence_list = []
        with open(my_file, 'r') as input_file_handler:
            try:
                for record in SeqIO.parse(input_file_handler, 'stockholm'):
                    # Convert all nucleotides to upper case
                    sequence_list.append(str(record.seq).upper())
            except ValueError:
                continue

        apsi_list = []
        for pair in itertools.combinations(sequence_list, r=2):
            first_sequence_gaps, second_sequence_gaps = pair
            sequence_identity = pwseqid.get_sequence_identity(first_sequence_gaps, second_sequence_gaps,
                                                              False, False)
            apsi_list.append(sequence_identity)

        if len(apsi_list) > 0:
            apsi = sum(apsi_list)/len(apsi_list)
        else:
            apsi = float('nan')
        #sys.stderr.write('{} has ASPI {:f} and {:d} sequences.\n'.format(my_file, apsi, len(sequence_list)))
        file_name_to_apsi[my_file] = apsi

    return file_name_to_apsi


def find_apsi_bin(apsi, sorted_apsi_thresholds):
    sorted_apsi_threshold_iterator = iter(sorted_apsi_thresholds)

    lower_threshold = next(sorted_apsi_threshold_iterator)
    upper_threshold = next(sorted_apsi_threshold_iterator)
    if lower_threshold < apsi <= upper_threshold:
        bin_name = 'apsi-' + str(lower_threshold) + '-' + str(upper_threshold)
        return bin_name

    for new_threshold in sorted_apsi_threshold_iterator:
        lower_threshold = upper_threshold
        upper_threshold = new_threshold
        if lower_threshold < apsi <= upper_threshold:
            bin_name = 'apsi-' + str(lower_threshold) + '-' + str(upper_threshold)
            return bin_name

    return None


def assign_files_to_apsi_bins(file_name_to_apsi, sorted_apsi_thresholds):
    file_name_to_apsi_bin = {}
    for filename, apsi in file_name_to_apsi.items():
        current_bin = find_apsi_bin(apsi, sorted_apsi_thresholds)
        file_name_to_apsi_bin[filename] = current_bin
    return file_name_to_apsi_bin


def compute_asl_per_file(file_list):
    file_name_to_asl = {}

    for my_file in file_list:
        sequence_lengths = []
        with open(my_file, 'r') as input_file_handler:
            try:
                for record in SeqIO.parse(input_file_handler, 'stockholm'):
                    sequence_lengths.append(len(str(record.seq.ungap('-'))))
            except ValueError:
                continue
        if len(sequence_lengths) > 0:
            asl = sum(sequence_lengths)/float(len(sequence_lengths))
        else:
            asl = float('nan')

        file_name_to_asl[my_file] = asl
    return file_name_to_asl


def find_asl_bin(asl, sorted_asl_thresholds):
    sorted_asl_thresholds_iterator = iter(sorted_asl_thresholds)

    l = next(sorted_asl_thresholds_iterator)
    for h in sorted_asl_thresholds_iterator:
        if l < asl <= h:
            bin_name = 'asl-' + str(l) + '-' + str(h)
            return bin_name
        l = h

    return None


def assign_files_to_asl_bins(file_name_to_asl, sorted_asl_thresholds):
    file_name_to_asl_bin = {}
    for filename, asl in file_name_to_asl.items():
        current_bin = find_asl_bin(asl, sorted_asl_thresholds)
        file_name_to_asl_bin[filename] = current_bin
    return file_name_to_asl_bin


def print_bin_sizes(filename_to_bin):
    current_bin_to_files = collections.defaultdict(set)

    for file_name, current_bin in filename_to_bin.items():
        current_bin_to_files[current_bin].add(file_name)

    sys.stdout.write('Bin_name\tClique_count\n')
    for current_bin in sorted(current_bin_to_files.keys()):
        sys.stdout.write('{}\t{:d}\n'.format(str(current_bin), len(current_bin_to_files[current_bin])))


def plot_histogram(values, bin_number, plot_range, show_mean, title, xlabel, ylabel, output_file):
    plt.figure()
    if plot_range:
        plt.hist(values, bins=bin_number, range=plot_range)
    else:
        plt.hist(values, bins=bin_number)
    if show_mean:
        plt.axvline(float(sum(values))/len(values), color='black', linestyle='dashed', linewidth=2)
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.savefig(output_file)
    plt.close()


def print_values_to_file(print_map, header, out_file_name):
    with gzip.open(out_file_name, 'wb') as out_file:
        writer = csv.writer(out_file, delimiter='\t')
        writer.writerow(header)
        for k in sorted(print_map.keys()):
            v = print_map[k]
            writer.writerow([str(k), str(v)])


if __name__ == '__main__':

    ####################################################################################################################
    # Command line argument specification and parsing
    ####################################################################################################################
    parser = argparse.ArgumentParser(description=
                                     """
                                     Assign structural alignments to bins based on sequence identity or length.

                                     Based on criteria specified via command line arguments, the average pairwise
                                     sequence identity (APSI) or average sequence length (ASL)of each input alignment
                                     is computed.
                                     During APSI computation, the Stockholm files are assumed to contain already aligned
                                     sequence and hence no new alignments are computed.
                                     Afterwards, each alignment is assigned to a user-specified bin and the binned
                                     subsets are written to the given output folder.
                                     """)

    parser.add_argument('input_directory', type=is_valid_dir,
                        help="""
                             A directory containing the set of structural alignments to be split into bins.

                             The directory is assumed to contain a set of Stockholm files, where each file
                              contains a single structural alignment to be assigned to a bin.
                             """)
    parser.add_argument('output_directory', type=is_not_valid_path,
                        help="""
                             The output folder the binned data set is written to.

                             The given path is not allowed to exist prior to executing the program and will be created
                             at runtime, including potentially non-existing parent directories.
                             """)
    parser.add_argument('-i', '--apsi', nargs='+', type=is_float_between_0_and_1,
                        help="""
                             The APSI thresholds at which the data set is to be binned.

                             Each threshold has to be a float in the interval [0,1] and at least two thresholds have to
                             be given.
                             The binning is implemented as follows: Let 0.2, 0.4, 0.6, 0.8, 1.0 be the given thresholds.
                             Then four bins are created: (0.2, 0.4], (0.4, 0.6], (0.6, 0.8] (0.8, 1.0]
                             Each structural alignment in the input data set is assigned to the bin corresponding to
                             its APSI, or no filtered out if such a bin does not exist.
                             """)
    parser.add_argument('-l', '--length', nargs='+', type=is_positive_integer,
                        help="""
                             The ASL thresholds at which the data set is to be binned.

                             Each threshold is an integer >0 and at least two thresholds have to be given.
                             The binning is implemented as follows: Let 0, 100, 200, 300, 1000 be the given thresholds.
                             Then four bins are created: (0, 100], (100,200], (200,300], (300, 1000]
                             """)
    parser.add_argument('-n', '--no_statistics', action='store_true',
                        help="""
                             Disable computing statistics and creating analysis plots.

                             By default, a set of statistics are computed and analysis plots are created for each bin
                             and stored in a subdirectory 'statistics/' within each bin's output directory.
                             Set this flag to disable this.
                             """)
    parser.add_argument('-r', '--representative_directory', type=is_valid_dir,
                        help="""
                             Location of directory containing single sequence representatives for input alignments.

                             Instead of assigning input alignments to bins, their representatives are assigned to bins.
                             However, the binning is still evaluated on the criteria given for the input alignments
                             and not on the representatives.
                             """)

    args = parser.parse_args()

    in_folder = args.input_directory
    out_folder = args.output_directory
    apsi_thresholds = args.apsi
    asl_thresholds = args.length
    rep_folder = args.representative_directory

    if args.no_statistics:
        compute_statistics = False
    else:
        compute_statistics = True

    if not apsi_thresholds and not asl_thresholds:
        raise ValueError('No binning rule is specified. Please refer to -i and -l parameter, respectively.')

    if apsi_thresholds and len(apsi_thresholds) < 2:
        raise ValueError('Please specify at least two APSI thresholds to bin the data set at.')

    if asl_thresholds and len(asl_thresholds) < 2:
        raise ValueError('Please specify at least two ASL thresholds to bin the data set at.')

    ####################################################################################################################
    # Program logic
    ####################################################################################################################
    # read files from input directory
    in_files = [os.path.join(in_folder, f) for f in os.listdir(in_folder) if os.path.isfile(os.path.join(in_folder, f))]

    if rep_folder:
        file_to_rep_file = {}
        for in_f in in_files:
            rep_file_basename = os.path.basename(in_f)
            rep_file_basename = rep_file_basename.replace('.', '_rep.')
            rep_file = os.path.join(rep_folder, rep_file_basename)
            if not os.path.isfile(rep_file):
                raise ValueError('Representative file {} for alignment file {} does not exist.'.format(rep_file, in_f))
            file_to_rep_file[in_f] = rep_file

    file_classes = [os.path.splitext(os.path.basename(full_path))[0] for full_path in in_files]
    file_2_class = dict(zip(in_files, file_classes))

    file_to_apsi = None
    file_to_apsi_bin = None
    file_to_asl = None
    file_to_asl_bin = None
    if apsi_thresholds:
        apsi_thresholds_sorted = sorted(set(apsi_thresholds))
        file_to_apsi = compute_apsi_per_file(in_files)
        file_to_apsi_bin = assign_files_to_apsi_bins(file_to_apsi, apsi_thresholds_sorted)
    if asl_thresholds:
        asl_thresholds_sorted = sorted(set(asl_thresholds))
        file_to_asl = compute_asl_per_file(in_files)
        file_to_asl_bin = assign_files_to_asl_bins(file_to_asl, asl_thresholds_sorted)

    # if binning both one APSI and ASL is chosen, create a joint binning
    if apsi_thresholds and asl_thresholds:
        file_to_bin = {}
        for curr_file in in_files:
            apsi_bin = file_to_apsi_bin[curr_file]
            asl_bin = file_to_asl_bin[curr_file]
            if apsi_bin and asl_bin:
                new_bin = apsi_bin + '-' + asl_bin
            elif apsi_bin:
                new_bin = apsi_bin
            elif asl_bin:
                new_bin = asl_bin
            else:
                new_bin = None
            file_to_bin[curr_file] = new_bin
    elif apsi_thresholds:
        file_to_bin = file_to_apsi_bin
    elif asl_thresholds:
        file_to_bin = file_to_asl_bin
    else:
        raise ValueError('No binning rule is specified. Please refer to -i and -l parameter, respectively.')

    # copy files to output folder, one sub-folder per bin is created
    os.makedirs(out_folder)
    bin_names = sorted(set(file_to_bin.values()))
    # remove bin that might be None
    bin_names = [x for x in bin_names if x]
    # map each bin's name to its files, used for computing bin-wise statistics below
    bin_name_to_file_list = collections.defaultdict(list)
    for curr_bin_name in bin_names:
        os.makedirs(os.path.join(out_folder, curr_bin_name))
    for curr_file, curr_bin in file_to_bin.items():
        if curr_bin:
            bin_name_to_file_list[curr_bin].append(curr_file)
            if not rep_folder:
                shutil.copy2(curr_file, os.path.join(out_folder, curr_bin))
            else:
                shutil.copy2(file_to_rep_file[curr_file], os.path.join(out_folder, curr_bin))

    # size of bins, measured in number of alignments, is always printed to stdout
    print_bin_sizes(file_to_bin)

    if not compute_statistics:
        sys.exit(0)

    for curr_bin_name, curr_file_list in bin_name_to_file_list.iteritems():
        curr_statistics_out_folder = os.path.join(out_folder, curr_bin_name, 'statistics')
        os.makedirs(curr_statistics_out_folder)
        if not file_to_apsi:
            file_to_apsi = compute_apsi_per_file(in_files)
        if not file_to_asl:
            file_to_asl = compute_asl_per_file(in_files)

        bin_apsi_values = []
        for curr_file in curr_file_list:
            bin_apsi_values.append(file_to_apsi[curr_file])
        bin_file_to_apsi = {}
        for curr_file in curr_file_list:
            bin_file_to_apsi[curr_file] = file_to_apsi[curr_file]

        plot_histogram(values=bin_apsi_values, bin_number=30, plot_range=(0, 1), show_mean=True,
                       title='Bin: {}'.format(curr_bin_name),
                       xlabel='APSI in subalignment (based on Rfam alignments)',
                       ylabel='Subalignment count, total: {:d}'.format(len(bin_apsi_values)),
                       output_file=os.path.join(curr_statistics_out_folder, 'histogram-apsi.pdf'))
        print_values_to_file(print_map=bin_file_to_apsi,
                             header=('Subalignment', 'APSI'),
                             out_file_name=os.path.join(curr_statistics_out_folder, 'apsi.tsv.gz'))

        bin_asl_values = []
        for curr_file in curr_file_list:
            bin_asl_values.append(file_to_asl[curr_file])
        bin_file_to_asl = {}
        for curr_file in curr_file_list:
            bin_file_to_asl[curr_file] = file_to_asl[curr_file]
        plot_histogram(values=bin_asl_values, bin_number=30, plot_range=None, show_mean=True,
                       title='Bin: {}'.format(curr_bin_name),
                       xlabel='ASL in subalignment',
                       ylabel='Subalignment count, total: {:d}'.format(len(bin_asl_values)),
                       output_file=os.path.join(curr_statistics_out_folder, 'histogram-asl.pdf'))
        print_values_to_file(print_map=bin_file_to_asl,
                             header=('Subalignment', 'ASL'),
                             out_file_name=os.path.join(curr_statistics_out_folder, 'asl.tsv.gz'))

        # number of instances per class in bin
        bin_classes = []
        for curr_file in curr_file_list:
            bin_classes.append(file_2_class[curr_file])
        bin_class_2_size = collections.defaultdict(int)
        for bin_class in bin_classes:
            bin_class_2_size[bin_class] += 1
        bin_class_sizes = list(bin_class_2_size.values())
        plot_histogram(values=bin_class_sizes, bin_number=30, plot_range=None, show_mean=True,
                       title='Bin: {}'.format(curr_bin_name),
                       xlabel='Instances in class',
                       ylabel='Class count, total: {:d}'.format(len(bin_class_2_size.keys())),
                       output_file=os.path.join(curr_statistics_out_folder, 'histogram-class-sizes.pdf'))
        print_values_to_file(print_map=bin_class_2_size,
                             header=('Class', 'Instances'),
                             out_file_name=os.path.join(curr_statistics_out_folder, 'class-sizes.tsv.gz'))
        # number of classes with at least a certain number of instances
        instance_count_thresholds = range(1, 11)
        instance_threshold_2_class_count = {}
        for instance_threshold in instance_count_thresholds:
            instance_threshold_2_class_count[instance_threshold] =\
                sum([x >= instance_threshold for x in bin_class_sizes])
        print_values_to_file(print_map=instance_threshold_2_class_count,
                             header=('MinimalNumberOfInstancesInClass', 'Classes'),
                             out_file_name=os.path.join(curr_statistics_out_folder,
                                                        'class-count-minimal-instance-count.tsv.gz'))