#!/usr/bin/env python

__author__ = 'Alexander Junge (alexander.junge@gmail.com)'

import logging
import tempfile
import shutil
import errno
import os
import subprocess
import time


class EDeNClustering:
    """
    Wrapper running a C++ EDeN clustering once the method fit_predict is called.
    """

    def __init__(self, eden_absolute_path, gspan_file_path, eden_parameters):
        """
        Generates a clustering object wrapping a call to C++ EDeN clustering.

        :param eden_absolute_path: a string, the location of the C++ EDeN binary
        :param gspan_file_path: a string, the location of the GSPAN file to be clustered
        :param eden_parameters: a string, parameters to be passed on to EDeN
        """
        self.eden_path = eden_absolute_path
        self.gspan_path = gspan_file_path
        self.eden_params = eden_parameters

        self.target_values = self.__compute_target_values__()

    @staticmethod
    def __load_clustering__(cluster_path):
        object_2_cluster = {}
        with open(cluster_path, 'r') as cluster_file:
            for cluster_number, line in enumerate(cluster_file):
                for obj in line.split():
                    if obj in object_2_cluster:
                        raise IOError('Input C++ EDeN clustering result read from {} assigns object {} to '
                                      'more than one cluster.'.format(cluster_path, obj))
                    else:
                        object_2_cluster[obj] = cluster_number
        logging.debug('Loaded cluster assignments for {:d} objects.'.format(len(object_2_cluster)))
        return object_2_cluster

    def fit_predict(self, matrix):
        """
        Runs a C++ EDeN clustering and returns the result.

        :param matrix: a (dis)similarity matrix. Note that this matrix is not used since an external C++ call based on
                       a previously computed GSPAN file is made that does not depend on an input (dis)similarity matrix.
        :return: list of strings, encoding the clustering result. The cluster assignment of object at index i is found
                 at position i of this list.
        """
        if matrix:
            logging.warning('Given (dis)similarity matrix ignored since a C++ EDeN clustering is run using GSPAN files')

        tmp_dir = tempfile.mkdtemp(dir=os.getcwd())
        logging.debug('Created temporary directory for clustering at {}.'.format(tmp_dir))
        tmp_shell_file = tempfile.mkstemp(suffix='.sh', dir=tmp_dir)[1]

        eden_out_file = os.path.join(tmp_dir, 'EDeN.log')
        eden_err_file = os.path.join(tmp_dir, 'EDeN.err')
        cluster_program = """
                          #!/usr/bin/env bash
                          
                          date
                          {} --action CLUSTER --file_type GRAPH --cluster_type K_QUICK_SHIFT {} -i {} -y {} 1>{} 2>{}
                          date
                          """.format(self.eden_path, self.eden_params, self.gspan_path, tmp_dir, eden_out_file,
                                     eden_err_file)

        instance_2_cluster = None
        cluster_result_path = None
        try:
            with open(tmp_shell_file, 'w') as f:
                f.write(cluster_program)

            with open(os.devnull, 'w') as dev_null:
                subprocess.call('qsub -N {} -q default -l nodes=1:ppn=8 {}'.format(tmp_shell_file, tmp_shell_file),
                                stdout=dev_null, stderr=dev_null, shell=True) 
            cluster_result_path = os.path.join(tmp_dir, 'kquickshift_cluster')

            # check every ten seconds if output has been created
            while not os.path.exists(cluster_result_path):
                time.sleep(10)
            time.sleep(1000)

            instance_2_cluster = EDeNClustering.__load_clustering__(cluster_result_path)
        finally:
            try:
                # shutil.rmtree(tmp_dir)  # delete tmp directory
                pass
            except OSError as exc:
                if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
                    raise  # re-raise exception

        obj_count = len(instance_2_cluster)
        all_objects = range(obj_count)
        cluster_assignment = []
        for curr_obj in all_objects:
            curr_obj = str(curr_obj)
            if not curr_obj in instance_2_cluster:
                raise IOError('Input C++ EDeN clustering result read from {} does not assign object {} '
                              'to a cluster.'.format(cluster_result_path, curr_obj))
            else:
                cluster_assignment.append(instance_2_cluster[curr_obj])

        # sanity check that number of objects is the same as in the cluster result and target clustering
        if len(self.load_eden_target_values()) != len(cluster_assignment):
            raise IOError('Loaded C++ clustering ({:d} objects) and EDeN target clustering  ({:d} objects) contain different numbers of objects.'.format(len(cluster_assignment), len(self.load_eden_target_values())))

        return cluster_assignment

    def __compute_target_values__(self):
        tmp_dir = tempfile.mkdtemp(dir=os.getcwd())
        tmp_class_file = tempfile.mkstemp(prefix='class', dir=tmp_dir)[1]
        tmp_dict2_file = tempfile.mkstemp(prefix='dict2', dir=tmp_dir)[1]
        tmp_target_file = tempfile.mkstemp(prefix='target', dir=tmp_dir)[1]

        cmd1 = "zcat -f {} | grep \"^t\" | awk -v column=4 '{{print $column}}'  | " \
               "sed 's/^\([^.]*\).*/\\1/' > {}".format(self.gspan_path, tmp_class_file)
        cmd2 = "cat {} | sort | uniq -c | cat -n  | " \
               "awk '{{print $1\"\\t\"$3\"\\t\"$2}}' > {}".format(tmp_class_file, tmp_dict2_file)
        cmd3 = "cat {} | awk -v HF={} 'BEGIN{{while(getline <HF){{H[$2]=$1}}}} {{print H[$1]}}' " \
               "> {}".format(tmp_class_file, tmp_dict2_file, tmp_target_file)

        target_values = None
        try:
            with open(os.devnull, 'w') as dev_null:
                subprocess.call(cmd1, stdout=dev_null, stderr=dev_null, shell=True)
                subprocess.call(cmd2, stdout=dev_null, stderr=dev_null, shell=True)
                subprocess.call(cmd3, stdout=dev_null, stderr=dev_null, shell=True)

            target_values = []
            with open(tmp_target_file, 'r') as target_file:
                for line in target_file:
                    target = line.strip()
                    target_values.append(target)
        finally:
            try:
                # shutil.rmtree(tmp_dir)  # delete tmp directory
                pass
            except OSError as exc:
                if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
                    raise  # re-raise exception

        return target_values

    def load_eden_target_values(self):
        """

        :return: list of strings, encoding the clustering result. The target cluster assignment of object at index i is
                 found at position i of this list.
        """
        return self.target_values