#!/usr/bin/env python
import argparse
import os
import collections
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import seaborn as sns
sns.set_context("talk")
sns.set_style("whitegrid")

__author__ = 'Alexander Junge (alexander.junge@gmail.com)'

parser = argparse.ArgumentParser(description='''
Reproduces Figures and Tables in Results section of the RNAscClust publication.
Since this is supposed to be called only after the 'run_clustering_docker.sh' script and is not intended for general
applications, all file paths are hard coded.
                            ''')
parser.add_argument('--no_knn_analysis', action='store_true',
                    help='set this to avoid KNN analysis step')
args = parser.parse_args()

if args.no_knn_analysis:
    perform_knn_analysis = False
else:
    perform_knn_analysis = True

eval_result_dir = 'cluster_evaluation'
data_set_sub_dirs = ('cliques-high', 'cliques-low', 'cliques-medium', 'rfamome')
data_set_plot_names = ('Rfam-cliques High', 'Rfam-cliques Low', 'Rfam-cliques Medium', 'Rfam-ome')
cluster_method_sub_dirs = ('RESULTS_SC', 'RESULTS_GC')
cluster_method_plot_names = ('RNAscClust', 'GraphClust')
eval_sub_folder = 'cluster_evaluation'
knn_analysis_filename = 'knn_analysis.tsv'
cluster_analysis_filename = 'cluster_analysis.tsv'

knn_measure_to_plot_name = {'precision_weighted': 'Precision',
                            'recall_weighted': 'Recall',
                            'f1_weighted': 'F1-Score'}
cluster_measure_to_plot_name = {'Homogeneity': 'Homogeneity',
                                'Completeness': 'Completeness',
                                'V-measure': 'V-measure',
                                'RandIndex': 'Rand Index',
                                'AdjustedRandIndex': 'Adjusted Rand Index',
                                "Matthew's correlation coefficient": "Matthew's correlation coefficient",
                                'ClusterCount': 'Cluster Count',
                                'TrueClassCount': 'True Class Count'}

plot_output_folder = os.path.join(eval_result_dir)

knn_data = collections.defaultdict(dict)
for data_set in data_set_plot_names:
    for meas in knn_measure_to_plot_name.values():
        # noinspection PyTypeChecker
        knn_matrix = [[nn] + [float('nan')] * 4 for nn in [1, 3, 5, 7]]
        knn_data[data_set][meas] = knn_matrix

cluster_data = {'tool': ['AC', 'AC', 'AC', 'AC', 'GC', 'GC', 'GC', 'GC'],
                'dataset': ['Rfam-cliques High\n(Mean PSI 0.8)', 'Rfam-cliques Medium\n(Mean PSI 0.64)',
                            'Rfam-cliques Low\n(Mean PSI 0.49)', 'Rfam-ome',
                            'Rfam-cliques High\n(Mean PSI 0.8)', 'Rfam-cliques Medium\n(Mean PSI 0.64)',
                            'Rfam-cliques Low\n(Mean PSI 0.49)', 'Rfam-ome'],
                'Homogeneity': [float('nan')] * 8,
                'Completeness': [float('nan')] * 8,
                'V-measure': [float('nan')] * 8,
                'Rand Index': [float('nan')] * 8,
                'Adjusted Rand Index': [float('nan')] * 8}
cluster_method_plot_name_to_abbreviation = {'RNAscClust': 'AC',
                                            'GraphClust': 'GC'}
data_set_plot_name_to_description = {'Rfam-cliques High': 'Rfam-cliques High\n(Mean PSI 0.8)',
                                     'Rfam-cliques Low': 'Rfam-cliques Low\n(Mean PSI 0.49)',
                                     'Rfam-cliques Medium': 'Rfam-cliques Medium\n(Mean PSI 0.64)',
                                     'Rfam-ome': 'Rfam-ome'}

for i, data_set_dir in enumerate(data_set_sub_dirs):
    data_set_plot_name = data_set_plot_names[i]
    for j, method_dir in enumerate(cluster_method_sub_dirs):
        method_plot_name = cluster_method_plot_names[j]
        curr_knn_analysis_file = os.path.join(eval_result_dir, data_set_dir, method_dir, eval_sub_folder,
                                              knn_analysis_filename)
        curr_cluster_analysis_file = os.path.join(eval_result_dir, data_set_dir, method_dir, eval_sub_folder,
                                                  cluster_analysis_filename)

        if not os.path.isfile(curr_cluster_analysis_file):
            raise ValueError('Following file was not found %s.' % curr_cluster_analysis_file)

        with open(curr_cluster_analysis_file) as cluster_file:
            metric_names = cluster_file.readline().rstrip('\n\r').split(',')
            metric_values = map(float, cluster_file.readline().rstrip('\n\r').split(', '))
            metric_values = map(lambda x: round(x, 2), metric_values)
            current_performance_measures = {}
            for idx, metric_name in enumerate(metric_names):
                current_performance_measures[cluster_measure_to_plot_name[metric_name.strip()]] = metric_values[idx]
            method_abbreviation = cluster_method_plot_name_to_abbreviation[method_plot_name]
            data_set_description = data_set_plot_name_to_description[data_set_plot_name]
            write_indices = set([i for i, t in enumerate(cluster_data['tool']) if t == method_abbreviation])
            assert len(write_indices) == 4
            write_indices = write_indices.intersection(set([i for i, t in enumerate(cluster_data['dataset'])
                                                            if t == data_set_description]))
            assert len(write_indices) == 1
            write_indices = list(write_indices)[0]
            for metric, metric_val in current_performance_measures.iteritems():
                if metric in cluster_data:
                    cluster_data[metric][write_indices] = metric_val

        if perform_knn_analysis:
            if not os.path.isfile(curr_knn_analysis_file):
                raise ValueError('Following file was not found %s.' % curr_knn_analysis_file)
            with open(curr_knn_analysis_file) as knn_file:
                for line in knn_file:
                    measure_name, k, measure_value, measure_sd = line.rstrip().split('\t')
                    line_no = int(k) // 2
                    measure_plot_name = knn_measure_to_plot_name[measure_name]
                    if method_plot_name == 'GraphClust':
                        measure_index = 3
                        sd_index = 4
                    else:
                        measure_index = 1
                        sd_index = 2
                    measure_row = knn_data[data_set_plot_name][measure_plot_name][line_no]
                    assert measure_row[0] == int(k)
                    measure_row[measure_index] = float(measure_value)
                    measure_row[sd_index] = float(measure_sd)

########################################################################################################################
# K-nearest neighbor analysis
########################################################################################################################
mycolors = ['#1b9e77', '#d95f02', '#7570b3']
if perform_knn_analysis:
    # one plot per #neighbors
    dataset_plot_order = ['Rfam-ome', 'Rfam-cliques High', 'Rfam-cliques Medium', 'Rfam-cliques Low']
    clique_low_left = True  # if True, the Rfam-cliques Low data set is plotted on the left; if False, on the right
    if clique_low_left:
        dataset_plot_order = [dataset_plot_order[i] for i in ( 3, 2, 1, 0)]
    dataset_descriptions_plot_order = [data_set_plot_name_to_description[dataset] for dataset in dataset_plot_order]
    dataset_descriptions_plot_order = [d if d != 'Rfam-ome' else 'Rfam-ome\n(Mean PSI 0.82)' for d in dataset_descriptions_plot_order]

    k_to_plot_name = {1: 'figure4.pdf',
                      3: 'supplement_figure1.pdf'}
    k_to_alternative_plot_name = {1: 'figure4_alternative.pdf',
                                  3: 'supplement_figure1_alternative.pdf'}

    target_K = [1, 3]
    for iK, K in enumerate(target_K):
        ###############################################
        # depiction using barplot
        fig = plt.figure(figsize=(9, 9), dpi=600)
        fig.subplots_adjust(hspace=0.2)
        width = 0.3  # the width of the bars
        K_measures = dict()
        # reshape raw data format
        for nn, maxSI in enumerate(knn_data):
            #         fig.subplots_adjust(hspace=0.5)
            d = knn_data[maxSI]
            for n, measure in enumerate(d):

                #             print measure
                if measure in K_measures:
                    K_measures[measure] += [[maxSI] + [m for m in d[measure] if m[0] == K]]
                else:
                    K_measures[measure] = [[maxSI] + [m for m in d[measure] if m[0] == K]]

        # Now use the pruned data to plot
        # Plot bars in following order
        measure_plot_order = ('Precision', 'Recall', 'F1-Score')

        for imeas, measure in enumerate(measure_plot_order):

            ax = plt.subplot(3, 1, imeas + 1, adjustable='box')

            means1 = []
            means2 = []
            std1 = []
            std2 = []
            for dataset in dataset_plot_order:
                for data_set_name, values in K_measures[measure]:
                    k_val, mean_RAC, std_dev_RAC, mean_GC, std_dev_GC = values
                    if data_set_name == dataset:
                        means1.append(mean_RAC)
                        means2.append(mean_GC)
                        std1.append(std_dev_RAC)
                        std2.append(std_dev_GC)
            # print means1, std1
            # print means2, std2

            rects1 = ax.bar([0, 1, 2, 3], means1, width, color=mycolors[2], yerr=std1, ecolor='black')
            rects2 = ax.bar(np.array([0, 1, 2, 3]) + width, means2, width, color=mycolors[1], yerr=std2, ecolor='black')

            ax.set_ylim(0.50, 1.00)
            ax.set_yticks([0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
            ax.set_ylabel(measure)
            ax.set_xticks(np.array([0, 1, 2, 3]) + width)
            if imeas == 0:
                # draw legend once in location loc
                plt.legend((rects1[0], rects2[0]), cluster_method_plot_names, ncol=2, loc=(0.0, 1.0))
                # add NN annotation
                plt.text(3, 1.03, '%d-Nearest Neighbor' % K, fontsize=14)
            if imeas == len(measure_plot_order) - 1:
                ax.set_xticklabels(dataset_descriptions_plot_order)
            else:
                ax.set_xticklabels([])

        fig.savefig(os.path.join(plot_output_folder, k_to_plot_name[K]), dpi=600, bbox_inches='tight')
        
         # #############################################################
        # # Write down KNN analysis in Latex friendly format and order
        measure_plot_order = ('Precision', 'Recall', 'F1-Score')
        with open(os.path.join(plot_output_folder, 'table-CV-{}NN.txt'.format(K)), 'w') as tab_f:
            tab_f.write('% CV KNN for K = {}\n'.format(K))
            tab_f.write( "Dataset & " +  "& ".join(measure_plot_order))
            tab_f.write( "\\\\\\midrule\n% dataset, mean_RAC, std_dev_RAC, mean_GC, std_dev_GC\\\\\n")
            for dataset in dataset_plot_order:
                tab_f.write(dataset,)
#                tab_f.write( "  & ")
                for imeas, measure in enumerate(measure_plot_order):
                    for data_set_name, values in K_measures[measure]:
                        k_val, mean_RAC, std_dev_RAC, mean_GC, std_dev_GC = values
                        if data_set_name == dataset:
#                            tab_f.write( "& ".join(map(str, [mean_RAC, std_dev_RAC, mean_GC, std_dev_GC]))) #comma seperated string of list elements
                            tab_f.write( "& {}$\\pm${} & {}$\\pm${}".format(mean_RAC, std_dev_RAC, mean_GC, std_dev_GC)) #comma seperated string of list elements

                tab_f.write("\\\\\n")
            tab_f.write("\\botrule\n")


    
        # ############################################### TODO?!
        # # depiction using line graph
        # # Set figure size and hdistance
        # fig = plt.figure(figsize=(7, 12), dpi=600)
        # fig.subplots_adjust(hspace=0.1)
        #
        # width = 0.6  # the width of the bars
        # measure_plot_to_y_lim = {'Precision': [0.58, 1.02], 'Recall': [0.58, 1.02], 'F1-Score': [0.58, 1.02]}
        # measure_plot_to_y_ticks = {'Precision': [0.6, 0.7, 0.8, 0.9, 1.0],
        #                            'Recall': [0.6, 0.7, 0.8, 0.9, 1.0],
        #                            'F1-score': [0.6, 0.7, 0.8, 0.9, 1.0]}
        #
        # for n, measure in enumerate(main_plot_measures):
        #     ax = plt.subplot(2, 1, n + 1, adjustable='box')
        #     rects1 = ax.plot(clique_cluster_data['X'][0:3], clique_cluster_data[measure][0:3], '-o', c=mycolors[2])
        #     rects2 = ax.plot(clique_cluster_data['X'][3:6], clique_cluster_data[measure][3:6], '-o', c=mycolors[1])
        #
        #     ax.set_ylabel(measure, rotation=90, size=13)
        #     plt.xticks(clique_cluster_data['X'], [])  # hide x tick labels
        #     ax.set_xlim(-0.05, 2.05)
        #     ax.set_frame_on(False)
        #
        #     # print "TICKS: ", 10 * max(clique_cluster_data[measure]), range(0, 10, 1), [r / 10.0 for r in range(0, 10, 1) if
        #     # r < 10 * max(clique_cluster_data[measure]) and r > 10 * min(clique_cluster_data[measure])]
        #     ax.set_yticks(main_plot_measure_to_y_ticks[measure])
        #     ax.set_ylim(main_plot_measure_to_y_lim[measure])
        #     plt.tick_params(right='off', top='off')
        #     ax.set_xmargin(0.8)
        #
        #     if n == 0:
        #         # draw legend once in location loc
        #         plt.legend((rects1[0], rects2[0]), ('RNAscClust', 'GraphClust'), ncol=2, loc=(0.15, 1), prop={'size': 13})
        #
        # plt.xticks(clique_cluster_data['X'], clique_cluster_data['dataset'])  # single x ticks labels at the bottom
        # fig.savefig(os.path.join(plot_output_folder, k_to_alternative_plot_name[K]), dpi=600, bbox_inches='tight')


########################################################################################################################
# Cluster analysis
########################################################################################################################
# determine indices corresponding to Rfam-ome set - the results for this data set are depicted as tables
# mycolorsB = ['#fdc086','#1f78b4','#7fc97f']
mycolorsB = ['#e41a1c', '#377eb8', '#4daf4a']
rfamome_indices = [i for i, d in enumerate(cluster_data['dataset']) if d == 'Rfam-ome']
assert len(rfamome_indices) == 2
rfamome_cluster_data = {}
clique_cluster_data = {}
for k, v in cluster_data.iteritems():
    rfamome_vec = []
    clique_vec = []
    for i, entry in enumerate(v):
        if i in rfamome_indices:
            rfamome_vec.append(entry)
        else:
            clique_vec.append(entry)
    rfamome_cluster_data[k] = rfamome_vec
    clique_cluster_data[k] = clique_vec

clique_cluster_data['X'] = [0, 1, 2, 0, 1, 2]

color = dict()
color['AC'] = mycolors[0]
color['GC'] = mycolors[1]

clique_low_left = True  # if True, the Rfam-cliques Low data set is plotted on the left; if False, on the right
if clique_low_left:
    rev_order = (2, 1, 0, 5, 4, 3)
    clique_cluster_data['dataset'] = [clique_cluster_data['dataset'][i] for i in rev_order]
    clique_cluster_data['Homogeneity'] = [clique_cluster_data['Homogeneity'][i] for i in rev_order]
    clique_cluster_data['Completeness'] = [clique_cluster_data['Completeness'][i] for i in rev_order]
    clique_cluster_data['V-measure'] = [clique_cluster_data['V-measure'][i] for i in rev_order]
    clique_cluster_data['Rand Index'] = [clique_cluster_data['Rand Index'][i] for i in rev_order]
    clique_cluster_data['Adjusted Rand Index'] = [clique_cluster_data['Adjusted Rand Index'][i] for i in rev_order]

# Set figure size and hdistance
fig = plt.figure(figsize=(7, 11), dpi=600)
fig.subplots_adjust(hspace=0.1)

width = 0.6  # the width of the bars
main_plot_measures = ['V-measure', 'Adjusted Rand Index']
main_plot_measure_to_y_lim = {'V-measure': [0.78, 1.02], 'Adjusted Rand Index': [0.58, 1.02]}
main_plot_measure_to_y_ticks = {'V-measure': [0.8, 0.85, 0.9, 0.95, 1],
                                'Adjusted Rand Index': [0.6, 0.7, 0.8, 0.9, 1.0]}

for n, measure in enumerate(main_plot_measures):
    ax = plt.subplot(2, 1, n + 1, adjustable='box')
    rects1 = ax.plot(clique_cluster_data['X'][0:3], clique_cluster_data[measure][0:3], '-o', c=mycolors[2])
    rects2 = ax.plot(clique_cluster_data['X'][3:6], clique_cluster_data[measure][3:6], '-o', c=mycolors[1])
 
    ax.set_ylabel(measure, rotation=90, size=13)
    plt.xticks(clique_cluster_data['X'], [])  # hide x tick labels
    ax.set_xlim(-0.05, 2.05)
    ax.set_frame_on(False)
 
    # print "TICKS: ", 10 * max(clique_cluster_data[measure]), range(0, 10, 1), [r / 10.0 for r in range(0, 10, 1) if
    # r < 10 * max(clique_cluster_data[measure]) and r > 10 * min(clique_cluster_data[measure])]
#     ax.set_yticks(main_plot_measure_to_y_ticks[measure])
    ax.set_yticks([0.6, 0.7, 0.8, 0.9, 1.0])

#     ax.set_ylim(main_plot_measure_to_y_lim[measure]) instead fix for all
    ax.set_ylim([0.6, 1.0])
    plt.tick_params(right='off', top='off')
    ax.set_xmargin(0.8)
 
    if n == 0:
        # draw legend once in location loc
        plt.legend((rects1[0], rects2[0]), ('RNAscClust', 'GraphClust'), ncol=2, loc=(0.15, 1), prop={'size': 13})
 
plt.xticks(clique_cluster_data['X'], clique_cluster_data['dataset'])  # single x ticks labels at the bottom
fig.savefig(os.path.join(plot_output_folder, 'figure5-new.pdf'), dpi=600, bbox_inches='tight')
# #Set figure size and hdistance

fig = plt.figure(figsize=(7, 13), dpi=600)
fig.subplots_adjust(hspace=0.1)

width = 0.2  # the width of the bars
xoffset = 0.4
main_plot_measure_to_y_lim = {'V-measure': [0.5, 1.0], 'Adjusted Rand Index': [0.5, 1.0]}

for n, measure in enumerate(main_plot_measures):
    ax = plt.subplot(2, 1, n + 1, adjustable='box')
    rects1 = ax.plot(np.array([0, 1, 2])+width/2.0,  clique_cluster_data[measure][0:3], ls="--", marker="o", markersize=8, color='black', linewidth=1.0)
    rects2 = ax.plot(np.array([0, 1, 2]) + width+width/2.0, clique_cluster_data[measure][3:6], ls="--", marker="o", markersize=8,  color='black', linewidth=1.0)
    rects1 = ax.bar([0, 1, 2],  clique_cluster_data[measure][0:3], width,  color=mycolorsB[1])
    rects2 = ax.bar(np.array([0, 1, 2]) + width, clique_cluster_data[measure][3:6], width, color=mycolorsB[2])
    
    ax.set_xticks(np.array([-0.2, 0, 1, 2])+width )
    ax.set_ylabel(measure, rotation=90, size=13)
    plt.xticks(clique_cluster_data['X'], [])  # hide x tick labels
    ax.set_ylim(main_plot_measure_to_y_lim[measure])

#     ax.set_xlim(-0.05, 2.05)
#     ax.set_frame_on(False)

    # print "TICKS: ", 10 * max(clique_cluster_data[measure]), range(0, 10, 1), [r / 10.0 for r in range(0, 10, 1) if
    # r < 10 * max(clique_cluster_data[measure]) and r > 10 * min(clique_cluster_data[measure])]
#     ax.set_yticks(main_plot_measure_to_y_ticks[measure])
#     ax.set_ylim(main_plot_measure_to_y_lim[measure])
    plt.tick_params(right='off', top='off')
#     ax.set_xmargin(0.8)
    plt.tick_params(axis='both', which='major', labelsize=12)
    if n == 0:
        # draw legend once in location loc
        plt.legend((rects1[0], rects2[0]), ('RNAscClust', 'GraphClust'), ncol=1, loc=(1, 0.9), prop={'size': 13})

plt.xticks(np.array([-0.2, 0, 1, 2])+width , [""]+clique_cluster_data['dataset'])  # single x ticks labels at the bottom
fig.savefig(os.path.join(plot_output_folder, 'figure5-barplot.pdf'), dpi=600, bbox_inches='tight')

# ############################################### TODO?!
# # depiction using barplot
# fig = plt.figure(figsize=(9, 9), dpi=600)
# fig.subplots_adjust(hspace=0.2)
# width = 0.3  # the width of the bars
#
# for imeas, measure in enumerate(main_plot_measures):
#     ax = plt.subplot(3, 1, imeas + 1, adjustable='box')
#
#     means1 = []
#     means2 = []
#     std1 = []
#     std2 = []
#     for dataset in main_plot_measures:
#         for data_set_name, values in K_measures[measure]:
#             k_val, mean_RAC, std_dev_RAC, mean_GC, std_dev_GC = values
#             if data_set_name == dataset:
#                 means1.append(mean_RAC)
#                 means2.append(mean_GC)
#                 std1.append(std_dev_RAC)
#                 std2.append(std_dev_GC)
#     # print means1, std1
#     # print means2, std2
#
#     rects1 = ax.bar([0, 1, 2, 3], means1, width, color=mycolors[2], yerr=std1, ecolor='black')
#     rects2 = ax.bar(np.array([0, 1, 2, 3]) + width, means2, width, color=mycolors[1], yerr=std2, ecolor='black')
#
#     ax.set_ylim(None, 1.05)
#     #         ax.set_xlabel()
#     ax.set_ylabel(measure)
#     ax.set_xticks(np.array([0, 1, 2, 3]) + width)
#     if imeas == 0:
#         # draw legend once in location loc
#         plt.legend((rects1[0], rects2[0]), cluster_method_plot_names, ncol=2, loc=(0.0, 1.0))
#         # add NN annotation
#         plt.text(3, 1.25, '%d-Nearest Neighbor' % K, fontsize=14)
#     if imeas == len(measure_plot_order) - 1:
#         ax.set_xticklabels(dataset_descriptions_plot_order)
#     else:
#         ax.set_xticklabels([])
# fig.savefig(os.path.join(plot_output_folder, k_to_plot_name[K]), dpi=600, bbox_inches='tight')

###############################################
# part for supplement

fig = plt.figure(figsize=(7, 16), dpi=600)
fig.subplots_adjust()

supp_plot_measures = ['Homogeneity', 'Completeness', 'Rand Index']
supp_plot_measures_to_y_lim = {'Homogeneity': [0.848, 1.01], 'Completeness': [0.68, 0.902], 'Rand Index': [0.848, 1.01]}

for n, measure in enumerate(supp_plot_measures):
    ax = plt.subplot(3, 1, n + 1, adjustable='box')
    rects1 = ax.plot(clique_cluster_data['X'][0:3], clique_cluster_data[measure][0:3], '-o', c=mycolors[2])
    rects2 = ax.plot(clique_cluster_data['X'][3:6], clique_cluster_data[measure][3:6], '-o', c=mycolors[1])

    ax.set_ylabel(measure, rotation=90, size=13)
    plt.xticks(clique_cluster_data['X'], [])  # hide x tick labels
    ax.set_xlim(-0.05, 2.05)
    ax.set_frame_on(False)

    # print "TICKS: ", 10 * max(clique_cluster_data[measure]), range(0, 10, 1), [r / 10.0 for r in range(0, 10, 1) if
    # r < 10 * max(clique_cluster_data[measure]) and r > 10 * min(clique_cluster_data[measure])]
    ax.set_ylim(supp_plot_measures_to_y_lim[measure])
    plt.tick_params(right='off', top='off')
    ax.set_xmargin(0.8)

    if n == 0:
        # draw legend once in location loc
        plt.legend((rects1[0], rects2[0]), ('RNAscClust', 'GraphClust'), ncol=2, loc=(0.15, 1), prop={'size': 13})

plt.xticks(clique_cluster_data['X'], clique_cluster_data['dataset'])  # single x ticks labels at the bottom
fig.savefig(os.path.join(plot_output_folder, 'supplement_figure2.pdf'), dpi=600, bbox_inches='tight')

# Save Rfam-ome clustering evaluations as tables
with open(os.path.join(plot_output_folder, 'table2.txt'), 'w') as tab_f:
    tab_f.write('Performance measure &  RNAscClust & GraphClust \\\\\\midrule\n')
    for meas in main_plot_measures:
        vec = rfamome_cluster_data[meas]
        tab_f.write(' & '.join([meas, str(vec[0]), str(vec[1])]) + '\\\\\n')
    tab_f.write('\\botrule')

with open(os.path.join(plot_output_folder, 'supplement_table1.txt'), 'w') as tab_f:
    tab_f.write('Performance measure &  RNAscClust & GraphClust \\\\\\midrule\n')
    for meas in supp_plot_measures:
        vec = rfamome_cluster_data[meas]
        tab_f.write(' & '.join([meas, str(vec[0]), str(vec[1])]) + '\\\\\n')
    tab_f.write('\\bottomrule')
