#!/usr/bin/env python

'''
  cluster_gspan
    ~~~~~~~~~~~~~
Script module cluster_gspan converts a GSPAN format graph file into sparse
feature vectors matrix.
Matrix feature vector is stored under output directory
and named X_array_S.npz where S is the number of vectorized clustering elements

Later: TODO: Run metric evaluation, cluster using dis-/similarity matrix.
Version 0.1
@author:    Milad Miladi
            miladim@informatik.uni-freiburg.de
'''
import os
import sys
import argparse
import logging


def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist" % arg)
    return arg


def is_valid_file(arg):
    if not os.path.isfile(arg):
        parser.error("Given file %s does not exist" % arg)
    return arg



def plot_results(names_to_plot, lists_to_plot, out_analysis_dir):
    import numpy

    from matplotlib import pyplot as plot
    from matplotlib.backends.backend_pdf import PdfPages

    # The PDF document
    pdf_name = '/AlignClust_reanalyze.pdf'
    pdf_pages = PdfPages(out_analysis_dir + pdf_name)

    fig = plot.figure(figsize=(15, 12), dpi=100)
    fig.subplots_adjust(left=0.2, wspace=0.3, hspace=0.3)
    for ip, plot_name in enumerate(names_to_plot):

        # Plot stuffs !
        pl = fig.add_subplot(220+ip+1)
        pl.set_xlabel(plot_name)
    #     plot.subplot2grid(grid_size, (ip % nb_plots_per_page, 0))
    #     plot.hist(lists_to_plot[ip], 100, normed=0, alpha=0.75)

        pl.grid(True)
        pl.hist(lists_to_plot[ip], bins=100, facecolor='blue', color='blue')

    #     plot.show()

    # Write the PDF document to the disk
    # plot.tight_layout()
    pdf_pages.savefig(fig)
    pdf_pages.close()


# =========================================================================
# Main() section
# Executed only if invoked  as main module, e.g. ran by command line
# =========================================================================
def main(args):
    logging.basicConfig(level=args.loglevel or logging.INFO)

    from GSPANVectorizer import GSPANVectorizer
    vectorizer = GSPANVectorizer(args.input_gspan, args.out_dir, saved_npz_vector=args.load_matrix,
                                 max_num_instances=args.num_instances, graphclust_format=args.graphclust_gspan,
			         max_distance=3, max_radius=3, simplify_gspan=True, num_jobs=3)
    if not args.only_load:
        vectorizer.do_vectorize()  # Vectorize the gspan file
    feature_vector = vectorizer.get_feature_vector()  # Get feature vector
    calc_cluster_metrics(feature_vector, vectorizer.ref_labels, vectorizer.ref_labels_id, args.out_dir)

# end main


if __name__ == "__main__":
    print "Args: ", sys.argv

    parser = argparse.ArgumentParser()
    argparse.ArgumentParser(description="""Generates and saves EDeN sparse feature
                                        vector from a GSPAN graph""")

    parser.add_argument('-v', '--verbose', help='Verbose (debug) logging', action='store_const',
                        const=logging.DEBUG, dest='loglevel')

    parser.add_argument('--input-gspan', type=is_valid_file, required=True,
                        help='GSPAN file to vectorize')

    parser.add_argument('--out-dir', type=is_valid_path, required=True,
                        help='Output directory for saving feature matrix ')

    parser.add_argument('--load-matrix', type=is_valid_file, required=False,
                        help='Load (partially) computed matrix ')

    parser.add_argument('--num-instances', type=int, default=-1,
                        help='Maximum number of instances to be loaded and vectorized ')

    parser.add_argument('--only-load', action='store_true',
                        help='Assume load matrix is not partial and fully computed')

    parser.add_argument('--graphclust-gspan', action='store_true',
                        help='input file is generated by GraphClust ')


    args = parser.parse_args()
    if args.load_matrix and args.load_matrix is None:
        parser.error("Parameter --only-load requires specifying matrix file through --load-matrix")

    main(args)
