#!/usr/bin/perl

#############################################################################
##############             RNAAlignClust              ##########################
#
# This program gets as input a Rfam like directory 
# , containing subdirectories of multiple alignments,
# Then maps the consensus structure of each alignment ot the its sequences
# For the consensus structure two option is available:
#	a) using the Rfam like .cons file in each directory 
#	b) run PETfold to produce consensus structure
# Then a constraint folding is done using RNAfold, where some sequences might 
# filtered out beacuse of not folding with compatiable to the consensus
# Output:
#	For each alignment, a fasta file containg the filtered sequences 
#	and their corresponding secondary structure with tag #S is produced
#	this output can be used as input by alignmnet2gspan.pl
################################################################################
#
#  Milad Miladi miladim@informatik.uni-freiburg.de
#
################################################################################

#use feature ':5.10';
use strict 'vars';
use warnings;
use Getopt::Long;
use Pod::Usage;
use File::Copy;
use List::Util qw(min max);
use File::Temp qw(tempdir);
use File::Path qw(rmtree);

use Cwd qw(getcwd abs_path);

use FindBin;                 # locate this script
use lib "$FindBin::Bin/../";  # use the parent directory
use Config::AlignClustConfig;

use lib "$FindBin::Bin/../lib/";  # use the parent directory
use StructureLibrary::Tools;
use StructureLibrary::Sequence;
use MultipleAlignment::Tools;
use MultipleAlignment::PETfold;
use ReadConfigFile;


###############################################################################
# GLOBAL VARIABLES
###############################################################################
my $CURRDIR = getcwd;
my $RUN_PETFOLD = "$FindBin::Bin/run_PETfold.pl";
my $RUN_align_to_SS = "$FindBin::Bin/alignment_to_secondary_structures.pl";

my $CLUSTERSUBMIT =    "$FindBin::Bin/../lib/StructureLibrary/generic_submit_to_cluster.sh";
my $STOCKHOLM_SUFFIX = ".stk"; #".stockholm";
my $CONSENSUS_SUFFIX = ".cons";
my $CLUSTAL_SUFFIX = ".clustal";
my $ALN_FASTA_SUFFIX = ".aln.fasta";
my %FAMILIES = ();
my %RFAMSEED = ();
#my %RFAMFILTERED = ();

=head1 NAME


iterate_over_alignments.pl -input-dir "alignments directory" 

=head1 SYNOPSIS

Options:

		
        -help    		brief help message
        -man     		full documentation
        -input-dir		directory containing stockholm files
        -stdout  		redirect result fasta to stdout
    	-local-config		if a local configuration is to be used, path to the local configuration file
        -debug	 		debug mode
        -result-dir	 	output directory for results
	-tmp-dir		path of temporaty directory
	-keep-tmp		if this flag is set, the temporary directory is not deleted after execution
        -cutoff-ratio-consensus=i
        -cutoff-ratio-MFE=i [=20%]  
        -use-petfold	[=20%]
        -read-stockholm
		-read-aln-fasta 
        -threshold-reliablity=f
        -SGE
        -sge-logdir=s
	-sge-errdir=s
	-sge-submit-host=s
	-sge-submit-user=s
         

=head1 DESCRIPTION

=cut


###############################################################################
# PARSE COMMAND LINE OPTIONS
###############################################################################

# command line options
my ($i_help, $i_man, $i_dir, $i_stdout, $i_debug, $i_local_config, 
	$i_result_dir, $i_tmp_dir, $i_keep_tmp, $i_cutoff_cons, $i_cutoff_mfe, 
	$i_petfold, $i_stockholm, $i_clustal, $i_aln_fasta, $i_threshold_reli, 
	$i_sge, $i_sge_logDir, $i_sge_errDir,
	$i_sge_submit_host,    $i_sge_submit_user);
$i_cutoff_cons = 20;
$i_cutoff_mfe = 20;
$i_threshold_reli = 0.001;
my $options = GetOptions ("help"	=> \$i_help,
                         "man"  	=> \$i_man,
                         "stdout"  	=> \$i_stdout,
			 "local-config=s" => \$i_local_config,
                         "debug"	=> \$i_debug,
                         "input-dir=s"	=> \$i_dir,
                         "result-dir=s"	=> \$i_result_dir,
			 "tmp-dir=s" => \$i_tmp_dir, 
			 "keep-tmp" => \$i_keep_tmp,
                         "cutoff-ratio-consensus=i" => \$i_cutoff_cons,
                         "cutoff-ratio-MFE=i" => \$i_cutoff_mfe,
                         "use-petfold" => \$i_petfold,
                         "read-stockholm" => \$i_stockholm,
                         "read-clustal" => \$i_clustal,
						 "read-aln-fasta" => \$i_aln_fasta,
                         "threshold-reliablity=f" => \$i_threshold_reli,
                         "SGE" => \$i_sge,
			  "sge-logdir=s"    => \$i_sge_logDir,
			  "sge-errdir=s"    => \$i_sge_errDir,
			  "sge-submit-host=s" => \$i_sge_submit_host,
			  "sge-submit-user=s" => \$i_sge_submit_user

                         );

                      
pod2usage(-exitstatus => 1, -verbose => 1) if $i_help;
pod2usage(-exitstatus => 0, -verbose => 2) if $i_man;
($options) or pod2usage(2);
($i_dir) or pod2usage("error: -input-dir is not specified!\n");
(-e $i_dir) or pod2usage("error: no such directory: $i_dir\n");

if ($i_local_config) {
	(-e $i_local_config) or pod2usage("error: local configuration file does not exist: $i_local_config\n");
}

if ($i_result_dir)
{
	(-e $i_result_dir) or pod2usage("error: no such directory: $i_result_dir\n");
	$i_result_dir = abs_path($i_result_dir);
}
else
{
	$i_result_dir = abs_path($CURRDIR) ;
} 

($i_cutoff_cons >= 0 and $i_cutoff_cons <= 100) or pod2usage("error: cutoff ratio consesnusis percentage and must be an integer between 0 and 100");
($i_cutoff_mfe >= 0 and $i_cutoff_mfe <= 100) or pod2usage("error: cutoff ratio mfe is percentage and must be an integer between 0 and 100");

if ($i_debug && $i_stdout) 
{
	pod2usage("error: at most one of debug and stdout switches can be on");
}

my $activated_formats = 0;
if ($i_clustal)   {$activated_formats++;}
if ($i_stockholm) {$activated_formats++;} 
if ($i_aln_fasta) {$activated_formats++;}

($activated_formats == 1) or  pod2usage("error: only one of reading format should be turned on");
if ($i_clustal)
{
	print STDERR "Warning:\tread_clustal option is experimental, \n\t\tclustal files would be treated as stockholm files without a consensus strcuture\n";
	($i_petfold) or pod2usage("Error read_clustal only works with pet fold prediction, use_petfold would be mandatory\n");
}

# Temporary directory is either specified by the user or an empty, unique temporary directory is created at runtime.
my $tmpdir;
if ($i_tmp_dir) {
  if (-d $i_tmp_dir) {
    $tmpdir = $i_tmp_dir;
  } else {
    pod2usage("Error: Given temporary directory does not exist: $i_tmp_dir\n");	
  }
} else {
  # Keep newly created temp dir if this was specified by the user. CLEANUP decides if temporary directory is deleted after execution.
  # FIXME Double check that this acutally works.
  if ($i_keep_tmp) {
    $tmpdir = tempdir(CLEANUP => 0);
  } else {
    $tmpdir = tempdir(CLEANUP => 1);	
  }
}
print "Using the following temporary directory: $tmpdir\n";

###############################################################################
# Overwrite global CONFIG with local configuration
###############################################################################
my $CONFIG_REF;
if ($i_local_config) {
    $CONFIG_REF = Config::AlignClustConfig::load_config($i_local_config);
} else {
    $CONFIG_REF = Config::AlignClustConfig::load_config();
}
my %CONFIG = %{$CONFIG_REF};
if ($i_debug) {
	ReadConfigFile::print_config(\%CONFIG);
}

(defined $i_sge_submit_host) or $i_sge_submit_host = $CONFIG{SGE_HOST};
(defined $i_sge_submit_host || !$i_sge) or pod2usage("error: SGE is used but sge-submit-host is not defined!\n");

# set the appropriate exports for old and new cluster respectively
##This should be defined using config file for usage on other systems
  my $SGE_BIN_PATH=$CONFIG{SGE_BIN_PATH};
  my $SGE_EXPORTS=$CONFIG{SGE_EXPORTS};
  my $PETFOLD_BIN=$CONFIG{PETFOLD_BIN};

my $CURRUSER = getlogin;
(defined $i_sge_submit_user) and $CURRUSER=$i_sge_submit_user;
#TODO the followind should be in there
#if ($CURRUSER eq "") { die ("error: user name cannot be retrieved for host SGE, please specify username via -sge-submit-user option\n");}

my $num_jobs_per_submit = $CONFIG{NUM_JOBS_PER_SUBMIT};

###############################################################################
# end handler for temporary directory
# adds an error handler that deletes the directory in case of error
# SIGUSR{1/2} are sent by the sge prior to the uncatchable SIGKILL if the
# option -notify was set
###############################################################################
$SIG{'INT'}  = 'end_handler';
$SIG{'TERM'} = 'end_handler';
$SIG{'ABRT'} = 'end_handler';
$SIG{'USR1'} = 'end_handler';
$SIG{'USR2'} = 'end_handler';

sub end_handler {
  print STDERR "signal ", $_[0], " caught, cleaning up temporary files\n"; # TODO: check cleanup
  die();

  # change into home directory. deletion of the temporary directory will
  # fail if it is the current working directory
  chdir();
  File::Temp::cleanup();
  die();
}

###############################################################################
# EXECUTION CODE
###############################################################################
#
#my $string = "((((...((.))...(())...((.))...((..))(..)..))))";
#print $string."\n";
### remove lonely base-pairs
#$string =~ s/\.\(\./\.\.\./g;
#$string =~ s/\)\(\./\)\.\./g;
#$string =~ s/\.\)\./\.\.\./g;
#$string =~ s/\.\)\(/\.\.\(/g;
## remove base-pairs that result in a 0,1,2-sized hairpin loop
#$string =~ s/\(\)/../g;
#$string =~ s/\(\.\)/\.\.\./g;
#$string =~ s/\(\.\.\)/\.\.\.\./g;
#print $string."\n";
#die "testing\n";


#------
my $num_erroneous_seq = 0;


my $num_files;
my @aln_files;
my @cons_struct_files;

if($i_clustal) 
{
	@aln_files = recursively_return_files_in_dir(abs_path($i_dir), $CLUSTAL_SUFFIX);
}
if ($i_stockholm) #read stockholm input
{
	@aln_files = recursively_return_files_in_dir(abs_path($i_dir), $STOCKHOLM_SUFFIX);
}

if ($i_aln_fasta)
{
	@aln_files = recursively_return_files_in_dir(abs_path($i_dir), $ALN_FASTA_SUFFIX);
}
$num_files = @aln_files;

if (!($i_stockholm || $i_clustal || $i_aln_fasta) && !$i_petfold)
{
	@cons_struct_files = recursively_return_files_in_dir(
						abs_path($i_dir), $CONSENSUS_SUFFIX);
	$num_files = @cons_struct_files;
}


die "No recognized files in the input directory $i_dir\n" if ($num_files == 0);
## iterate over consensus structures for each RFAM dir	
my $grouped_SGE_options="";
my @submitted_jobs;
foreach my $i (0..($num_files-1))
{
	#######################
	## check for duplicates(e.g. in recursive directories)
	my @parts_cons = get_path_base_suffix($aln_files[$i]); 
	my $file_path = $parts_cons[0];
	my $file_name = $parts_cons[1];
	my $file_extension = $parts_cons[2];

	##filter rfam id 
	my $rfam_id = "";
	if($file_name =~ /(RF\S+)\.alignment/){
		$rfam_id = $1;
	} else {
		$rfam_id = $file_name;
	}
	## check whether this RFAM ID has already been seen!
	if(exists($FAMILIES{$rfam_id})){
		die "ERROR: this RFAM family $rfam_id has occured twice!\n";
	} else {
		$FAMILIES{$rfam_id} = "";
	}
	
    #######################	
	
	my $invoke_options = "-input-aln-file=$aln_files[$i] ";
    $invoke_options .= "-input-cons-file=$cons_struct_files[$i] " if (!($i_stockholm || $i_clustal || $i_aln_fasta) && !$i_petfold);
   
	$invoke_options .= "-stdout "								if ($i_stdout);
    $invoke_options .= "-local-config=$i_local_config "           if ($i_local_config);     
    $invoke_options .= "-debug "                                if ($i_debug);
	$invoke_options .= "-result-dir=$i_result_dir "				if ($i_result_dir);
	$invoke_options .= "-tmp-dir=$tmpdir "; # Tmp directory is either user-specified or automatically created
	$invoke_options .= "-cutoff-ratio-consensus=$i_cutoff_cons "	if ($i_cutoff_cons);
	$invoke_options .= "-cutoff-ratio-MFE=$i_cutoff_mfe "		if ($i_cutoff_mfe);
	$invoke_options .= "-use-petfold "							if ($i_petfold);
	$invoke_options .= "-read-stockholm "						if ($i_stockholm);
	$invoke_options .= "-read-clustal "							if ($i_clustal);
	$invoke_options .= "-read-aln-fasta "                       if ($i_aln_fasta);
	$invoke_options .= "-threshold-reliablity=$i_threshold_reli "	if ($i_threshold_reli);
	if ($i_sge)
	{	
		
		$grouped_SGE_options .= "\"$invoke_options\" ";
		## submit a single SGE job for the bin set of files, inluding last bin set
		if ((($i+1) % $num_jobs_per_submit) == 0 or $i+1 == $num_files) 
		{
			$i_sge_logDir = "$i_result_dir/SGE_log" if ( !$i_sge_logDir );
			mkdir($i_sge_logDir);

  			$i_sge_errDir = $i_sge_logDir if ( !$i_sge_errDir );
  			mkdir($i_sge_errDir);

    			my $syscall = "cd $CURRDIR; "
		   	. "$SGE_EXPORTS "
		   	. "$SGE_BIN_PATH/qsub -v PETFOLDBIN=\'$PETFOLD_BIN\' -o $i_sge_errDir/ -e $i_sge_errDir/ "
		   	. "$CLUSTERSUBMIT $CURRDIR $RUN_align_to_SS $grouped_SGE_options;";
    		say STDOUT "starting SGE job:";
    		say STDOUT $syscall;
    		my $qsub_out = `$syscall`;
    		my ($job_num) = $qsub_out =~ /Your job (\d+)/;  
    		print "qsub_out:$qsub_out\n";
    		print "job_num:$job_num\n";
    		push (@submitted_jobs, $job_num); 	
    			
			$grouped_SGE_options = "";
		}
	}
	else
	{
		print "local call:\n\t$CLUSTERSUBMIT $CURRDIR $RUN_align_to_SS \"$invoke_options\"\n";
		system("$CLUSTERSUBMIT $CURRDIR $RUN_align_to_SS \"$invoke_options\"");
	}
	select(undef, undef, undef, 0.1);  #sleep of 100 milliseconds
}
# Wait for submitted jobs to finish
if ($i_sge)
{
	print "\n";
	my $num_unfinished_jobs;
	do{
	my $qstat_out = `qstat`;
	my @lines = split /\n/, $qstat_out;
	$num_unfinished_jobs = 0;
	
		foreach my $line (@lines)
		{
			if (my ($job_num) = $line =~ /\s*(\d+)\s+.*\s+\d+\s*/){
				$num_unfinished_jobs++ if grep {$_ eq $job_num} @submitted_jobs;
			} 
		}
		print STDERR "\rWaiting for $num_unfinished_jobs SGE jobs to finish.." if ($num_unfinished_jobs > 0 );
	} while($num_unfinished_jobs > 0);
	
	print STDERR "All jobs done\n";
}	
print STDERR "NUMBER OF ERRONEOUS SEQUENCES: $num_erroneous_seq!\n" if ($num_erroneous_seq != 0);	

# Remove user-specified temporary directory if they should not be kept
if (defined $i_tmp_dir) {
  if (!$i_keep_tmp) {
    rmtree($tmpdir);
  }
}
