#!/usr/bin/perl

############################################################################
# This script runs PETfold on a fasta alignment
# Returns the consensus structure of the alignment with 
#   filterd base pairs based on their PETfold reliablities
# 
############################################################################

#use feature ':5.10';
use File::Basename;
use lib dirname($0);    # search skript directory for modules
use strict 'vars';
use warnings;
use Getopt::Long;
use Pod::Usage;
use List::Util qw/ min max /;


use Cwd qw(getcwd abs_path);
use File::Temp qw(tempdir);
use File::Copy;
use POSIX qw(ceil);

use FindBin;                 # locate this script
use lib "$FindBin::Bin/../";  # use the parent directory
use Config::AlignClustConfig;

use lib "$FindBin::Bin/../lib/";  # use the parent directory

use MultipleAlignment::PETfold;


##script name
my $SCRIPT_NAME = basename($0);


=head1 NAME


ss ssss dfd
=head1 SYNOPSIS

Options:

  HELP
  -help   brief help message
  -man    full documentation

COMPULSORY
  -align-fasta <STRING> alignment file name
  -threshold      <FLOAT> e.g. 0.001
  			Reliability threshold, report basepairs of predicted consensus structure above threshold
  			 
  -debug  additional debug output
  -local-config   if a local configuration is to be used, path to the local configuration file


  DEFAULT VALUES
  -threshold   0.1
=head1 DESCRIPTION

=cut

###############################################################################
# sub declrations
###############################################################################
#sub parse_real_pp;

###############################################################################
# PARSE COMMAND LINE OPTIONS
###############################################################################

# command line options
my (
  $i_help,               $i_man,          $i_debug, $i_local_config,
  $i_align_fas, $i_threshold, $i_consensus_out, $i_tmp_dir, $i_job_id
);

my ( $i_add_seq_graph_win, $i_add_seq_graph_t, $i_change_seq_graph_alph );
$i_threshold = 0.1;
my $options = GetOptions(
  "help"            => \$i_help,
  "man"             => \$i_man,
  "debug"           => \$i_debug,
  "local-config=s" => \$i_local_config,
  "align-fasta=s"   => \$i_align_fas,
  "working-dir=s"	=> \$i_tmp_dir,
  "consensus-out=s"	=> \$i_consensus_out,
  "threshold=f"     => \$i_threshold, 
  "job-id=s"		=> \$i_job_id
  );

pod2usage( -exitstatus => 1, -verbose => 1 ) if $i_help;
pod2usage( -exitstatus => 0, -verbose => 2 ) if $i_man;
($options) or pod2usage(2);

if ($i_local_config) {
  (-e $i_local_config) or pod2usage("error: local configuration file does not exist: $i_local_config\n");
}

# check compulsory options
($i_tmp_dir) or pod2usage("Error: no tmp directory has been defined\n");
(-e $i_tmp_dir) or pod2usage("Error: tmp directory does not exist!\n");
$i_tmp_dir = abs_path($i_tmp_dir);

($i_align_fas) or pod2usage("Error: the option -align-fasta is compulsory!\n");
($i_consensus_out) or pod2usage("Error: the option -consensus-out is compulsory!\n");
( -e $i_align_fas ) or pod2usage("Error: no such file - $i_align_fas!\n");
$i_align_fas = abs_path($i_align_fas);

# check other options and set default values

# set the appropriate exports for old and new cluster respectively

###############################################################################
# Overwrite global CONFIG with local configuration
###############################################################################
my $CONFIG_REF;
if ($i_local_config) {
    $CONFIG_REF = Config::AlignClustConfig::load_config($i_local_config);
} else {
    $CONFIG_REF = Config::AlignClustConfig::load_config();
}
my %CONFIG = %{$CONFIG_REF};

###############################################################################
# GLOBAL VARIABLES
###############################################################################
#my $PETFOLD_BIN="/home/miladim/software/PETfold/bin/PETfold";  
#my $RNADIST_BIN="/home/miladim/software/bin/RNAdistance";  
my $petfold_loc = $CONFIG{PETFOLD_BIN};

my $REL_FILE_NAME = "reliabilities.pp";
my $OUT_FOLDER = "/PETfold_output";
my $PET_LOG = "pet_log";
my $PET_SUBOPT_LOG = "pet_subopt_log";
my $SUBOPT_PETFOLD = " -o 10000 ";

###############################################################################
# end handler for temporary directory
# adds an error handler that deletes the directory in case of error
# SIGUSR{1/2} are sent by the sge prior to the uncatchable SIGKILL if the
# option -notify was set
###############################################################################
$SIG{'INT'}  = 'end_handler';
$SIG{'TERM'} = 'end_handler';
$SIG{'ABRT'} = 'end_handler';
$SIG{'USR1'} = 'end_handler';
$SIG{'USR2'} = 'end_handler';

sub end_handler {
  print STDERR "signal ", $_[0], " caught, cleaning up temporary files\n";

  # change into home directory. deletion of the temporary directory will
  # fail if it is the current working directory
  chdir();
  File::Temp::cleanup();
  die();
}

###############################################################################
# EXECUTION CODE
###############################################################################

# set up tmp directory

$OUT_FOLDER .= "_$i_job_id" if ($i_job_id);

$OUT_FOLDER = $i_tmp_dir . $OUT_FOLDER;

##create out directory if necessary
system "mkdir $OUT_FOLDER" if (! (-e $OUT_FOLDER) );
$OUT_FOLDER = abs_path($OUT_FOLDER);

##call PETfold with parameters
my $params_PETfold = "-f $i_align_fas -r $OUT_FOLDER/$REL_FILE_NAME ";#$SUBOPT_PETFOLD" ;
my $call_PETfold = "$petfold_loc/PETfold $params_PETfold";
print STDOUT "call_PETfold: $call_PETfold\n";
system ("$call_PETfold > $OUT_FOLDER/$PET_LOG");

### parse PETfold suboptimals
#my $parser_pipe = '| awk \'{if($1=="PETfold"){mea=$4};if($1~/^Score/){print mea,$3};if($1=="Suboptimal"){print $3,$4}}\' | tr - . | sort -k2gr ';
#print "parser_pipe: $parser_pipe\n";
#system ("cat $OUT_FOLDER/$PET_LOG $parser_pipe > $OUT_FOLDER/$PET_SUBOPT_LOG"); 
#
### calculate RNA structure distances
#my $call_RNAdist = "$RNADIST_BIN -Xf -Dh < $OUT_FOLDER/$PET_SUBOPT_LOG " . ' | awk \'{print $2,NR}\' | sort -k1gr | awk \'{print NR,$2,NR+$2}\' | sort -k3g | cut -d" " -f2 | head -40 ' ;#| head -10';
#system ("$call_RNAdist");
#system ("sed -n '5p' $OUT_FOLDER/$PET_SUBOPT_LOG");
#
## ~/software/bin/RNAdistance -Xf -Dh < PETfold_output/pet_subopt_log | awk '{print NR,$2}' | head -5

##R script
#data_rel_dist <- read.table(file="/home/milad/workspace/AlignClust/PETfold_output/pet_rel_dist_subs")
#pc <- princomp(data_rel_dist[,2:3], cor=TRUE, scores=TRUE) 
#ks = kmeans(pc$scores[,1:2],4)

## parse PETfold reliabilities
my $struct = MultipleAlignment::PETfold::parse_outlog ("$OUT_FOLDER/$PET_LOG");
my $reliabilities = MultipleAlignment::PETfold::parse_reali_pp ("$OUT_FOLDER/$REL_FILE_NAME");
my $filtered_struct = MultipleAlignment::PETfold::filter_structure($struct, $reliabilities);
MultipleAlignment::PETfold::write_struct_to_file($i_consensus_out, $filtered_struct);
print "filtered_struct is:\n$filtered_struct\n" if ($i_debug);




