#!/usr/bin/env python

"""
    run_rnascclust
    ~~~~~~~~~~~~~
    Wrapper module to invoke the RNAAlignClust pipeline

@author:    Milad Miladi, Alexander Junge
            miladim@informatik.uni-freiburg.de, ajunge@rth.dk
            
    Sample invocation call command:
    $ python bin/run_rnascclust.py --alignments-dir data/cliques-low/alignments/ \
      --predict-structure  --result-path . --local-config Config/bilp13.config
"""

import os
import sys
import shutil
import argparse
import bz2
import glob
import subprocess
import string
from itertools import izip
from Bio import AlignIO, SeqIO
from Bio.Alphabet import IUPAC
from Bio.Align import MultipleSeqAlignment

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


BIN_PATH = os.path.dirname(os.path.realpath(__file__))

import signal
__my_child_pid=None
import signal
def signal_handler(sig, frame):
        print('You pressed Ctrl+C!')
        print('child pid was', __my_child_pid)
        os.kill(__my_child_pid, signal.SIGTERM)
        sys.exit(0)


        
signal.signal(signal.SIGINT, signal_handler)

def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist" % arg)
    return arg

def is_valid_file(arg):
    if not os.path.isfile(arg):
        parser.error("Given file %s does not exist" % arg)
    return arg


def get_alignments(path_stks, min_class_size, target_dir):
    if os.path.isdir(path_stks) is False:
        logger.error("extract_representatives: stk path not found: {}".format(path_stks))
    
    os.mkdir(target_dir)

    nonfiltered_stk_files = glob.glob(path_stks+"/*.stk")
    # sort files such that we get the order as bash ls command. RF00005.0, RF00005.10, RF00005.11, RF00005.1 
    alphabet = string.printable  # 0-9A-Za-z_-. 
    nonfiltered_stk_files = sorted( nonfiltered_stk_files, 
                                    key=lambda word: [alphabet.index(c) for c in word+"..."])

    # Count size of the each class, stk_file base name must be class label
    labels_count = dict()
    for stk_file in nonfiltered_stk_files:
        label = os.path.basename(stk_file).split('.')[0] # Filename ohne extension!
        labels_count[label] = labels_count.get(label, 0) + 1  # Initialize/increment key value
#     from collections import Counter
#     print sorted(labels_count, key=labels_count.get) #lambda var: labels_count[var])
#     print Counter(labels_count)
#     print Counter(labels_count.values())
#     total = sum(v for v in labels_count.values() if v >= 1)
#     min_3 = sum(v for v in labels_count.values() if v >= 3)
#     min_5 = sum(v for v in labels_count.values() if v >= 5)
#     print "Number of families: Total: {}, min3: {}, min_5: {}, loss-3-to-5: {}%".format(total, min_3, min_5, 100*float(min_3 -min_5)/min_3)
    
    # Extract alignments that their reference label contains minimum number of alignments
    stk_files = [stk for stk in nonfiltered_stk_files 
                 if labels_count[os.path.basename(stk).split('.')[0]] >= min_class_size]
    
    print "Stk files reduced from", len(nonfiltered_stk_files), len(stk_files)
    # Copy filtered files into target directory under RESULTS_AC
    for f in stk_files:
        shutil.copy(f, target_dir)
        
    copied_stk_files = glob.glob(target_dir+"/*.stk")
    return sorted(copied_stk_files,
                  key=lambda word: [alphabet.index(c) for c in word+"..."])

def generate_representatives (stk_files, out_fasta):
    '''Extracts a fasta file of representatives from all the alignments 
    available under the given path.
    Assumes first sequence of the alignment is the representative of the alignment'''
    
    # Collect first seqs of all alignments
    alignment_list = []        
    for stk_file in stk_files:
        stk_handle = open (stk_file, 'r')
        #dependent on the assumption that read() keeps the input file order
        # TODO: Check alphabet
        alignment = AlignIO.read(stk_handle, "stockholm", alphabet=IUPAC.ambiguous_rna)  
        assert (len(alignment) > 0)
        rep = alignment[0]
        # TODO remove gaps
        rep.description = "{} {}".format(rep.id, 
                                         os.path.splitext(os.path.basename(stk_file))[0])
        gap_chars = '[-.~_]'  # valid gap symbols all be converted to "-"
        for gap in gap_chars:
            rep.seq = rep.seq.ungap(gap)

        alignment_list.append(rep)
        stk_handle.close()
    
    # Write single line sequences inside fasta file 
    from Bio.SeqIO.FastaIO import FastaWriter
    out_handle = open(out_fasta, 'w')
    writer = FastaWriter(out_handle, wrap=0)
    writer.write_file(alignment_list)
    out_handle.close()
#     with open(out_fasta, "w") as output_handle:
#         SeqIO.write(alignment_list, output_handle, "fasta")
#     
    logger.info("{} representatives were extracted".format(len(alignment)))

def sanity_check(stk_files, represent_fasta):
    '''Verifies alignments and the representative fasta have same size and order'''
    
    with open(represent_fasta, "r") as in_fasta_handle:
        fa_descs = [rec.description for rec in SeqIO.parse(in_fasta_handle, "fasta")]
    
    # Count size of the each class, stk_file base name must be class label
    labels_count = dict()
    for stk_file in stk_files:
        label = os.path.basename(stk_file)
        labels_count[label] = labels_count.get(label, 0) + 1  # Initialize/increment key value
        
        # Check alignments can be parsed by biopython and are not empty
        stk_handle = open (stk_file, 'r')
        #dependent on the assumption that read() keeps the input file order
        alignment = AlignIO.read(stk_handle, "stockholm")
        assert (len(alignment) > 0)
        stk_handle.close()
    #end for 
    
    if len(fa_descs) != len(stk_files):
        logger.error("Number of alignment files:{} and sequences of representative "\
        "fasta:{} must match ".format(len(stk_files), len(fa_descs)))
        raise AssertionError
    
    
    # Fasta descriptors (second entrY OF each fasta header) should match to stk filename 
    for desc, file in izip(fa_descs, stk_files):
        file = os.path.splitext(os.path.basename(file))[0]
        
        if (len(desc.split()) < 2) or (desc.split()[1] != file):
            logger.error("Mismatch between representative fasta and alignment files")
            logger.error("Fasta-descriptor:{} alignment-file:{}".format(desc, file))
            raise AssertionError
    #end for
    logger.info("Sanity check of the representatives passed!")
#         stk_handle = open (stk_file, 'r')
        #dependent on the assumption that read() keeps the input file order
#         alignment = AlignIO.read(stk_handle, "stockholm") 

# =========================================================================
#
# =========================================================================
print "Args: ", sys.argv
parser = argparse.ArgumentParser()
'''
        Options from perl script not integrated
        
                         "stdout"      => \$i_stdout,
                         "cutoff-ratio-consensus=i" => \$i_cutoff_cons,
                         "cutoff-ratio-MFE=i" => \$i_cutoff_mfe,
                         "read-stockholm" => \$i_stockholm,
                         "read-clustal" => \$i_clustal,
                         "read-aln-fasta" => \$i_aln_fasta,
                         "threshold-reliablity=f" => \$i_threshold_reli,
                         );

'''
argparse.ArgumentParser(description="""RNAAlignClust Pipeline """)
                
parser.add_argument('--alignments-dir', "-i", type=is_valid_path, required=True,
                    help='Path to the directory containing stockholm files')
parser.add_argument('--result-path', type=is_valid_path, required=True,
                    help='Path to creat output directory RESULTS_AC/ under\
                    it stores intermediate and final results ')
parser.add_argument('--local-config', type=is_valid_file,
                    help='if a local configuration is to be used, path to the \
                    local configuration file')
parser.add_argument('--sge-parallel', action='store_true', default=False,
                    help='Run in parallel multicore/SGE mode ')

parser.add_argument('--rep-fasta', type=is_valid_file, required=False,
                    help='A fasta file containing one representative sequence \
                    from each alignment under the input-directory.\
                    Representative could be e.g. the human sequence of each alignment')

parser.add_argument('--temp-path', type=is_valid_path, required=False,
                    default='/tmp/',  # TODO: find the host temp path in generic way
                    help='Path to store intermediate temporary direcotry named TEMP_AC ')

parser.add_argument('--structure-is-given', action='store_true', default=False,
                    help='Do not use PETfold to predict the secondary consensus "\
                    structure of each alignment')
parser.add_argument('--predict-structure', dest='structure_is_given', action='store_false',
                    help='Use PETfold to predict the secondary consensus \
                    structure of each alignment')
parser.add_argument('--keep-temp', '-k', action='store_true',
                    help='Keep files after run finished')
parser.add_argument('--min-class-size', type=int, default = 3,
                    help='Discard families where number of subalignments is less than min')

parser.add_argument('--debug', '-d', action='store_true',
                    help='Debug mode, verbose information logged')
args = parser.parse_args()



# =========================================================================
# 
# =========================================================================
results_SC_dir = args.result_path.rstrip('/') + '/RESULTS_SC/'


logger.info("Remove old result files")
logger.info("Creating and cleaning up the results directory")
try:
    if os.path.isdir(results_SC_dir):
        shutil.rmtree(results_SC_dir)
    os.mkdir(results_SC_dir)    
except:
    logger.error("Unable to modify/remove content of the result directory with path %s" % results_SC_dir)
    raise



temp_AC_dir = args.temp_path.rstrip('/') + '/TEMP_AC/'
try:
    if os.path.isdir(temp_AC_dir):
        shutil.rmtree(temp_AC_dir)
    os.mkdir(temp_AC_dir)    
except:
    logger.error("Unable to modify/remove content of the temp directory with path %s" % temp_AC_dir)
    raise

# Extract list of alignment files, filter them if needed

extracted_alignments = get_alignments(args.alignments_dir, args.min_class_size, 
                                      results_SC_dir + "INPUT/" )
represent_fasta_file = results_SC_dir + 'representatives.fa' # Todo extract human

# If representatives are provided by user
if args.rep_fasta:
    try:
        shutil.copy(args.rep_fasta, represent_fasta_file)
    except:
        logger.error("Unable to copy provided fasta into results directory with path %s" % represent_fasta_file)
        raise
        
else:  # otherwise extract a sequence of each alignment as rep
    logger.warning("Fasta file for representative sequences is not provided,\
        considering the first sequence of each alignment as its representative!")
    generate_representatives(extracted_alignments, represent_fasta_file)

# Verify the provided or generated representatives against the alignment files 
sanity_check(extracted_alignments, represent_fasta_file )


# =========================================================================
#
# =========================================================================
logger.info("Generate Graph Encoded Alignments")
''' Example:
perl bin/alignment_to_secondary_structures.pl -input-aln-file=RF02106_filtered.2 
-debug -result-dir=. -tmp-dir=./tmp/ -cutoff-ratio-consensus=20 -cutoff-ratio-MFE=20 
-use-petfold -read-stockholm -threshold-reliablity=0.001 --local-config=Config/bilp13.config
'''

perl_args = ["--input-dir", results_SC_dir+"INPUT/",
              "--result-dir", results_SC_dir,
              "--tmp-dir",  temp_AC_dir,
              "--use-petfold" if not args.structure_is_given else "",
              "--keep-temp" if args.keep_temp else "",
              "--debug" if args.debug else "",
              "--sge" if args.sge_parallel else "",
              "--read-stockholm"
              ]
if args.local_config:
    perl_args += ["--local-config", args.local_config]

print perl_args


# Run gspan graph encoder scripts
with open(results_SC_dir+'phase1-graph_encoder.stdout.log', 'w') as logfile:
    encoder_pipe = subprocess.Popen(["perl", BIN_PATH + "/iterate_over_alignments.pl"] + perl_args, 
                     stdout=logfile#subprocess.PIPE#, stderr=subprocess.PIPE #logfile,                    preexec_fn=os.setsid
                     )
    __my_child_pid = encoder_pipe.pid
    encoder_pipe.communicate()
        
    

if len(glob.glob(results_SC_dir + "GSPAN/*.bz2")) == 0:
    logger.error("No gspan Encoded Graph was generated from the first phase, Exit..")
    sys.exit()

logger.info("Graph encoding Alignment finished")

gspan_file = results_SC_dir + "encoded_alignments.gspan"

if os.path.isfile(gspan_file):  # Remove old gspan file if exist
    os.remove(gspan_file)
    
# Concatenate all generated gspan graphs into a single file
# subprocess.call("cat {}GSPAN/*.bz2 > {}.bz2".format(results_SC_dir, gspan_file), shell=True)

# Decompress the bz2 gspan files and write them into a single file
with open(gspan_file, 'wb') as outfile:
    for bz2file in sorted(glob.glob(results_SC_dir + "GSPAN/*.bz2")):
        with bz2.BZ2File(bz2file, 'rb') as infile:
            for chunk in iter(lambda : infile.read(100 * 1024), b''):
                outfile.write(chunk)

# ==================================== =====================================
# NSPDK creates sparse feature vectors
# =========================================================================
logger.info("Generate Sparse Feature Vector for the Graph Encoded Alignments")
nspdk_path = BIN_PATH + '/GraphClust/NSPDK'
nspdk_args =  " -R 3 -D 3  -gt DIRECTED -of -fg %s" % gspan_file # Todo use config file params

print [nspdk_path, nspdk_args ]
vectorizer_pipe = subprocess.call (nspdk_path + nspdk_args, shell=True)

# =========================================================================
# Hybrid pipeline Candidate CLuster Selection and Iterative clustering
# =========================================================================
logger.info("Candidate CLuster Selection and Iterative clustering")

hybrid_path = BIN_PATH + '/GraphClust/MASTER_Hybrid_debug.pl'
hybrid_args = ['--fasta',  represent_fasta_file,
               '--root-dir', results_SC_dir,
               '--feature-vector', gspan_file + ".feature_bin",
               '--config', BIN_PATH + '/GraphClust/config.cnf',
               '--no-ssh',
               '--sge'  if args.sge_parallel else ""
               ]

print ["perl", hybrid_path] + hybrid_args

hybrid_pipe = subprocess.call (["perl", hybrid_path] + hybrid_args )

cluster_dirs = results_SC_dir+"/RESULTS/[0-9]*"
# One liner collector of the clustering results
collector_cmd = "for c in " + cluster_dirs + "; do cat $c/cluster.all | awk -v c=$(basename $c) '{print c,$11}' | sort; done "
collector_cmd += " > " +results_SC_dir+"clustering_results_overall.txt"
exit_code = os.system(collector_cmd)
if (exit_code != 0):
    logger.error("Failure in collecting cluster results")
    raise
# with open(results_SC_dir+'rnascclust_result.txt', 'w') as resultfile:
#     cluster_dirs = results_SC_dir+"/RESULTS/[0-9]*"
#     collector_cmd = "for c in " + cluster_dirs + "; do cat $c/cluster.all | awk -v c=$(basename $c) '{print c,$11}' | sort; done"
#     process = subprocess.Popen(collector_cmd.split(), shell=True, stdout=resultfile)
#     process.communicate()


