package MultipleAlignment::Tools;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(read_stockholm
map_consensus_structure_to_single_alignment_sequence
			);
#@EXPORT_OK = qw(			);

#StructureLibrary::Structure::LOAD_CONFIG();
##############

#######################################################################
# Converts a clustal file to stk format
# INPUT: clustal file, stk file
# OUTPUT: 
#######################################################################
sub convert_clustal_to_stk
{
	my $clust_file = shift;
	my $stk_file = shift;
	my $FUNCTION = "convert_clustal_to_stk ";
	open(IN_HANDLE, "<$clust_file") || die "ERROR in $FUNCTION:\n Couldn't open file $file/n";
	open(OUT_HANDLE, ">$stk_file") || die "ERROR in $FUNCTION:\n Couldn't open file $file/n";
	
	print OUT_HANDLE "# STOCKHOLM 1.0";
	if ( <IN_HANDLE> !~ /^CLUSTAL/ ) {
    	die("Input not in CLUSTAL format. Missing header starting with CLUSTAL\n");
	}
	while(my $line = <IN_HANDLE>){
		print OUT_HANDLE $line;
	}
	close IN_HANDLE;
	close OUT_HANDLE;
}

#######################################################################
# Parse a clustal file
# the input order of tags and seuqences does NOT CHANGE
# non uniqueness of tags is supported
# INPUT: 
# OUTPUT: 
#######################################################################
sub parse_clustal_aref{
	my $file = shift;
	my $FUNCTION = "parse_clustal ";
	
	open(IN_HANDLE, "<$file") || die "ERROR in $FUNCTION:\n Couldn't open file $file/n";
	
	my %alignments = ();
	my @inorder_tags = ();
	while(my $line = <IN_HANDLE>){
		chomp($line);
		
		# get entry line and add it to alignments hash
#		if ($line =~ /^(\S+\/\d+-\d+)\s+(\S+)$/){ Rfam based headers is not necessary 
		if ($line =~ /^(\S+)\s+(\S+)$/){
			if(exists ($alignments{$1})){
				$alignments{$1} .= $2;
			} else {
				$alignments{$1} = $2;
				push (@inorder_tags, $1);
			}
		}
	}
	## copy hash contents to an array
	my @inorder_seqs = map {$alignments{$_}} @inorder_tags;
	close IN_HANDLE;
	
	return (\@inorder_tags, \@inorder_seqs);
}

############################################################
##
## Parse stockholm format.
##
## @param file name (with path)
## @param only_GC return only GC and GF markups and discard others
##
## @returns (\%seqs, \%markups, \@names)
##
############################################################
sub read_stockholm {
  my $stk_file_name = shift;	
  my $only_GC = shift;
  open ($fh,"<$stk_file_name" ) or die "error: couldn't open stockholm file $stk_file_name for reading";
  
  my %markups = ();    # result record
  my $header = <$fh>;
  if ( $header !~ /^# STOCKHOLM 1.0$/ ) {
    die("Input not in Stockholm 1.0 format. Missing header \"# STOCKHOLM 1.0\".Found $header\n");

    #return 0;
  }

  my %seqs  = ();    ## hash of sequences
  my @names = ();    ## list of sequence names

  while ( my $line = <$fh> ) {
    if ( $line =~ /^\/\/(\s)*$/ ) { ## end reading at // marker
      last;
    } elsif ( $line =~ /^\s+$/ ) {    ## skip empty lines
                                      # skip
    } elsif ( $line =~ /^#=(G[FC])\s(\S*)\s+(.*)$/ ) { ## handle per column and per residue information
      my $tag = $1;
      next if ($tag ne "GC" and $only_GC); #skip non GC tags if only_GC is set

      my $feature = $2;
      my $val = $3;
      chomp $val;
      if(! exists $markups{$tag}{$feature} )
      {
      	 $markups{$tag}{$feature} = "";
      }
      $markups{$tag}{$feature} .= $val;
    } 
    elsif ($line =~ /^#=(G[SR])\s(\S+)\s+(.*)$/ ) { ## handle per column and per residue information
      if (!$only_GC) {	
	my $tag = $1;
	my $name = $2;
	my $feature_val = $3;
	chomp $feature_val;
	$markups{$tag}{$name} = $feature_val;
      }
    }
    elsif ( $line =~ /^(\S+)\s+(.*)/ ) {    ## handle sequence information
      my $name = $1;
      my $sequence = $2;
      chomp $sequence;
      if ( !exists $seqs{$name} ) {
        push @names, $name;
        $seqs{$name} = "";
      }
    #  print "name:$name, sequence:$sequence\n";
      $seqs{$name} .= $sequence;
    }
  }

close $fh;
   ## log & debug	
#  print "read_stockholm:\n"  ; 
#  foreach my $tags (keys %markups) 
#  {
#  	foreach (keys $markups{$tags})
#  	{
#  		print "markups{$tags}{$_}=" . $markups{$tags}{$_} ."\n";
#  	}
#  }
#  foreach (keys %seqs) {print "??? $_ ???\n";}
  return (\%seqs, \%markups, \@names);
}

#######################################################################
# Projects secondary struture to 
# INPUT:
#	secondary structure in dotbracket format
#	sequence 
# OUTPUT: 
#	pruned structure which is compatible with the sequence + sequence + bp_ratio
#######################################################################
sub map_consensus_structure_to_single_alignment_sequence{
	my ($consensus, $ali_seq) = @_;
	
	my $structure = $consensus;
	my $seq = $ali_seq;
	
	my $FUNCTION = "map_consensus_structure_to_single_alignment_sequence in ".
		"consensus_to_single_structures.pl";
	
	## check input parameters
	die "the the consensus structure in dot-bracket format and the alignment ".
		"sequence are compulsory in $FUNCTION" unless ($structure && $seq);
	## check that the lengths are the same
	my $structure_len = length($structure);
	my $seq_len = length ($seq);
	die "The consensus structure has a different length than the alignment sequence!\n".
		"$seq\n$structure\n" unless ($structure_len == $seq_len);
		

	## remove base-pairs that are incompatible with the sequence
	## first convert base-pairs into the hash format
	my $struct_href = StructureLibrary::Structure::structure2basepairs($structure,
		1, "().-", 1);

	my $num_bp_very_initial = keys (%{$struct_href});

	## remove non-compatible base-pairs from hash
	StructureLibrary::Structure::remove_non_compatible_bps_wrt_seq($seq, $struct_href);
	## convert back to dot-bracket notation
	$structure = StructureLibrary::Structure::basepairs2dotbracket($struct_href, $structure_len);
	
#	print STDERR "$structure <-- after removing incomp base-pairs\n";
	
	## remove all columns with gaps in the alignment
	for my $i (0 .. ($structure_len - 1)){
		
		if((substr($seq, $i, 1) eq '.') || (substr($seq, $i, 1) eq '-')){
			substr($seq, $i, 1, '#');
			substr($structure, $i, 1, '#');
		}
	}
	$seq =~ tr/#//d;
	$structure =~ tr/#//d;
	
	print STDERR "$seq <-- after removing gaps\n";
	print STDERR "$structure <-- after removing gaps\n";
	
	## re-calculate the sequence length
#	$len = length($seq);
	

	## remove unstable or impossible structures, such as lonely
	## base pairs and hairpin loops with 0-2 unpaired bases
	$structure = recursively_remove_impossible_structure_constructs($structure);
	
	## final check
	my $shref = StructureLibrary::Structure::structure2basepairs($structure,
		1, "().-", 1);
	## remove non-compatible base-pairs from hash
	StructureLibrary::Structure::check_bps_wrt_seq($seq, $shref);
	
	## check lengths of structure and sequence again
	$structure_len = length($structure);
	$seq_len = length($seq);
	die "After mapping: The consensus structure has a different length than the alignment sequence!\n".
		"$seq\n$structure\n" unless ($structure_len == $seq_len);
	
	## calculate a ratio of how much of the consensus structure has been deleted
	my $num_bp_before = keys (%{$struct_href});
#	print "num_bp_very_initial: $num_bp_very_initial\tnum_bp_before: $num_bp_before\n";
	my $num_bp_after = keys (%{$shref});
	my $ratio = -1; # the ratio is -1 if there was no consensus structure in the first place
	#$ratio = $num_bp_after / $num_bp_before if ($num_bp_before);
	$ratio = $num_bp_after / $num_bp_very_initial if ($num_bp_very_initial);
	
	return ($structure, $seq, $ratio);
}


 
####
# The method removes lonely base-pairs and also hairpin loops of
# sizes 1 to 2 unpaired bases within the loop
sub recursively_remove_impossible_structure_constructs{
	my $structure = shift;
	my $FUNCTION = "recursively_remove_impossible_structure_constructs ".
					"in consensus_to_single_structures.pl";
	my $cleaned_struct = $structure;

	my $str_href = StructureLibrary::Structure::structure2basepairs($structure,
		1, "().-");
	
	my %map_open_to_close = ();
	my %map_close_to_open = ();
	
	my ($i, $j) = -1;
	foreach (keys(%{$str_href})){
		if (($i, $j) = split(":", $_)){
			if($i < $j){
				$map_open_to_close{($i-1)} = $j - 1;
				$map_close_to_open{($j-1)} = $i - 1;
			} elsif ($i > $j) {
				print STDERR "WARNING in $FUNCTION: The base-pair hash is in the".
					" wrong format as i is greater than j in the key. Swapping i ".
					"and j.\n";
					$map_open_to_close{($j-1)} = $i - 1;
					$map_close_to_open{($i-1)} = $j - 1;
			} else {
				die "ERROR in $FUNCTION: i cannot be equal j!\n";
			}
		} else {
			die("ERROR in $FUNCTION:\nThe base-pair keys in hash have the wrong format $_\n");
		}
	}
	
	
	## remove lonely base-pairs
	my $index = -1;
	my $open = -1;
	my $close = -1;
	my $len = length($cleaned_struct);
	## remove opening base in ".(." and the corresponding closing base
	$index = index($cleaned_struct, ".(.");
	if($index != -1){
		$open = $index + 1;
		if(exists($map_open_to_close{$open})){
			$close = $map_open_to_close{$open};
			substr($cleaned_struct, $open, 1, ".");
			substr($cleaned_struct, $close, 1, ".");
		}else {
			die "Cannot find opening base $open in $cleaned_struct!\n";
		}
		$index = -1;
	}
	## remove opening base in ")(." and the corresponding closing base
	$index = index($cleaned_struct, ")(.");
	if($index != -1){
		$open = $index + 1 ;
		if(exists($map_open_to_close{$open})){
			$close = $map_open_to_close{$open};
			substr($cleaned_struct, $open, 1, ".");
			substr($cleaned_struct, $close, 1, ".");
		}else {
			die "Cannot find opening base $open in $cleaned_struct!\n";
		}
		$index = -1;
	}
	## remove closing base in ".)." and the corresponding opening base
	$index = index($cleaned_struct, ".).");
	if($index != -1){
		$close = $index + 1;
		if(exists($map_close_to_open{$close})){
			$open = $map_close_to_open{$close};
			substr($cleaned_struct, $open, 1, ".");
			substr($cleaned_struct, $close, 1, ".");
		}else {
#			foreach (keys (%map_close_to_open)){
#				print "close $_ and open $map_close_to_open{$_}\n";
#			}
			die "Cannot find closing base $close in $cleaned_struct!\n";
		}
		$index = -1;
	}	
	## remove closing base in ".)(" and the corresponding opening base
	$index = index($cleaned_struct, ".)(");
	if($index != -1){
		$close = $index + 1;
		if(exists($map_close_to_open{$close})){
			$open = $map_close_to_open{$close};
			substr($cleaned_struct, $open, 1, ".");
			substr($cleaned_struct, $close, 1, ".");
		}else {
			die "Cannot find closing base $close in $cleaned_struct!\n";
		}
		$index = -1;
	}
	## remove opening base in "^(." and the corresponding closing base
	elsif(substr($cleaned_struct, 0, 2) eq "(."){
		$open = 0;
		if(exists($map_open_to_close{$open})){ #resolved bug in Sita's code. exist was not invoked
			$close = $map_open_to_close{$open};
			substr($cleaned_struct, $open, 1, ".");
			substr($cleaned_struct, $close, 1, ".");
		}else {
			die "Cannot find opening base $open in $cleaned_struct!\n";
		}
	}
	## remove closing base in ".)$" and the corresponding opening base
	elsif(substr($cleaned_struct, ($len-2), 2) eq ".)"){
		$close = $len - 1;
		if(exists($map_close_to_open{$close})){#resolved bug in Sita's code. exist was not invoked
			$open = $map_close_to_open{$close};
			substr($cleaned_struct, $open, 1, ".");
			substr($cleaned_struct, $close, 1, ".");
		}else {
			die "**Cannot find closing base $close in $cleaned_struct!\n";
		}
	}
	## remove base-pairs that result in a 0,1,2-sized hairpin loop
	$cleaned_struct =~ s/\(\)/../g;
	$cleaned_struct =~ s/\(\.\)/\.\.\./g;
	$cleaned_struct =~ s/\(\.\.\)/\.\.\.\./g;
	
	if($structure eq $cleaned_struct){
		return $cleaned_struct;
	} else {
		recursively_remove_impossible_structure_constructs($cleaned_struct);
	}
}


1;
