# ReadConfigFile.pm
#
# Extracts configurations from RNAalignClust configuration files.
# Author: Alexander Junge <alexander.junge@gmail.com>

package ReadConfigFile;

use strict;
use warnings;

##########################################################

=item read_config_file()

Reads a configuration file and returns a hash mapping
parameters to values.

Input:
        $config_file_path - string - Path to configuration file

Output:
        Hash: string -> string
        Maps parameter name to value set in the configuration file.

=cut

###########################################################

sub read_config_file {
    my $config_file_path = shift;
    open my $config_handle, '<', $config_file_path or die "Cannot open local configuration file: $!";

    my %config_hash;
    my %str_formats_hash;
    my $str_formats_given = 0;
        
    while (my $line = <$config_handle>) {
        chomp($line);
        $line =~ /^\s*$/ and next; # skip empty lines containing
        $line =~ /^\s*#/ and next; # skip comment lines starting with #
          
        my ($key, $val) = split /=/, $line;
        #($val) = $val =~ /["']([^"']*)["']/; # removes single and double quotes
        
        # Slightly dirty workaround to read STR_FORMATS paramter into a hash
        if ($key eq 'STR_FORMATS_DOTBRACKET') {
            $str_formats_hash{'DOTBRACKET'} = $val;
            $str_formats_given = 1;
        } elsif ($key eq 'STR_FORMATS_CONSTRAINT') {
            $str_formats_hash{'CONSTRAINT'} = $val;
            $str_formats_given = 1;
        } else {
            $config_hash{$key} = $val;
        }
    }
    if ($str_formats_given) {
        $config_hash{'STR_FORMATS'} = \%str_formats_hash;
    }
    #my $str_hash_ref = $config_hash{'STR_FORMATS'};
    #my %str_hash = %$str_hash_ref;
    #warn "DOTBRACKET: " . $str_hash{'DOTBRACKET'};
    #warn "CONSTRAING: " . $str_hash{'CONSTRAINT'};
    return \%config_hash;
}


###########################################################

=item print_config()

Prints the current configuration to stderr.

Input:
        $config_hash_ref - hash reference
        Reference to the hash storing the current configuration.

=cut

###########################################################

sub print_config {
    my $config_hash_ref = shift;
    my %config_hash = %$config_hash_ref;
    
    warn "The following configuration is used:\n";
    foreach (keys %config_hash) {
        my $val = $config_hash{$_};
        warn $_ . "=" . (defined($val) ? $val : "<UNDEFINED>") . "\n";
    }
    warn "\n";
}


###########################################################

=item update_config()

Overwrites values set in the given default configuration with
values set in the given local configuration.

Input:
        $default_config_ref - hash reference
        Reference to the hash storing the default configuration
        
        $local_config_ref - hash reference
        Reference to the hash storing the local configuration

Output:
        hash reference
        Reference to the hash storing the default configuration
        with values taken from the local configuration.

=cut

###########################################################

sub update_config {
    my $default_config_ref = shift;
    my %default_config_hash = %$default_config_ref;
    my $local_config_ref = shift;
    my %local_config_hash = %$local_config_ref;
    
    foreach (keys %local_config_hash) {
        if (exists $default_config_hash{$_}) {
            $default_config_hash{$_} = $local_config_hash{$_};
        } else {
            die sprintf("Local configuration defines parameter %s which was not found " .
                        "in default configuration.\n", $_);
            # FIXME fix quite uninformative message above
        }
    }
    return \%default_config_hash;
}

1;